
#include <stdio.h>

#include <NDS.h>

#include "_const.h"

#include "plugin.h"
#include "plugin_def.h"

#include "libgif/gif_lib.h"

// CallBack from plugin_dll.cpp

void cbLoadLibrary(void)
{
}

void cbFreeLibrary(void)
{
}

void cbQueryInterfaceLibrary(void)
{
}

// -----------------

typedef struct {
  u32 Palettes[256];
  u32 PalettesCount;
  u32 Width,Height;
} TGIFInfo;

TGIFInfo GIFInfo;
u8 *pBitmap=NULL;
u32 BitmapWidth=0;

// ------------------------------------------------------------------------------------

static void _PrintGifError(void)
{
  _consolePrintf("GIF error:%s\n",GifLastError());
}

static void _PrintGifErrorStr(char *str)
{
  _consolePrintf("GIF error:%s\n",str);
}

extern "C" {
extern int readFunc_FileHandle;
}

int readFunc_FileHandle=0;

static int readFunc(GifFileType* GifFile, GifByteType* buf, int count)
{
  return(fread(buf,1,count,readFunc_FileHandle));
}

#define GAMMA(x)	(x)

bool LoadGIF(int FileHandle)
{
  readFunc_FileHandle=FileHandle;
  
  if(FileHandle==0){
    _PrintGifErrorStr("SourceData Null.");
    return(false);
  }
  
  {
    GifFileType *GifFile;
    
    _consolePrintf("OpenFile.\n");
    if ((GifFile = DGifOpen((void*)FileHandle, readFunc)) == NULL) {
      _PrintGifError();
      return false;
    }
    
    /* Scan the content of the GIF file and load the image(s) in: */
    GifRecordType RecordType;
    
    do {
      _consolePrintf("DGifGetRecordType\n");
      if (DGifGetRecordType(GifFile, &RecordType) == GIF_ERROR) {
        _PrintGifError();
        return false;
      }
      
      switch (RecordType) {
        case IMAGE_DESC_RECORD_TYPE:
          _consolePrintf("IMAGE_DESC_RECORD_TYPE\n");
          
          _consolePrintf("DGifGetImageDesc\n");
          if (DGifGetImageDesc(GifFile) == GIF_ERROR) {
            _PrintGifError();
            return false;
          }
          
          s32 Width=GifFile->Image.Width;
          s32 Height=GifFile->Image.Height;
          u32 PalettesCount;
          
/*
_consolePrintf("p:GifFile=%x\n",(u32)GifFile);
_consolePrintf("p:&Image.ColorMap=%x\n",&GifFile->Image.ColorMap);
_consolePrintf("p:&SColorMap=%x\n",&GifFile->SColorMap);
*/

          _consolePrintf("Update Color map\n",0);
          
          _consolePrintf("GifFile->Image.ColorMap = %x\n",(int)GifFile->Image.ColorMap);
          _consolePrintf("GifFile->SColorMap = %x\n",(int)GifFile->SColorMap);
          
          ColorMapObject *ColorMap = (GifFile->Image.ColorMap ? GifFile->Image.ColorMap : GifFile->SColorMap);
          
          PalettesCount=ColorMap->ColorCount;
          
          for(u32 idx=0;idx<256;idx++){
            GIFInfo.Palettes[idx]=0 | BIT15;
          }
          
          for(u32 idx=0;idx<PalettesCount;idx++){
            GifColorType* pColor = &ColorMap->Colors[idx];
            
            u32 pal=0;
            
            pal|=(GAMMA(pColor->Blue) & 0xff)<< 0;
            pal|=(GAMMA(pColor->Green) & 0xff) << 8;
            pal|=(GAMMA(pColor->Red) & 0xff) << 16;
            
            GIFInfo.Palettes[idx]=pal;
          }
          
          BitmapWidth=(Width+4+4) & ~3;
          
          _consolePrintf("malloc(%d);\n",BitmapWidth*Height);
          
          if(pBitmap!=NULL) free(pBitmap);
          pBitmap=(u8*)safemalloc(BitmapWidth*Height);
          if(pBitmap==NULL){
            _PrintGifErrorStr("out of memory.\n");
            return(false);
          }
          MemSet8DMA3(GifFile->SBackGroundColor,pBitmap,BitmapWidth*Height);
          
          _consolePrintf("GetImage.\n");
          
          if (GifFile->Image.Interlace) {
            /* Need to perform 4 passes on the images: */
            const short InterlacedOffset[] = { 0, 4, 2, 1 }; /* The way Interlaced image should. */
            const short InterlacedJumps[] = { 8, 8, 4, 2 };    /* be read - offsets and jumps... */
            for (u32 InterlaceCount=0; InterlaceCount<4; InterlaceCount++){
              MWin_ProgressShow("Decode interlace image.",Height);
              for (s32 y = InterlacedOffset[InterlaceCount]; y < Height; y += InterlacedJumps[InterlaceCount]) {
                MWin_ProgressSetPos(y);
                if (DGifGetLine(GifFile, &pBitmap[(Height-1-y)*BitmapWidth], Width) == GIF_ERROR) {
                  _PrintGifError();
                  return false;
                }
              }
              MWin_ProgressHide();
            }
            }else{
            MWin_ProgressShow("Decode image.",Height);
            for (s32 y = 0; y < Height; y++) {
              MWin_ProgressSetPos(y);
              if (DGifGetLine(GifFile, &pBitmap[(Height-1-y)*BitmapWidth], Width) == GIF_ERROR) {
                _PrintGifError();
                return false;
              }
            }
            MWin_ProgressHide();
          }
          
          _consolePrintf("GIFInfo Update.\n");
          
          GIFInfo.Width=Width;
          GIFInfo.Height=Height;
          GIFInfo.PalettesCount=PalettesCount;
          break;
        case EXTENSION_RECORD_TYPE:
          _consolePrintf("EXTENSION_RECORD_TYPE\n");
          /* Skip any extension blocks in file: */
          GifByteType *Extension;
          int ExtCode;
          if (DGifGetExtension(GifFile, &ExtCode, &Extension) == GIF_ERROR) {
            _PrintGifError();
            return false;
          }
          while (Extension != NULL) {
            if (DGifGetExtensionNext(GifFile, &Extension) == GIF_ERROR) {
              _PrintGifError();
              return false;
            }
          }
          break;
        case TERMINATE_RECORD_TYPE:
          _consolePrintf("TERMINATE_RECORD_TYPE\n");
          break;
        default:      /* Should be traps by DGifGetRecordType. */
          _consolePrintf("unknown RECORD_TYPE\n");
          break;
      }
    }
    while (RecordType != TERMINATE_RECORD_TYPE);
    
    _consolePrintf("CloseFile.\n");
    /* Close file when done */
    if (DGifCloseFile(GifFile) == GIF_ERROR) {
      _PrintGifError();
      return false;
    }
  }
  
  if(pBitmap==NULL){
    return(false);
    }else{
    return(true);
  }
}

// ------------------------------------------------------------------------------------

extern void Free(void);

bool Start(int FileHandle)
{
  if(LoadGIF(FileHandle)==false){
    _consolePrintf("GIF LoadError.\n");
    Free();
    return(false);
  }
  
  _consolePrintf("PalettesCount=%d\n",GIFInfo.PalettesCount);
  _consolePrintf("Size=(%d,%d) %dpixel\n",GIFInfo.Width,GIFInfo.Height,GIFInfo.Width*GIFInfo.Height);
  
  return(true);
}

void Free(void)
{
  if(pBitmap!=NULL){
    free(pBitmap); pBitmap=NULL;
  }
}

typedef struct {
  s32 x,y,w,h;
} TRect;

void RefreshScreen(u32 OffsetX,u32 OffsetY,u16 *pBuf,u32 BufWidth,u32 DrawWidth,u32 DrawHeight,bool HalfMode)
{
//  _consolePrintf("%d,%d 0x%x,%d %d,%d %d\n",OffsetX,OffsetY,(u32)pBuf,BufWidth,DrawWidth,DrawHeight,HalfMode);
  
  u32 BaseX=OffsetX,BaseY=OffsetY;
  
  TRect ClientRect;
  
  ClientRect.x=0;
  ClientRect.y=0;
  ClientRect.h=DrawHeight;
  ClientRect.w=DrawWidth;
  
#define Gravity(c,cg) { \
  c+=cg; \
  cg=c&7; \
  c=c>>3; \
  if((c&(~0x1f))!=0) c=(c<0) ? 0x00 : 0x1f; \
}

  u32 gr=0,gg=0,gb=0;
  
  s32 DrawH,DrawW;
  
  DrawH=GIFInfo.Height-BaseY-1;
  DrawW=GIFInfo.Width-BaseX-1;
  
  if(DrawH>ClientRect.h) DrawH=ClientRect.h;
  if(DrawW>ClientRect.w) DrawW=ClientRect.w;
  
  u32 *PaletteTable=GIFInfo.Palettes;
  
  
  if(HalfMode==false){
    for(s32 y=0;y<DrawH;y++){
      u8 *pSrcBM=&pBitmap[(GIFInfo.Height-(BaseY+y)-1)*BitmapWidth];
      u16 *pDstBM=&pBuf[ClientRect.x+((ClientRect.y+y)*BufWidth)];
      
      pSrcBM+=BaseX;
      for(s32 x=0;x<DrawW;x++){
        u32 pal;
        u32 r,g,b;
        
        pal=PaletteTable[pSrcBM[x]];
        
        b=(pal >> 0) & 0xff;
        g=(pal >> 8) & 0xff;
        r=(pal >> 16) & 0xff;
        
        Gravity(b,gb);
        Gravity(g,gg);
        Gravity(r,gr);
        
        pDstBM[x]=RGB15(r,g,b) | BIT(15);
      }
    }
    
    }else{ // half
    
    for(s32 y=0;y<DrawH/2;y++){
      u8 *pSrcBM0=&pBitmap[(GIFInfo.Height-(BaseY+(y*2)+0)-1)*BitmapWidth];
      u8 *pSrcBM1=&pBitmap[(GIFInfo.Height-(BaseY+(y*2)+1)-1)*BitmapWidth];
      u16 *pDstBM=&pBuf[ClientRect.x+((ClientRect.y+y)*BufWidth)];
      
      if((pSrcBM0!=NULL)&&(pSrcBM1!=NULL)){
        pSrcBM0+=BaseX;
        pSrcBM1+=BaseX;
        for(s32 x=0;x<DrawW/2;x++){
          u32 pal;
          u32 r=0,g=0,b=0;
          
          pal=PaletteTable[pSrcBM0[(x*2)+0]];
          b+=(pal >> 0) & 0xff; g+=(pal >> 8) & 0xff; r+=(pal >> 16) & 0xff;
          pal=PaletteTable[pSrcBM0[(x*2)+1]];
          b+=(pal >> 0) & 0xff; g+=(pal >> 8) & 0xff; r+=(pal >> 16) & 0xff;
          pal=PaletteTable[pSrcBM1[(x*2)+0]];
          b+=(pal >> 0) & 0xff; g+=(pal >> 8) & 0xff; r+=(pal >> 16) & 0xff;
          pal=PaletteTable[pSrcBM1[(x*2)+1]];
          b+=(pal >> 0) & 0xff; g+=(pal >> 8) & 0xff; r+=(pal >> 16) & 0xff;
          
          b/=4; g/=4; r/=4;
          
          Gravity(b,gb);
          Gravity(g,gg);
          Gravity(r,gr);
          
          pDstBM[x]=RGB15(r,g,b) | BIT(15);
        }
      }
    }
  }
#undef Gravity
}

s32 GetWidth(void)
{
  return(GIFInfo.Width);
}

s32 GetHeight(void)
{
  return(GIFInfo.Height);
}

int GetInfoIndexCount(void)
{
  return(0);
}

bool GetInfoStrL(int idx,char *str,int len)
{
  return(false);
}

bool GetInfoStrW(int idx,UnicodeChar *str,int len)
{
  return(false);
}

bool GetInfoStrUTF8(int idx,char *str,int len)
{
  return(false);
}

