
#include <stdio.h>

#include <NDS.h>

#include "_const.h"

#include "plugin.h"
#include "plugin_def.h"

// CallBack from plugin_dll.cpp

void cbLoadLibrary(void)
{
}

void cbFreeLibrary(void)
{
}

void cbQueryInterfaceLibrary(void)
{
}

// -----------------

#include "boolean.h"
#include "jpeg_inside.h"

static u32 DecodedMCUY;
static b8 InitialDrawFlag;

// ------------------------------------------------------------------------------------

void Jpeg_ShowFrameInfo(void)
{
  _consolePrintf("MCUCount (%d*%d)=%d\n",FrameInfo.MCUWidthCount,FrameInfo.MCUHeightCount,FrameInfo.MCUCount);
  _consolePrintf("PicSize (%d*%d)=%d\n",FrameInfo.Width,FrameInfo.Height,FrameInfo.Width*FrameInfo.Height);
  
  _consolePrintf("MaxFact H=%d F=%d\n",FrameInfo.SampleMaxHFact,FrameInfo.SampleMaxVFact);
  _consolePrintf("DRI %d\n",DelayRestartInterval);
  
  _consolePrintf("BitStreamStartPosition=%d\n",BitStreamStartPosition);
}

// ------------------------------------------------------------------------------------

bool Start(int FileHandle)
{
  DecodedMCUY=0;
  InitialDrawFlag=True;
  
  pHuffmanTree=(THuffmanTree*)safemalloc(4*sizeof(THuffmanTree));
  pQuantizeTable=(TQuantizeTable*)safemalloc(4*sizeof(TQuantizeTable));
  
  if((pHuffmanTree==NULL)||(pQuantizeTable==NULL)){
    _consolePrintf("malloc HuffmanTree or QuantizeTable out of memory.\n");
    return(False);
  }
  
  MemSet32DMA3(0,pHuffmanTree,4*sizeof(THuffmanTree));
  MemSet32DMA3(0,pQuantizeTable,4*sizeof(TQuantizeTable));
  
  rfsInit(FileHandle,0);
  rfsStartFrame();
  
  if(AnalizeJpegMarker()==False){
    _consolePrintf("JpegHeader AnalizeError.\n");
    safefree(pHuffmanTree); pHuffmanTree=NULL;
    safefree(pQuantizeTable); pQuantizeTable=NULL;
    return(false);
  }
  
  rfsClose();
  
  Jpeg_ShowFrameInfo();
  
  rbsInit(FileHandle,BitStreamStartPosition);
  
  if(Jpeg_InsideInit(FrameInfo.MCUHeightCount,FrameInfo.Width)==False){
    _consolePrintf("Jpeg LoadError.\n");
    safefree(pHuffmanTree); pHuffmanTree=NULL;
    safefree(pQuantizeTable); pQuantizeTable=NULL;
    return(false);
  }
  
  return(true);
}

void Free(void)
{
  rfsClose();
  
  if(pHuffmanTree!=NULL){
    free(pHuffmanTree); pHuffmanTree=NULL;
  }
  
  if(pQuantizeTable!=NULL){
    free(pQuantizeTable); pQuantizeTable=NULL;
  }
  
  Jpeg_InsideFree();
}

typedef struct {
  s32 x,y,w,h;
} TRect;

void RefreshScreen(u32 OffsetX,u32 OffsetY,u16 *pBuf,u32 BufWidth,u32 DrawWidth,u32 DrawHeight,bool HalfMode)
{
//  _consolePrintf("%d,%d 0x%x,%d %d,%d %d\n",OffsetX,OffsetY,(u32)pBuf,BufWidth,DrawWidth,DrawHeight,HalfMode);
  
  u32 BaseX=OffsetX,BaseY=OffsetY;
  
  TRect ClientRect;
  
  ClientRect.x=0;
  ClientRect.y=0;
  ClientRect.h=DrawHeight;
  ClientRect.w=DrawWidth;
  
  LoadIWRAMMappedCode(YUVMode);
  
  if(16<=BaseY){
    u32 ReqReadyMCUY;
    
    ReqReadyMCUY=BaseY/16;
    
    if((DecodedMCUY+1)<ReqReadyMCUY){
      s32 PrgBase=DecodedMCUY+1;
      MWin_ProgressShow("SkipImage",ReqReadyMCUY-PrgBase);
      for(u32 y=DecodedMCUY+1;y<ReqReadyMCUY;y++){
        MWin_ProgressSetPos(y-PrgBase);
        DrawImageOneLine(y,LCBM_False);
      }
      DecodedMCUY=ReqReadyMCUY-1;
      MWin_ProgressHide();
    }
  }
  
  s32 DrawH,DrawW;
  
  DrawH=FrameInfo.Height-BaseY-1;
  DrawW=FrameInfo.Width-BaseX-1;
  
  if(DrawH>ClientRect.h) DrawH=ClientRect.h;
  if(DrawW>ClientRect.w) DrawW=ClientRect.w;
  
  u32 ycnt,ylimit;
  
  ylimit=(ClientRect.h/16)+2;
  
  if(InitialDrawFlag==True) MWin_ProgressShow("Decode...",ylimit);
  
  for(ycnt=0;ycnt<ylimit;ycnt++){
    u16 *drawbm;
    u32 srcypos;
    
    srcypos=(BaseY/16)+ycnt;
    
    if(srcypos<LCBMCount){
      u32 cachednum=LCBM_FindCache(srcypos);
      
      if(cachednum!=LCBM_False){
        LCBM[cachednum].Life=LCBMLifeCur; LCBMLifeCur++;
        drawbm=LCBM[cachednum].BM;
        }else{
        u32 freenum=LCBM_FindFree();
        
        if(freenum!=LCBM_False){
          if(LCBM[freenum].BM!=NULL){
            _consolePrintf("JpegPlugin InsideError. malloc LCBM but NULL.\n");
            ShowLogHalt();
          }
          LCBM[freenum].BM=(u16*)safemalloc(LCBMWidth*LCBMHeight*2);
          if(LCBM[freenum].BM==NULL){
            freenum=LCBM_FindOldCache();
          }
          }else{
          freenum=LCBM_FindOldCache();
        }
        
        LCBM[freenum].Life=LCBMLifeCur; LCBMLifeCur++;
        LCBM[freenum].ypos=srcypos;
        if(srcypos<FrameInfo.MCUHeightCount){
          DrawImageOneLine(srcypos,freenum);
        }
        if(DecodedMCUY<srcypos) DecodedMCUY=srcypos;
        drawbm=LCBM[freenum].BM;
      }
      }else{
      drawbm=NULL;
    }
    
    if(drawbm!=NULL){
      int yofs=((ycnt*LCBMHeight)-(BaseY % 16));
      u16 *srcbm=&drawbm[BaseX+(0*LCBMWidth)];
      
      if(HalfMode==false){
        u16 *dstbm=&pBuf[ClientRect.x+((ClientRect.y+yofs+0)*BufWidth)];
        
        for(u32 y=0;y<LCBMHeight;y++){
          s32 base=yofs+y;
          
          if((0<=base)&&(base<DrawH)){
            MemCopy16DMA3(srcbm,dstbm,DrawW*2);
          }
          srcbm+=LCBMWidth;
          dstbm+=BufWidth;
        }
        
        }else{
        
        u16 *dstbm=&pBuf[ClientRect.x+((ClientRect.y+((yofs+0)/2))*BufWidth)];
        
        for(u32 y=0;y<LCBMHeight/2;y++){
          s32 base=yofs+(y*2);
          
          if((0<=base)&&(base<DrawH)){
            u16 *srcbm0=&srcbm[0*LCBMWidth];
            u16 *srcbm1=&srcbm[1*LCBMWidth];
            for(int x=0;x<DrawW/2;x++){
              u32 r=0,g=0,b=0;
              u32 rgb;
              
              rgb=srcbm0[x*2+0];
              r+=(rgb >> 0) & 0x1f; g+=(rgb >> 5) & 0x1f; b+=(rgb >> 10) & 0x1f;
              rgb=srcbm0[x*2+1];
              r+=(rgb >> 0) & 0x1f; g+=(rgb >> 5) & 0x1f; b+=(rgb >> 10) & 0x1f;
              rgb=srcbm1[x*2+0];
              r+=(rgb >> 0) & 0x1f; g+=(rgb >> 5) & 0x1f; b+=(rgb >> 10) & 0x1f;
              rgb=srcbm1[x*2+1];
              r+=(rgb >> 0) & 0x1f; g+=(rgb >> 5) & 0x1f; b+=(rgb >> 10) & 0x1f;
              
              dstbm[x]=BIT15 | RGB15(r/4,g/4,b/4);
            }
          }
          srcbm+=LCBMWidth*2;
          dstbm+=BufWidth;
        }
      }
    }
    
    if(InitialDrawFlag==True) MWin_ProgressSetPos(ycnt);
  }
  
  if(InitialDrawFlag==True){
    MWin_ProgressHide();
    InitialDrawFlag=False;
  }
}

s32 GetWidth(void)
{
  return(FrameInfo.Width);
}

s32 GetHeight(void)
{
  return(FrameInfo.Height);
}

int GetInfoIndexCount(void)
{
  return(5);
}

bool GetInfoStrL(int idx,char *str,int len)
{
  switch(idx){
    case 0: snprintf(str,len,"Pixels=%dx%d=%d",FrameInfo.Width,FrameInfo.Height,FrameInfo.Width*FrameInfo.Height); return(true); break;
    case 1: snprintf(str,len,"MCUCount=%dx%d=%d",FrameInfo.MCUWidthCount,FrameInfo.MCUHeightCount,FrameInfo.MCUCount); return(true); break;
    case 2: snprintf(str,len,"MaxFact=(H,F)=%d,%d",FrameInfo.SampleMaxHFact,FrameInfo.SampleMaxVFact); return(true); break;
    case 3: snprintf(str,len,"DRI=%d",DelayRestartInterval); return(true); break;
    case 4: snprintf(str,len,"BitStreamStartPosition=%d",BitStreamStartPosition); return(true); break;
  }
  return(false);
}

bool GetInfoStrW(int idx,UnicodeChar *str,int len)
{
  return(false);
}

bool GetInfoStrUTF8(int idx,char *str,int len)
{
  return(false);
}

