/*
  MDX player

  Made by Daisuke Nagano <breeze.nagano@nifty.ne.jp>
  Jan.16.1999


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "../plugin.h"
#include "../plugin_def.h"
#include "std.h"

#include "version.h"
#include "mdx.h"
//#include "mdxopl3.h"

/* ------------------------------------------------------------------ */
#define PATH_BUF_SIZE 1024

static PDX_DATA* _get_pdx(MDX_DATA* mdx,void *pdxbuf,int pdxbufsize);
static int self_construct(void);
static void self_destroy(void);

/* ------------------------------------------------------------------ */

/* ------------------------------------------------------------------ */

static MDX_DATA mdx;
static PDX_DATA *pdx = NULL;

bool StartMDX(void *mdxbuf,int mdxbufsize,void *pdxbuf,int pdxbufsize) {

  if (!self_construct()) {
    /* failed to create class instances */
    return(false);
  }

    MDX_DATA *pmdx=&mdx;

    /* load mdx file */

    if(mdx_open_mdx(pmdx,mdxbuf,mdxbufsize)==false) return(false);
    if ( pmdx == NULL ) return(false);

    pmdx->is_use_pcm8         = FLAG_FALSE;
    pmdx->is_use_fm           = FLAG_TRUE;
    pmdx->is_use_opl3         = FLAG_FALSE;
    pmdx->fm_wave_form        = 0;
    pmdx->master_volume       = 127;
    pmdx->fm_volume           = 127;
    pmdx->pcm_volume          = 127;
    pmdx->max_infinite_loops  = 2;
    pmdx->fade_out_speed      = 5;

    pmdx->is_output_titles    = FLAG_TRUE;
    pmdx->dump_voice=FLAG_TRUE;

//    pmdx->dsp_speed=48000; // == PCM8_MASTER_PCM_RATE

    /* voice data load */

    if ( mdx_get_voice_parameter( pmdx ) != 0 ) {
      _consolePrintf("mdx_get_voice_parameter( mdx ) == false\n");
      return(false);
    }
    
    /* load pdx data */
    pdx = _get_pdx(pmdx,pdxbuf,pdxbufsize);
    
    /* open io port */
    
    /* output Title, etc... */

    if ( pmdx->is_output_titles == FLAG_TRUE ) {
      mdx_output_titles( pmdx );
    }

    if(mdx_parse_mml_ym2151_Start(pmdx,pdx)==false) return(false);
    
    return(true);
}

int UpdateMDX(s16 *lbuf,s16 *rbuf) {

  /* MML data parse and play */
  
  if(mdx_parse_mml_ym2151_Update(lbuf,rbuf)==false){
    return(0);
    }else{
    return(pcm8_GetSampleBufferSize());
  }
}

void FreeMDX(void) {

  mdx_parse_mml_ym2151_Free();

    MDX_DATA *pmdx=&mdx;
    
  /* one playing finished */
    
    /* close io port */
    
    mdx_close_pdx( pdx );
    mdx_close_mdx( pmdx );

  self_destroy();
  
}

/* pdx loading */

extern int settingTotalTrack;
int settingTotalTrack=0;

static PDX_DATA*
_get_pdx(MDX_DATA* mdx,void *pdxbuf,int pdxbufsize)
{
  PDX_DATA* pdx = NULL;

  mdx->haspdx = FLAG_FALSE;
  mdx->tracks = settingTotalTrack;
  mdx->pdx_enable = FLAG_FALSE;
  return NULL;
}

/* class initializations */
extern void* _mdxmml_ym2151_initialize(void);
extern void* _mdx2151_initialize(void);
extern void* _pcm8_initialize(void);
extern void* _ym2151_c_initialize(void);
extern void  _mdxmml_ym2151_finalize(void* self);
extern void  _mdx2151_finalize(void* in_self);
extern void  _pcm8_finalize(void* in_self);
extern void  _ym2151_c_finalize(void* in_self);

static void* self_mdx2151 = NULL;
static void* self_mdxmml_ym2151 = NULL;
static void* self_pcm8 = NULL;
static void* self_ym2151_c = NULL;

void*
_get_mdx2151(void)
{
  return self_mdx2151;
}

void*
_get_mdxmml_ym2151(void)
{
  return self_mdxmml_ym2151;
}

void*
_get_pcm8(void)
{
  return self_pcm8;
}

void*
_get_ym2151_c(void)
{
  return self_ym2151_c;
}

static int
self_construct(void)
{
  self_mdx2151 = _mdx2151_initialize();
  if (!self_mdx2151) {
    goto error_end;
  }
  self_mdxmml_ym2151 = _mdxmml_ym2151_initialize();
  if (!self_mdxmml_ym2151) {
    goto error_end;
  }
  self_pcm8 = _pcm8_initialize();
  if (!self_pcm8) {
    goto error_end;
  }
#if 0
  self_ym2151_c = _ym2151_c_initialize();
  if (!self_ym2151_c) {
    goto error_end;
  }
#endif

  return FLAG_TRUE;

error_end:
#if 0
  if (self_ym2151_c) {
    _ym2151_c_finalize(self_ym2151_c);
    self_ym2151_c = NULL;
  }
#endif
  if (self_pcm8) {
    _pcm8_finalize(self_pcm8);
    self_pcm8 = NULL;
  }
  if (self_mdxmml_ym2151) {
    _mdxmml_ym2151_finalize(self_mdxmml_ym2151);
    self_mdxmml_ym2151 = NULL;
  }
  if (self_mdx2151) {
    _mdx2151_finalize(self_mdx2151);
    self_mdx2151 = NULL;
  }
  return FLAG_FALSE;
}

static void
self_destroy(void)
{
#if 0
  if (self_ym2151_c) {
    _ym2151_c_finalize(self_ym2151_c);
    self_ym2151_c = NULL;
  }
#endif
  if (self_pcm8) {
    _pcm8_finalize(self_pcm8);
    self_pcm8 = NULL;
  }
  if (self_mdxmml_ym2151) {
    _mdxmml_ym2151_finalize(self_mdxmml_ym2151);
    self_mdxmml_ym2151 = NULL;
  }
  if (self_mdx2151) {
    _mdx2151_finalize(self_mdx2151);
    self_mdx2151 = NULL;
  }
}

/* command line handling */

static void display_version( void ) {

  _consolePrintf(VERSION_ID " [ " VERSION_TEXT1 " / " VERSION_TEXT2 " ]\n");
  _consolePrintf("Made by NAGANO Daisuke <breeze.nagano@nifty.com> 1999-2005\n\n");
  _consolePrintf("original:\n");
  _consolePrintf("   X68k MXDRV music driver (c)1988-92 milk.,K.MAEKAWA, Missy.M, Yatsube\n");
  _consolePrintf("ym2151.[ch]:\n");
  _consolePrintf("   (c) 1997-2002 Jarek Burczynski (s0246@poczta.onet.pl, bujar@mame.net)\n");
  _consolePrintf("   Some of the optimizing ideas by Tatsuyuki Satoh\n");

  _consolePrintf("\n");
  _consolePrintf("This software is under the GPL.\n");
  _consolePrintf("This is free software; see the source for copying conditions.\n");
  _consolePrintf("There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A\n");
  _consolePrintf("PARTICULAR PURPOSE.\n\n");

}
