/*
  MDXplayer :  MML parser

  Made by Daisuke Nagano <breeze.nagano@nifty.ne.jp>
  Jan.15.1999

  reference : mdxform.doc  ( Takeshi Kouno )
            : MXDRVWIN.pas ( monlight@tkb.att.ne.jp )
 */

/* ------------------------------------------------------------------ */

#include "../plugin.h"
#include "../plugin_def.h"
#include "std.h"

#include "version.h"
#include "mdx.h"
//#include "mdxopl3.h"
#include "mdx2151.h"
#include "pcm8.h"

# define       timercmp(tvp, uvp, cmp)\
               ((tvp)->tv_sec cmp (uvp)->tv_sec ||\
               (tvp)->tv_sec == (uvp)->tv_sec &&\
               (tvp)->tv_usec cmp (uvp)->tv_usec)

/* ------------------------------------------------------------------ */
/* export symbols */

int mdx_parse_mml_ym2151( MDX_DATA *, PDX_DATA * );
int mdx_init_track_work_area_ym2151( void );

/* ------------------------------------------------------------------ */
/* local valiables */

static const int vol_conv[] = {
   85,  87,  90,  93,  95,  98, 101, 103,
  106, 109, 111, 114, 117, 119, 122, 125
};

/* ------------------------------------------------------------------ */
/* local instances */

typedef struct _mdxmml_ym2151_instances {
  MDX_DATA *mdx;
  PDX_DATA *pdx;

  int pcm8_disable;
  int fade_out;
  int is_sigint_caught;

  int all_track_finished;
  int fade_out_wait;
  int master_volume;
} mdxmml_ym2151_instances;

static ALIGNED_VAR_IN_DTCM TRACK_WORK mdx_track[MDX_MAX_TRACK_NUMBER];

/* ------------------------------------------------------------------ */
/* local functions */

static int  set_new_event( int );
static void set_tempo( int );
static void note_off( int );
static void note_on( int, int );
static void set_volume( int, int );
static void inc_volume( int );
static void dec_volume( int );
static void set_phase( int, int );
static void set_keyoff( int );
static void set_voice( int, int );
static void set_quantize( int, int );
static void set_detune( int, int, int );
static void set_portament( int, int, int );
static void set_freq( int, int );
static void set_opm_reg( int, int, int );
static void set_fade_out( int );
static void send_sync( int );
static void recv_sync( int );

static void set_plfo_onoff( int, int );
static void set_plfo( int, int, int, int, int, int );
static void set_alfo_onoff( int, int );
static void set_alfo( int, int, int, int, int, int );
static void set_hlfo_onoff( int, int );
static void set_hlfo( int, int, int, int, int, int );
static void set_lfo_delay( int, int );

static void do_quantize( int, int );

/* externs */
extern void pcm8_clear_buffer_flush_flag(void);
extern int pcm8_buffer_flush_flag(void);

/* ------------------------------------------------------------------ */
/* class interface */

extern void* _get_mdxmml_ym2151(void);

static mdxmml_ym2151_instances*
__get_instances(void)
{
  return (mdxmml_ym2151_instances*)_get_mdxmml_ym2151();
}

#define __GETSELF  mdxmml_ym2151_instances* self = __get_instances();

void*
_mdxmml_ym2151_initialize(void)
{
  mdxmml_ym2151_instances* self = NULL;

  self = (mdxmml_ym2151_instances *)malloc(sizeof(mdxmml_ym2151_instances));
  if (!self) {
    return NULL;
  }
  memset(self, 0, sizeof(mdxmml_ym2151_instances));

  self->mdx = NULL;
  self->pdx = NULL;
  self->pcm8_disable = FLAG_TRUE;
  self->fade_out = 0;
  self->is_sigint_caught = FLAG_FALSE;

  return self;
}

void
_mdxmml_ym2151_finalize(void* in_self)
{
  mdxmml_ym2151_instances* self = (mdxmml_ym2151_instances *)in_self;
  if (self) {
    free(self);
  }
}

/* ------------------------------------------------------------------ */
/* implementations */

static int fade_out_wait;
static int master_volume;
static int cursample;

bool mdx_parse_mml_ym2151_Start( MDX_DATA *orig_mdx, PDX_DATA *orig_pdx )
{
  __GETSELF;

  self->mdx = orig_mdx;
  self->pdx = orig_pdx;

  mdx_init_track_work_area_ym2151();

  self->pcm8_disable=FLAG_TRUE;
  if ( pcm8_open(self->mdx)==0 ) {
    self->pcm8_disable=FLAG_FALSE;
  }

  if (!ym2151_reg_init( self->mdx )) {
    /* failed to initialize opm! */
    return(false);
  }

  /* start parsing */

  fade_out_wait=0;
  master_volume=127;
  cursample=0;

/*
  gettimeofday(&st, NULL);
  srand((int)st.tv_usec%65536);
  gettimeofday(&et, NULL);
*/

  return(true);
}

#include "masterdef.h"

bool mdx_parse_mml_ym2151_Update(s16 *lbuf,s16 *rbuf)
{
  __GETSELF;
  
  cursample+=MASTER_PCM_SAMPLES_PER_FRAME;
  
  while(0<=cursample){
    cursample-=MASTER_PCM_RATE*(256-self->mdx->tempo)/4000;
/*
    float ms=((float)1024*(256-self->mdx->tempo))/(float)4000;
    cursample-=MASTER_PCM_RATE*((float)ms/1000);
*/
    
    if ( self->fade_out > 0 ) {
      if ( fade_out_wait==0 ) { fade_out_wait = self->fade_out; }
      fade_out_wait--;
      if ( fade_out_wait==0 ) { master_volume--; }
      if ( master_volume==0 ) { return(false); }
    }
    ym2151_set_master_volume( master_volume * self->mdx->fm_volume / 127 );
    pcm8_set_master_volume( master_volume * self->mdx->pcm_volume / 127 );

    int infinite_loops = 32767; /* large enough */
    int all_track_finished=FLAG_TRUE;
    
    int i;
    for ( i=0 ; i<self->mdx->tracks ; i++ ) {

      TRACK_WORK *pmdx_track=&mdx_track[i];
      
      if ( pmdx_track->waiting_sync == FLAG_TRUE )
	{ continue; }

      int count = pmdx_track->counter;
      
      if ( count < 0 ) { continue; } /* this track has finished */
      all_track_finished=FLAG_FALSE;

      pmdx_track->gate--;
      if ( pmdx_track->gate == 0 ) { note_off( i ); }

      if ( i<8 ) {
	ym2151_set_freq_volume( i ); /* do portament, lfo, detune */
      }

      count--;
      while ( count == 0 ) {
	count=set_new_event( i );
      }

      pmdx_track->counter = count;
      if ( infinite_loops > pmdx_track->infinite_loop_times ) {
	infinite_loops = pmdx_track->infinite_loop_times;
      }
    }

    if ( self->mdx->max_infinite_loops > 0 ) {
      if ( infinite_loops >= self->mdx->max_infinite_loops ) {
	self->fade_out = self->mdx->fade_out_speed;
      }
    }

    /* timer count */

    self->mdx->total_count++;
//    self->mdx->elapsed_time += 1000*1024*(256 - self->mdx->tempo)/4000;
    
    if(all_track_finished==FLAG_TRUE) return(false);
  }
  
  if((lbuf==NULL)||(rbuf==NULL)) return(true);
  
  do_pcm8(lbuf,rbuf);
  
  return(true);
}

void mdx_parse_mml_ym2151_Free(void)
{
  ym2151_all_note_off();
  pcm8_close();
  ym2151_shutdown();
}

/* ------------------------------------------------------------------ */


/* ------------------------------------------------------------------ */

#define __GETMDX \
MDX_DATA* mdx; \
__GETSELF; \
mdx = self->mdx;

int
mdx_init_track_work_area_ym2151( void )
{
  int i;

  __GETMDX;

  self->fade_out = 0;

  mdx->tempo        = 200;
  mdx->total_count  = 0;
  mdx->elapsed_time = 0;

  mdx->fm_noise_vol = 0;
  mdx->fm_noise_freq = 0;

  for ( i=0 ; i<mdx->tracks ; i++ ) {
    TRACK_WORK *pmdx_track=&mdx_track[i];
    
    pmdx_track->counter = 1;
    pmdx_track->gate = 1;
    pmdx_track->current_mml_ptr = mdx->mml_data_offset[i];

    pmdx_track->voice         = 0;
    pmdx_track->volume        = 64;
    pmdx_track->volume_normal = 8;
    pmdx_track->note          = 0;
    pmdx_track->phase         = MDX_PAN_C;
    pmdx_track->quantize1     = 8;
    pmdx_track->quantize2     = 0;
    pmdx_track->detune        = 0;
    if ( i<8 )
      ym2151_set_detune(i,0);
    pmdx_track->portament     = 0;
    if ( i<8 )
      ym2151_set_portament(i,0);

    pmdx_track->loop_depth = 0;
    pmdx_track->infinite_loop_times = 0;

    pmdx_track->p_lfo_flag = FLAG_FALSE;
    pmdx_track->a_lfo_flag = FLAG_FALSE;
    pmdx_track->h_lfo_flag = FLAG_FALSE;

    pmdx_track->p_lfo_form = 0;
    pmdx_track->p_lfo_clock = 255;
    pmdx_track->p_lfo_depth = 0;

    pmdx_track->a_lfo_form = 0;
    pmdx_track->a_lfo_clock = 255;
    pmdx_track->a_lfo_depth = 0;

    pmdx_track->waiting_sync = FLAG_FALSE;

    pmdx_track->keyoff_disable     = FLAG_FALSE;
    pmdx_track->last_volume_normal = FLAG_FALSE;
  }

  return 0;
}

extern int settingTranspose;
int settingTranspose=0;

static int
set_new_event( int t )
{
  int ptr;
  unsigned char *data;
  int count;
  int follower;
  __GETMDX;
  
  TRACK_WORK *pmdx_track=&mdx_track[t];

  data = mdx->data;
  ptr = pmdx_track->current_mml_ptr;
  count = 0;
  follower = 0;

#if 0
  if ( ptr+1 <= mdx->length && t>7 ) {
    fprintf(stderr,"%2d %2x %2x\n",t,data[ptr],data[ptr+1]);fflush(stderr);
  }
#endif

  if ( data[ptr] <= MDX_MAX_REST ) {  /* rest */
    note_off(t);
    count = data[ptr]+1;
    pmdx_track->gate = count+1;
    follower=0;

  } else if ( data[ptr] <= MDX_MAX_NOTE ) { /* note */
    if(settingTranspose==0){
      note_on( t, data[ptr]);
      }else{
      int note=data[ptr];
      note-=12;
      if(note<=0) note=1;
      note_on( t, note);
    }
    count = data[ptr+1]+1;
    do_quantize( t, count );
    follower = 1;

  } else {
    switch ( data[ptr] ) {

    case MDX_SET_TEMPO:
      set_tempo( data[ptr+1] );
      follower = 1;
      break;

    case MDX_SET_OPM_REG:
      set_opm_reg( t, data[ptr+1], data[ptr+2] );
      follower = 2;
      break;

    case MDX_SET_VOICE:
      set_voice( t, data[ptr+1] );
      follower = 1;
      break;

    case MDX_SET_PHASE:
      set_phase( t, data[ptr+1] );
      follower = 1;
      break;

    case MDX_SET_VOLUME:
      set_volume( t, data[ptr+1] );
      follower = 1;
      break;

    case MDX_VOLUME_DEC:
      dec_volume( t );
      follower = 0;
      break;

    case MDX_VOLUME_INC:
      inc_volume( t );
      follower = 0;
      break;

    case MDX_SET_QUANTIZE:
      set_quantize( t, data[ptr+1] );
      follower = 1;
      break;

    case MDX_SET_KEYOFF:
      set_keyoff( t );
      follower = 0;
      break;

    case MDX_REPEAT_START:
      pmdx_track->loop_counter[pmdx_track->loop_depth] = data[ptr+1];
      if ( pmdx_track->loop_depth < MDX_MAX_LOOP_DEPTH )
	pmdx_track->loop_depth++;
      follower = 2;
      break;

    case MDX_REPEAT_END:
      if (--pmdx_track->loop_counter[pmdx_track->loop_depth-1] == 0 ) {
	if ( --pmdx_track->loop_depth < 0 ) pmdx_track->loop_depth=0;
      } else {
	if ( data[ptr+1] >= 0x80 ) {
	  ptr = ptr+2 - (0x10000-(data[ptr+1]*256 + data[ptr+2])) - 2;
	} else {
	  ptr = ptr+2 + data[ptr+1]*256 + data[ptr+2] - 2;
	}
      }
      follower = 2;
      break;

    case MDX_REPEAT_BREAK:
      if ( pmdx_track->loop_counter[pmdx_track->loop_depth-1] == 1 ) {
	if ( --pmdx_track->loop_depth < 0 ) pmdx_track->loop_depth=0;
	ptr = ptr+2 + data[ptr+1]*256 + data[ptr+2] -2 +2;
      }
      follower = 2;
      break;

    case MDX_SET_DETUNE:
      set_detune( t, data[ptr+1], data[ptr+2] );
      follower = 2;
      break;

    case MDX_SET_PORTAMENT:
      set_portament( t, data[ptr+1], data[ptr+2] );
      follower = 2;
      break;

    case MDX_DATA_END:
      if ( data[ptr+1] == 0x00 ) {
	count = -1;
	note_off(t);
	follower = 1;
      } else {
	ptr = ptr+2 - (0x10000-(data[ptr+1]*256 + data[ptr+2])) - 2;
	pmdx_track->infinite_loop_times++;
	follower = 2;
      }
      break;
 
    case MDX_KEY_ON_DELAY:
      follower = 1;
      break;

    case MDX_SEND_SYNC:
      send_sync( data[ptr+1] );
      follower = 1;
      break;

    case MDX_RECV_SYNC:
      recv_sync( t );
      follower = 0;
      count = 1;
      break;

    case MDX_SET_FREQ:
      set_freq( t, data[ptr+1] );
      follower = 1;
      break;

    case MDX_SET_PLFO:
      if ( data[ptr+1] == 0x80 || data[ptr+1] == 0x81 ) {
	set_plfo_onoff( t, data[ptr+1]-0x80 );
	follower = 1;
      } else {
	set_plfo( t,
		  data[ptr+1], data[ptr+2], data[ptr+3],
		  data[ptr+4], data[ptr+5] );
	follower = 5;
      }
      break;

    case MDX_SET_ALFO:
      if ( data[ptr+1] == 0x80 || data[ptr+1] == 0x81 ) {
	set_alfo_onoff( t, data[ptr+1]-0x80 );
	follower = 1;
      } else {
	set_alfo( t,
		  data[ptr+1], data[ptr+2], data[ptr+3],
		  data[ptr+4], data[ptr+5] );
	follower = 5;
      }
      break;

    case MDX_SET_OPMLFO:
      if ( data[ptr+1] == 0x80 || data[ptr+1] == 0x81 ) {
	set_hlfo_onoff( t, data[ptr+1]-0x80 );
	follower = 1;
      } else {
	set_hlfo( t,
		  data[ptr+1], data[ptr+2], data[ptr+3],
		  data[ptr+4], data[ptr+5] );
	follower = 5;
      }
      break;

    case MDX_SET_LFO_DELAY:
      set_lfo_delay( t, data[ptr+1] );
      follower = 1;
      break;

    case MDX_SET_PCM8_MODE:
      follower = 0;
      break;

    case MDX_FADE_OUT:
      if ( data[ptr+1]==0x00 ) {
	follower = 1;
	set_fade_out( 5 );
      } else {
	follower = 2;
	set_fade_out( data[ptr+2] );
      }
      break;

    default:
      count = -1;
      break;
    }
  }

  ptr += 1+follower;
  pmdx_track->current_mml_ptr = ptr;

  return count;
}


static void
set_tempo( int val )
{
  __GETMDX;

  if (val<2) return; /* workaround!!! */
  mdx->tempo = val;
  return;
}

static void
note_off( int track )
{
  __GETMDX;

  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  if ( pmdx_track->keyoff_disable == FLAG_FALSE ) {
    pmdx_track->note=-1;
    if ( track<8 )
      ym2151_note_off(track);
    else
      pcm8_note_off(track-8);
  }
  pmdx_track->keyoff_disable = FLAG_FALSE;

  return;
}

static void
note_on( int track, int note )
{
  int last_note;
  PDX_DATA* pdx;
  __GETMDX;

  pdx = self->pdx;

  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  last_note = pmdx_track->note;
  note -= MDX_MIN_NOTE;
  pmdx_track->note = note;

  if ( pmdx_track->phase != MDX_PAN_N ) {
    if ( track<8 ){
      ym2151_note_on(track, pmdx_track->note);
    }
    else {
      if ( note<MDX_MAX_PDX_TONE_NUMBER && self->pcm8_disable == FLAG_FALSE && pdx != NULL ) {
	int n = note+pmdx_track->voice*MDX_MAX_PDX_TONE_NUMBER;
	pcm8_note_on( track-8,
		      pdx->tone[n].data,
		      pdx->tone[n].size, 
		      pdx->tone[n].orig_data,
		      pdx->tone[n].orig_size ); 
      }
    }
  }

  return;
}

static void
do_quantize( int t, int count )
{
  int gate;

  TRACK_WORK *pmdx_track=&mdx_track[t];
  
  __GETMDX;

  if ( pmdx_track->keyoff_disable == FLAG_TRUE ) {
    gate = count+2;
    pmdx_track->keyoff_disable = FLAG_FALSE;

  } else {
    if ( pmdx_track->quantize2>0 ) {
      gate = count - pmdx_track->quantize2;
      if ( gate <= 0 ) gate=1;

    } else {
      gate = count * pmdx_track->quantize1/8;
    }
  }

  /* FIXME : we must tune up the note-off job */
  if ( gate+1 < count )
    gate++;

  pmdx_track->gate = gate;

  return;
}

static void
set_phase( int track, int val )
{
  __GETMDX;

  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  pmdx_track->phase = val;

  if ( track<8 )
    ym2151_set_pan(track, val);
  else
    pcm8_set_pan(val);

  if ( val == MDX_PAN_N ) {
    if ( track<8 )
      ym2151_note_off(track);
    else
      pcm8_note_off(track-8);
  }

  return;
}

static void
set_volume( int track, int val )
{
  int v;
  int m;
  __GETMDX;

  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  if ( val < 0x10 ) {
    pmdx_track->volume_normal = val;
    v = vol_conv[val];
    m = FLAG_TRUE;
  }
  else if ( val >= 0x80 ) {
    v = 255-val;
    m = FLAG_FALSE;
  }
  else {
    v=0;
    m=FLAG_FALSE;
  }
  pmdx_track->volume = v;
  pmdx_track->last_volume_normal = m;

  if ( track<8 )
    ym2151_set_volume( track, v );
  else
    pcm8_set_volume( track-8, v );

  return;
}

static void
inc_volume( int track )
{
  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  __GETMDX;

  if ( pmdx_track->last_volume_normal == FLAG_TRUE ) {
    if ( ++pmdx_track->volume_normal > 0x0f ) {
      pmdx_track->volume_normal = 0x0f;
    }
    pmdx_track->volume = vol_conv[pmdx_track->volume_normal];
  }
  else {
    if ( ++pmdx_track->volume > 0x7f ) {
      pmdx_track->volume = 0x7f;
    }
  }

  if ( track<8 )
    ym2151_set_volume( track, pmdx_track->volume );
  else
    pcm8_set_volume( track-8, pmdx_track->volume );

  return;
}

static void
dec_volume( int track )
{
  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  __GETMDX;

  if ( pmdx_track->last_volume_normal == FLAG_TRUE) {
    if ( --pmdx_track->volume_normal < 0 ) {
      pmdx_track->volume_normal = 0;
    }
    pmdx_track->volume = vol_conv[pmdx_track->volume_normal];
  }
  else {
    if ( --pmdx_track->volume < 0 ) {
      pmdx_track->volume = 0;
    }
  }

  if ( track<8 )
    ym2151_set_volume( track, pmdx_track->volume );
  else
    pcm8_set_volume( track-8, pmdx_track->volume );

  return;
}

static void
set_keyoff( int track )
{
  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  __GETMDX;

  pmdx_track->keyoff_disable = FLAG_TRUE;

  return;
}

static void
set_voice( int track, int val )
{
  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  __GETMDX;

  if ( track<8 && val >= MDX_MAX_VOICE_NUMBER ) return;
  if ( track>7 && val >= MDX_MAX_PDX_BANK_NUMBER ) return;
  pmdx_track->voice = val;
  if ( track<8 ) {
    ym2151_set_voice( track, &mdx->voice[val] );
  }

  return;
}

static void
set_quantize( int track, int val )
{
  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  __GETMDX;

  if ( val < 9 ) {
    pmdx_track->quantize1 = val;
    pmdx_track->quantize2 = 0;
  }
  else {
    pmdx_track->quantize1 = 8;
    pmdx_track->quantize2 = 0x100-val;
  }

  return;
}

static void
set_detune( int track, int v1, int v2 )
{
  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  int v;
  __GETMDX;

  if ( track<8 ) {
    v = v1*256 + v2;
    if ( v1 >= 0x80 ) v = v-0x10000;
    pmdx_track->detune = v;
    ym2151_set_detune( track, v );
  }

  return;
}

static void
set_portament( int track, int v1, int v2 )
{
  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  int v;
  __GETMDX;

  if ( track<8 ) {
    v = v1*256 + v2;
    if ( v1 >= 0x80 ) v = v - 0x10000;
    pmdx_track->portament = v;
    ym2151_set_portament( track, v );
  }

  return;
}

static void
set_freq( int track, int val )
{
  __GETMDX;

  if ( track >= 8 ) {
    pcm8_set_pcm_freq( track-8, val );
  }
  else {
    ym2151_set_noise_freq( val );
  }
  return;
}

static void
set_fade_out( int speed )
{
  __GETSELF;

  self->fade_out = speed+1;
  return;
}

static void
set_opm_reg( int track, int v1, int v2 )
{
  __GETMDX;

  if ( track<8 ) {
    ym2151_set_reg( v1, v2 );
  }
  if ( v1 == 0x12 ) {
    mdx->tempo = v2;
  }

  return;
}

static void
send_sync( int track )
{
  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  __GETMDX;

  pmdx_track->waiting_sync = FLAG_FALSE;
  return;
}

static void
recv_sync( int track )
{
  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  __GETMDX;

  pmdx_track->waiting_sync = FLAG_TRUE;
  return;
}

static void
set_plfo_onoff( int track, int onoff )
{
  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  __GETMDX;

  if ( onoff == 0 ) {
    pmdx_track->p_lfo_flag = FLAG_FALSE;
  }
  else {
    pmdx_track->p_lfo_flag = FLAG_TRUE;
  }

  if ( track < 8) {
    ym2151_set_plfo( track,
		     pmdx_track->p_lfo_flag,
		     pmdx_track->p_lfo_form,
		     pmdx_track->p_lfo_clock,
		     pmdx_track->p_lfo_depth );
  }

  return;
}

static void
set_plfo( int track, int v1, int v2, int v3, int v4, int v5 )
{
  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  int t,d;
  __GETMDX;

  t = v2*256+v3;
  d = v4*256+v5;
  if ( d>=0x8000 ) {
    d = d-0x10000;
  }
  if ( v1 > 4  ) d*=256;
  /*fprintf(stderr,"%d %d %d %d\n", track, v1, t, d);fflush(stderr);*/
  pmdx_track->p_lfo_form  = v1;
  pmdx_track->p_lfo_clock = t;
  pmdx_track->p_lfo_depth = d;
  pmdx_track->p_lfo_flag  = FLAG_TRUE;

  if ( track < 8 ) {
    ym2151_set_plfo( track,
		     pmdx_track->p_lfo_flag,
		     pmdx_track->p_lfo_form,
		     pmdx_track->p_lfo_clock,
		     pmdx_track->p_lfo_depth );
  }

  return;
}

static void
set_alfo_onoff( int track, int onoff )
{
  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  __GETMDX;

  if ( onoff == 0 ) {
    pmdx_track->a_lfo_flag = FLAG_FALSE;
  }
  else {
    pmdx_track->a_lfo_flag = FLAG_TRUE;
  }

  if ( track < 8 ) {
    ym2151_set_alfo( track,
		     pmdx_track->a_lfo_flag,
		     pmdx_track->a_lfo_form,
		     pmdx_track->a_lfo_clock,
		     pmdx_track->a_lfo_depth );
  }
  return;
}

static void
set_alfo( int track, int v1, int v2, int v3, int v4, int v5 )
{
  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  int t,d;
  __GETMDX;

  t = v2*256+v3;
  d = v4*256+v5;
  if ( v4>=0x80 ) {
    d = d-0x10000;
  }
  pmdx_track->a_lfo_form  = v1;
  pmdx_track->a_lfo_clock = t;
  pmdx_track->a_lfo_depth = d;
  pmdx_track->a_lfo_flag  = FLAG_TRUE;

  if ( track < 8 ) {
    ym2151_set_alfo( track,
		     pmdx_track->a_lfo_flag,
		     pmdx_track->a_lfo_form,
		     pmdx_track->a_lfo_clock,
		     pmdx_track->a_lfo_depth );
  }

  return;
}

static void
set_hlfo_onoff( int track, int onoff )
{
  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  __GETMDX;

  if ( onoff == 0 ) {
    pmdx_track->h_lfo_flag = FLAG_FALSE;
  }
  else {
    pmdx_track->h_lfo_flag = FLAG_TRUE;
  }

  if ( track < 8 ) {
    ym2151_set_hlfo_onoff( track, pmdx_track->h_lfo_flag );
  }

  return;
}

static void
set_hlfo( int track, int v1, int v2, int v3, int v4, int v5 )
{
  int wa, sp, pd, ad, ps, as, sy;
  __GETMDX;

  wa = v1&0x03;
  sp = v2;
  pd = v3 & 0x7f;
  ad = v4;
  ps = (v5/16) & 0x07;
  as = v5 & 0x03;
  sy = (v1/64) & 0x01;

  /*fprintf(stderr,"%d %d %d %d\n", track, v1, v2, v3);fflush(stderr);*/
  if ( track < 8 ) {
    ym2151_set_hlfo( track, v1, v2, v3, v4, v5 );
  }

  return;
}

static void
set_lfo_delay( int track, int delay )
{
  TRACK_WORK *pmdx_track=&mdx_track[track];
  
  __GETMDX;

  pmdx_track->lfo_delay = delay;

  if ( track < 8 ) {
    ym2151_set_lfo_delay( track, delay );
  }
  return;
}
