/*
  MDXplayer : PCM8 emulater :-)

  Made by Daisuke Nagano <breeze.nagano@nifty.ne.jp>
  Jan.16.1999
 */

/* ------------------------------------------------------------------ */

#include "../plugin.h"
#include "../plugin_def.h"
#include "std.h"

#include "version.h"
#include "mdx.h"
#include "pcm8.h"
#include "ym2151.h"

extern void* ym2151_instance(void);

/* ------------------------------------------------------------------ */
/* local instances */

typedef struct _pcm8_instances pcm8_instances;
struct _pcm8_instances {
  MDX_DATA *emdx;
  PCM8_WORK work[PCM8_MAX_NOTE];

  int pcm8_opened;
  int pcm8_interrupt_active;
  int dev;
  
  int master_volume;
  int master_pan;

//  int is_encoding_16bit;
//  int is_encoding_stereo;
  int dsp_speed;

  unsigned char *sample_buffer;
  int *sample_buffer2;
  int sample_buffer_size;
  int dest_buffer_size;

  SAMP *ym2151_voice[2];

  unsigned char *pcm_buffer;
  int pcm_buffer_ptr;
  int pcm_buffer_size;

  int is_pcm_buffer_flushed;
};


/* ------------------------------------------------------------------ */
/* local defines */

#define PCM8_MAX_FREQ 5

#include "masterdef.h"

static const int adpcm_freq_list[] = {
  3900, 5200, 7800, 10400, 15600
};

/* ------------------------------------------------------------------ */
/* class interface */

extern void* _get_pcm8(void);

static pcm8_instances*
__get_instances(void)
{
  return (pcm8_instances*)_get_pcm8();
}

#define __GETSELF  pcm8_instances* self = __get_instances();

void*
_pcm8_initialize(void)
{
  pcm8_instances* self = NULL;

  self = (pcm8_instances *)malloc(sizeof(pcm8_instances));
  if (!self) {
    return NULL;
  }
  memset(self, 0, sizeof(pcm8_instances));

  self->emdx                  = NULL;
  self->pcm8_opened           = FLAG_FALSE;
  self->pcm8_interrupt_active = 0;
  self->dev                   = -1;
  
  self->master_volume         = 0;
  self->master_pan            = MDX_PAN_N;

  self->dsp_speed             = 0;

  self->sample_buffer         = NULL;
  self->sample_buffer2        = NULL;
  self->sample_buffer_size    = 0;

  self->ym2151_voice[0]       = NULL;
  self->ym2151_voice[1]       = NULL;

  self->pcm_buffer            = NULL;
  self->pcm_buffer_ptr        = 0;
  self->pcm_buffer_size       = 0;

  self->is_pcm_buffer_flushed = FLAG_FALSE;

  return self;
}

void
_pcm8_finalize(void* in_self)
{
  pcm8_instances* self = (pcm8_instances *)in_self;
  if (self) {
    free(self);
  }
}

/* ------------------------------------------------------------------ */
/* implementations */

int pcm8_open( MDX_DATA *mdx )
{
  int i;
  __GETSELF;

  self->emdx = mdx;

  self->dsp_speed = MASTER_PCM_RATE;
  
  if ( self->pcm_buffer != NULL ) free(self->pcm_buffer);
  self->pcm_buffer = NULL;

  self->sample_buffer_size = MASTER_PCM_SAMPLES_PER_FRAME;
//    self->dsp_speed * PCM8_SYSTEM_RATE / 1000;

  self->dest_buffer_size = self->sample_buffer_size*4;

  if ( self->sample_buffer == NULL ) {
    self->sample_buffer =
      (unsigned char *)malloc(sizeof(char)*self->sample_buffer_size*2*2);
  }
  if ( self->sample_buffer2 == NULL ) {
    self->sample_buffer2 =
      (int *)malloc(sizeof(int)*self->sample_buffer_size);
  }

  if ( self->sample_buffer == NULL || self->sample_buffer2 == NULL ) {
    if ( self->sample_buffer != NULL ) free( self->sample_buffer );
    if ( self->sample_buffer2 != NULL ) free( self->sample_buffer2 );
    return 1;
  }

  for ( i=0 ; i<2 ; i++ ) {
    if ( self->ym2151_voice[i] == NULL ) {
      self->ym2151_voice[i] = (SAMP *)malloc(self->sample_buffer_size*2);
    }
  }

  if ( self->ym2151_voice[0] == NULL || self->ym2151_voice[1] == NULL ) {
    if ( self->ym2151_voice[0] != NULL ) free( self->ym2151_voice[0] );
    if ( self->ym2151_voice[1] != NULL ) free( self->ym2151_voice[1] );
    return 1;
  }

  self->pcm8_opened = FLAG_TRUE;
  self->emdx->dsp_speed = self->dsp_speed;

  pcm8_init();

  return 0;
}

int pcm8_close( void )
{
  int i;
  int free_all = FLAG_FALSE;
  __GETSELF;

  self->pcm8_opened = FLAG_FALSE;

  pcm8_stop();

  if ( free_all == FLAG_TRUE ) {
    for ( i=1 ; i<=0 ; i-- ) {
      if ( self->ym2151_voice[i]!=NULL ) {
	free( self->ym2151_voice[i] );
	self->ym2151_voice[i]=NULL;
      }
    }

    if ( self->sample_buffer2 != (int *)NULL ) {
      free( self->sample_buffer2 );
      self->sample_buffer2 = NULL;
    }

    if ( self->sample_buffer!=NULL ) {
      free( self->sample_buffer );
      self->sample_buffer = NULL;
    }
  }

  return 0;
}

void pcm8_init( void )
{
  int i;
  __GETSELF;

  for ( i=0; i<PCM8_MAX_NOTE ; i++ ) {
    self->work[i].ptr     = NULL;
    self->work[i].top_ptr = NULL;
    self->work[i].end_ptr = NULL;
    self->work[i].volume  = PCM8_MAX_VOLUME;
    self->work[i].freq    = adpcm_freq_list[PCM8_MAX_FREQ-1];
    self->work[i].adpcm   = FLAG_TRUE;

    self->work[i].isloop  = FLAG_FALSE;
    self->work[i].fnum    = 0;
    self->work[i].snum    = 0;
  }

  self->master_volume = PCM8_MAX_VOLUME;
  self->master_pan = MDX_PAN_C;

  return;
}

/* ------------------------------------------------------------------ */

int pcm8_set_pcm_freq( int ch, int hz ) {

  __GETSELF;

  if ( self->pcm8_opened == FLAG_FALSE ) return 1;
  if ( ch >= PCM8_MAX_NOTE || ch < 0 ) return 1;
  if ( hz < 0 ) return 1;
  if ( hz >= PCM8_MAX_FREQ ) {
    self->work[ch].adpcm = FLAG_FALSE;
    self->work[ch].freq = 15600;
  } else {
    self->work[ch].freq = adpcm_freq_list[hz];
    self->work[ch].adpcm = FLAG_TRUE;
  }

  return 0;
}

int pcm8_note_on( int ch, int *ptr, int size, int* orig_ptr, int orig_size ) {

  __GETSELF;

  if (self->emdx->is_use_pcm8==FLAG_FALSE) {
	return 1;
      }
  
  if ( self->pcm8_opened == FLAG_FALSE ) return 1;
  if ( ch >= PCM8_MAX_NOTE || ch < 0 ) return 1;

  if ( self->work[ch].top_ptr!=NULL ) return 0; /* tie */

  if (self->work[ch].adpcm) {
    self->work[ch].ptr = ptr;
    self->work[ch].top_ptr = ptr;
    self->work[ch].end_ptr = ptr+size;
  } else {
    self->work[ch].ptr = orig_ptr;
    self->work[ch].top_ptr = orig_ptr;
    self->work[ch].end_ptr = orig_ptr+orig_size;
  }

  self->work[ch].isloop = FLAG_FALSE;

  return 0;
}

int pcm8_note_off( int ch ) {

  __GETSELF;

  if ( self->pcm8_opened == FLAG_FALSE ) return 1;
  if ( ch >= PCM8_MAX_NOTE || ch < 0 ) return 1;

  self->work[ch].ptr = NULL;
  self->work[ch].top_ptr = NULL;
  self->work[ch].end_ptr = NULL;

  self->work[ch].isloop = FLAG_FALSE;

  return 0;
}

int pcm8_set_volume( int ch, int val ) {

  __GETSELF;

  if ( self->pcm8_opened == FLAG_FALSE ) return 1;
  if ( ch >= PCM8_MAX_NOTE ) return 1;
  if ( val > PCM8_MAX_VOLUME ) return 1;
  if ( val < 0 ) return 1;

  self->work[ch].volume = val;

  return 0;
}

int pcm8_set_master_volume( int val ) {

  __GETSELF;

  if ( val > PCM8_MAX_VOLUME ) { return 1; }
  if ( val < 0 ) { return 1; }

  self->master_volume = val;

  return 0;
}

int pcm8_set_pan( int val ) {

  __GETSELF;

  if ( self->pcm8_opened == FLAG_FALSE ) { return 1; }
  self->master_pan = val;

  return 0;
}

/* internal helpers */

static void
__make_noise(pcm8_instances* self)
{
  int i=0;
  int j=0;
  int k=0;

  while ( i<self->sample_buffer_size ) {
    j = (int)(rand()%8192 - 4096) *1.5* self->emdx->fm_noise_vol / 127;
    for ( k=0 ; (k<(32-self->emdx->fm_noise_freq)/4+1&&i<self->sample_buffer_size) ; k++,i++ ) {
      self->sample_buffer2[i] = j;
    }
  }
}

/* ------------------------------------------------------------------ */

/* PCM8 main function: mixes all of PCM sound and OPM emulator */

void do_pcm8(s16 *lbuf,s16 *rbuf)
{
  int ch, i;
  int s=0;
  int *src, *dst;

  int is_dst_ran_out;
  int is_note_end;
  int f;
  int buffer_size;

  __GETSELF;

  buffer_size = self->sample_buffer_size;

  /* must I pronounce? */

  if ( self->pcm8_opened   == FLAG_FALSE  || 
       self->sample_buffer == NULL ) return;

  /* Execute YM2151 emulator */

  s16 *ym2151lbuf=self->ym2151_voice[0];
  s16 *ym2151rbuf=self->ym2151_voice[1];
  
  MemSet16DMA3(0,ym2151lbuf,self->sample_buffer_size*2);
  MemSet16DMA3(0,ym2151rbuf,self->sample_buffer_size*2);
  
  YM2151UpdateOne( ym2151_instance(), self->ym2151_voice, self->sample_buffer_size );

  /* reset buffer */

//  MemSet16DMA3(0,self->sample_buffer2,self->sample_buffer_size*2);

  if (self->emdx->haspdx==FLAG_TRUE) {
    for ( ch=0 ; ch<PCM8_MAX_NOTE ; ch++ ) {
      if ( !self->work[ch].ptr ) { continue; }
    
      /* frequency conversion */
      src = self->work[ch].ptr;
      dst = self->sample_buffer2;
    
      is_dst_ran_out=0;
      is_note_end=0;
      f=self->work[ch].fnum;
      s=self->work[ch].snum;
      
      while(is_dst_ran_out==0) {
	while( f>=0 ) {
	  s = *(src++) * self->work[ch].volume / PCM8_MAX_VOLUME;
	  if ( src >= self->work[ch].end_ptr ) {
	    src--;
	    is_note_end=1;
	  }
	  f -= self->dsp_speed;
	}
	while( f<0 ) {
	  *(dst++) += s;
	  f += self->work[ch].freq;
	  if ( dst >= self->sample_buffer2+self->sample_buffer_size ) {
	    is_dst_ran_out=1;
	    break;
	  }
	}
      }
      if ( is_note_end==1 ) {
	self->work[ch].ptr = NULL;
	self->work[ch].fnum = 0;
	self->work[ch].snum = 0;
      } else {
	self->work[ch].ptr = src;
	self->work[ch].fnum = f;
	self->work[ch].snum = s;
      }
    }
  }
  
  int *pcm8buf=self->sample_buffer2;
  
  MemCopy16DMA3(ym2151lbuf,lbuf,self->sample_buffer_size*2);
  MemCopy16DMA3(ym2151rbuf,rbuf,self->sample_buffer_size*2);
  return;
  
  int pcm8volume=self->master_volume/2;
  
  /* 16bit stereo */
  for ( i=0 ; i<self->sample_buffer_size ; i++ ) {
    int v;
    int l,r;
    
/*
    v = (*pcm8buf++)/2 * self->master_volume/PCM8_MAX_VOLUME;

    switch(self->master_pan) {
      case MDX_PAN_L: {
        l=v;
        r=0;
      } break;
      case MDX_PAN_R: {
        l=0;
        r=v;
      } break;
      default: {
        l=v;
        r=v;
      } break;
    }
*/
    v = (*pcm8buf++) * pcm8volume/PCM8_MAX_VOLUME;
    
    l=v;
    r=v;
    
    l += (int)(*ym2151lbuf++) * YM2151EMU_VOLUME;
    r += (int)(*ym2151rbuf++) * YM2151EMU_VOLUME;
    
    *lbuf++=l;
    *rbuf++=r;
  }
  
  return;

}

int pcm8_GetSampleBufferSize(void)
{
  __GETSELF;
  
  return(self->sample_buffer_size);
}

void
pcm8_wait_for_pcm_write(void)
{
}

void
pcm8_clear_buffer_flush_flag(void)
{
  __GETSELF;

  self->is_pcm_buffer_flushed = FLAG_FALSE;
}

int
pcm8_buffer_flush_flag(void)
{
  __GETSELF;

  return self->is_pcm_buffer_flushed;
}

/* ------------------------------------------------------------------ */

/* Timer hook */

void pcm8_start( void )
{
  __GETSELF;

  if ( self->pcm8_interrupt_active!=0 ) { return; }

//  t.it_value.tv_usec    = PCM8_SYSTEM_RATE * 1000; /* mili second */

  self->pcm8_interrupt_active = 1;
  return;
}

void pcm8_stop( void )
{
  __GETSELF;

  if ( self->pcm8_interrupt_active!=1 ) { return; }

  self->pcm8_interrupt_active=0;
  return;
}

