
#include <stdio.h>

#include <NDS.h>
#include <NDS/ARM9/CP15.h>

#include "plugin.h"
#include "plugin_def.h"
#include "fix.h"

#include "nsf_MMU.h"
#include "nsf_APU2A03.h"

u8 *RAM_Main=NULL; // 0x0000-
u8 *RAM_PPUIO=NULL; // 0x2000-
u8 *RAM_APUIO=NULL; // 0x4000-
u8 *RAM_EXROM=NULL; // 0x4020-
u8 *RAM_SRAM=NULL; // 0x6000-
u8 *RAM_ROM=NULL; // at bank (can't directaccess)
u8 *RAM_EXRAM=NULL; // EXRAM_Address SerialAccess

u8 EXRAM_Address;

u8 APU_MMC5_Multiple0,APU_MMC5_Multiple1;

u8 *pRAM_MainStack=NULL;

TPROMBank PROMBank[16];

//#define ROMMaxSize (128*1024)

void MMU_Init(void)
{
  u32 cnt;
  
  RAM_Main=(u8*)safemalloc(0x0800);
  pRAM_MainStack=&RAM_Main[0x100];
  
  RAM_PPUIO=(u8*)safemalloc(0x0008);
  
  RAM_APUIO=(u8*)safemalloc(0x0020);
  
  RAM_EXROM=(u8*)safemalloc(0x1fe0);
  
  RAM_SRAM=(u8*)safemalloc(0x2000);
  
//  RAM_ROM=(u8*)safemalloc(ROMMaxSize);
  RAM_ROM=NULL;
  
  RAM_EXRAM=(u8*)safemalloc(0x0080);
  
  EXRAM_Address=0x00;
  
  for(cnt=0;cnt<8;cnt++) MMU_SetBankSwitch(cnt,0);
  for(cnt=8;cnt<12;cnt++) MMU_SetBankSwitch(cnt,cnt-8);
  for(cnt=12;cnt<16;cnt++) MMU_SetBankSwitch(cnt,cnt-12);
  
  APU_MMC5_Multiple0=0x00;
  APU_MMC5_Multiple1=0x00;
}

void MMU_Free(void)
{
  if(RAM_Main!=NULL){
    safefree(RAM_Main); RAM_Main=NULL;
  }
  
  if(RAM_PPUIO!=NULL){
    safefree(RAM_PPUIO); RAM_PPUIO=NULL;
  }
  
  if(RAM_APUIO!=NULL){
    safefree(RAM_APUIO); RAM_APUIO=NULL;
  }
  
  if(RAM_EXROM!=NULL){
    safefree(RAM_EXROM); RAM_EXROM=NULL;
  }
  
  if(RAM_SRAM!=NULL){
    safefree(RAM_SRAM); RAM_SRAM=NULL;
  }
  
/*
  if(RAM_ROM!=NULL){
    safefree(RAM_ROM); RAM_ROM=NULL;
  }
*/
  RAM_ROM=NULL;
  
  if(RAM_EXRAM!=NULL){
    safefree(RAM_EXRAM); RAM_EXRAM=NULL;
  }
  
}

void MMU_LoadFromNSF(TNSFInfo *NSFInfo)
{
  u32 ofs;
  u32 cnt;
  
  RAM_APUIO[0x10]=0x10;
  RAM_APUIO[0x15]=0x0f;
  
//  for(cnt=0;cnt<ROMMaxSize;cnt++) RAM_ROM[cnt]=0x00;
  
  ofs=(NSFInfo->LoadAddress & 0x3fff);
  
//  for(cnt=0;cnt<NSFInfo->DataSize;cnt++) RAM_ROM[ofs+cnt]=NSFInfo->Data[cnt];
  RAM_ROM=&NSFInfo->Data[0]-ofs;
  
  if(NSFInfo->UseBankswitch==True){
    for(cnt=0;cnt<8;cnt++) MMU_SetBankSwitch(8+cnt,NSFInfo->BankswitchInitValues[cnt]);
    return;
  }
  
  if(NSFInfo->DataSize<0x2000){
    for(cnt=0;cnt<8;cnt++) MMU_SetBankSwitch(8+cnt,cnt % 2);
    return;
  }
  if(NSFInfo->DataSize<0x4000){
    for(cnt=0;cnt<8;cnt++) MMU_SetBankSwitch(8+cnt,cnt % 4);
    return;
  }
  
  for(cnt=0;cnt<8;cnt++) MMU_SetBankSwitch(8+cnt,cnt);
}

u8* MMU_GetAPUIOptr(void)
{
  return(RAM_APUIO);
}

void MMU_SetBankSwitch(u32 page,u32 _segment)
{
  PROMBank[page].ptr=&RAM_ROM[0x1000*_segment];
  PROMBank[page].Segment=_segment;
}

/*
u32 MMU_memread8main(u32 adr)
{
  return(RAM_Main[adr]);
}

u32 MMU_memread16main(u32 adr)
{
  return(MMU_memread8main(adr+0)+((u32)MMU_memread8main(adr+1) << 8));
}
*/

u32 MMU_memread8(u32 adr)
{
  if(adr<0x2000) return(RAM_Main[adr & 0x7ff]);
  
  if(0x8000<=adr) return(PROMBank[adr >> 12].ptr[adr & 0x0fff]);
  
  if(adr<0x4000) return(RAM_PPUIO[adr & 0x7]);
  
  if(adr<0x4020){
//    writelog(format('APU C==%6d R 0x%2x==0x%2x',[CPUCycle,adr-0x4000,Result]));
    
    return(RAM_APUIO[adr-0x4000]);
  }
  
  if(adr<0x6000) return(RAM_EXROM[adr-0x4020]);
  
  if(adr<0x8000) return(RAM_SRAM[adr-0x6000]);
  
  CPUHALT=True;
  
  return(0);
}

u32 MMU_memread16(u32 adr)
{
  return(MMU_memread8(adr+0)+((u32)MMU_memread8(adr+1) << 8));
}

/*
void MMU_memwrite8main(u32 adr,u32 data)
{
  RAM_Main[adr]=data;
}
*/

void MMU_memwrite8(u32 adr,u32 data)
{
  if(adr<0x2000){
    RAM_Main[adr & 0x7ff]=data;
    return;
  }
  
  if(0x8000<=adr){
    if(adr==0xf800){
      EXRAM_Address=data;
      APU_WriteEXRAM(adr,data);
      return;
    }
    CPUHALT=True;
    return;
  }
  
  if(adr<0x4000){
    RAM_PPUIO[adr & 0x7]=data;
    return;
  }
  
  if(adr<0x4020){
    RAM_APUIO[adr-0x4000]=data;
    APU_WriteAPUIO(adr,data);
    
    if(adr<0x4004){
//      _consolePrintf("APUW C=%6d $%02x=$%02x\n",CPUCycle,adr & 0xff,data);
    }
    return;
  }
  
  if(adr<0x6000){
    RAM_EXROM[adr-0x4020]=data;
    
    if((0x4040<=adr) && (adr<0x4100)){
      APU_WriteEXRAM(adr,data);
      return;
    }
    if( (adr==0x4800) || ((0x5000<=adr) && (adr<0x5016)) ){
      RAM_EXRAM[EXRAM_Address&0x7f]=data;
      if((EXRAM_Address & BIT7)!=0){
        EXRAM_Address=((EXRAM_Address+1) & 0x7f) | BIT7;
      }
//      writelog(format('EXRW $%04x=$%02x',[adr,data]));
      APU_WriteEXRAM(adr,data);
      return;
    }
    if(adr==0x5025){
      APU_MMC5_Multiple0=data;
      return;
    }
    if(adr==0x5026){
      APU_MMC5_Multiple1=data;
      return;
    }
    if((0x5ff0<=adr) && (adr<0x6000)){ // Bank switch
//      _consolePrintf("setbank %4x at %4x",adr,data);
      MMU_SetBankSwitch(adr & 0xF,data);
    }
    return;
  }
  
  if(adr<0x8000){
    RAM_SRAM[adr-0x6000]=data;
    return;
  }
  
  CPUHALT=True;
}

/*

void MMU_writestack8(u32 *REGS,u32 data)
{
  pRAM_MainStack[*REGS]=data;
  *REGS=(*REGS-1) & 0xff;
}

u32 MMU_readstack8(u32 *REGS)
{
  *REGS=(*REGS+1) & 0xff;
  return(pRAM_MainStack[*REGS]);
}

void MMU_writestack16(u32 *REGS,u32 data)
{
  MMU_writestack8(REGS,data >> 8);
  MMU_writestack8(REGS,data & 0xff);
}

u32 MMU_readstack16(u32 *REGS)
{
  u32 res;
  
  res=MMU_readstack8(REGS);
  res+=(u32)MMU_readstack8(REGS) << 8;
  
  return(res);
}

*/

