
#include <stdio.h>

#include <NDS.h>

#include "_const.h"

#include "plugin.h"
#include "plugin_def.h"

#include "boolean.h"

#include "libnsf/nsf_const.h"
#include "libnsf/nsf_MMU.h"
#include "libnsf/nsf_MC6502.h"
#include "libnsf/nsf_APU2A03.h"

// CallBack from plugin_dll.cpp

void cbLoadLibrary(void)
{
}

void cbFreeLibrary(void)
{
}

void cbQueryInterfaceLibrary(void)
{
}

// -----------------

static s32 fsize,fofs;

static bool LoadNSF(u8 *pb,u32 size)
{
#define readbyte(pos) (pb[pos])
#define readword(pos) (readbyte(pos)+((u16)readbyte(pos+1) << 8))
#define readstr(pos) (&pb[pos])
  
  if(size<0x80){
    _consolePrintf("DataSize:%d<0x80\n",size);
    return(false);
  }
  
  NSFInfo.DataSize=size-0x80;
  NSFInfo.Data=&pb[0x80];
  NSFInfo.PageCount=NSFInfo.DataSize/0x1000;
  
  for(u32 i=0;i<5;i++){
    NSFInfo.ID[i]=readbyte(0x00+i);
  }
  
  if((NSFInfo.ID[0]!='N')||(NSFInfo.ID[1]!='E')||(NSFInfo.ID[2]!='S')||(NSFInfo.ID[3]!='M')||(NSFInfo.ID[4]!=0x1a)){
    _consolePrintf("Illigal NSF format.\n");
    _consolePrintf("ID[]=");
    for(u32 i=0;i<8;i++){
      _consolePrintf("%02x,",pb[i]);
    }
    _consolePrintf("\n");
    return(false);
  }
  
  NSFInfo.TotalSongs=readbyte(0x06);
  NSFInfo.StartingSong=readbyte(0x07)-1;
  
  if(NSFInfo.TotalSongs==0){
    _consolePrintf("TotalSongs 0x%2x==0x00\n",NSFInfo.TotalSongs);
    return(false);
  }
  
  if(NSFInfo.StartingSong<=NSFInfo.TotalSongs){
    NSFInfo.StartingSong=NSFInfo.TotalSongs-1;
  }
  
  NSFInfo.LoadAddress=readword(0x08);
  
  if(NSFInfo.LoadAddress<0x8000){
    _consolePrintf("LoadAddress 0x%4x<0x8000\n",NSFInfo.LoadAddress);
    return(false);
  }
  
  NSFInfo.InitAddress=readword(0x0a);
  NSFInfo.PlayAddress=readword(0x0c);
  NSFInfo.strName=readstr(0x0e);
  NSFInfo.strArtist=readstr(0x2e);
  NSFInfo.strCopyright=readstr(0x4e);
  NSFInfo.NTSCTicksCount=readword(0x6e);
  
  NSFInfo.UseBankswitch=False;
  for(u32 i=0;i<8;i++){
    NSFInfo.BankswitchInitValues[i]=readbyte(0x70+i);
    if(NSFInfo.BankswitchInitValues[i]!=0x00) NSFInfo.UseBankswitch=True;
  }
  
  NSFInfo.PALTicksCount=readword(0x78);
  
  NSFInfo.VSyncMode=readbyte(0x7a) & BIT0;
  
  NSFInfo.ExtraSoundChipFlags=readbyte(0x7b);
  
  u8 excf=NSFInfo.ExtraSoundChipFlags;
  
  if((excf&ESCF_VRCVI)!=0) _consolePrintf("VRCVI,");
  if((excf&ESCF_VRCVII)!=0) _consolePrintf("VRCVII,");
  if((excf&ESCF_FDS)!=0) _consolePrintf("FDS,");
  if((excf&ESCF_MMC5)!=0) _consolePrintf("MMC5,");
  if((excf&ESCF_Namco106)!=0) _consolePrintf("Namco106,");
  if((excf&ESCF_SunsoftFME07)!=0) _consolePrintf("SunsoftFME07,");
  if((excf&ESCF_Future6)!=0) _consolePrintf("Future6,");
  if((excf&ESCF_Future7)!=0) _consolePrintf("Future7,");
  
  excf&=~ESCF_Namco106;
  
  if(excf!=0x00){
    _consolePrintf("notsupport ExtraSoundChip\n");
    return(false);
  }
  
  return(true);

#undef readbyte
#undef readword
#undef readstr
}

static void ShowNSFInfo(void)
{
#define astr(ttl,data) _consolePrintf("%s:%s\n",ttl,data);
#define au8(ttl,data) _consolePrintf("%s:$%02x\n",ttl,data);
#define au16(ttl,data) _consolePrintf("%s:$%04x\n",ttl,data);
#define au32(ttl,data) _consolePrintf("%s:$%04x\n",ttl,data);
  
  _consolePrintf("ID:%4s $%02x\n",(u32*)NSFInfo.ID,NSFInfo.ID[4]);

  au8("VersionNumber",NSFInfo.VersionNumber);

  au8("TotalSongs",NSFInfo.TotalSongs);
  au8("StartingSong",NSFInfo.StartingSong);
  au16("LoadAddress",NSFInfo.LoadAddress);
  au16("InitAddress",NSFInfo.InitAddress);
  au16("PlayAddress",NSFInfo.PlayAddress);
  astr("strName",NSFInfo.strName);
  astr("strArtist",NSFInfo.strArtist);
  astr("strCopyright",NSFInfo.strCopyright);
  au16("NTSCTicksCount",NSFInfo.NTSCTicksCount);
  
  astr("BankswitchInitValues[8]","...");
  if(NSFInfo.UseBankswitch==False){
    _consolePrintf("notused.\n");
    }else{
    for(u32 cnt=0;cnt<8;cnt++){
      _consolePrintf("%02x,",NSFInfo.BankswitchInitValues[cnt]);
    }
    _consolePrintf("\n");
  }
  
  au16("PALTicksCount",NSFInfo.PALTicksCount);
  au8("VSyncMode",NSFInfo.VSyncMode);
  au8("ExtraSoundChipFlags",NSFInfo.ExtraSoundChipFlags);

  au32("DataPtr",(u32)&NSFInfo.Data[0]);
  au32("DataSize",NSFInfo.DataSize);
  au32("PageCount",NSFInfo.PageCount);

#undef astr
#undef au8
#undef au16
#undef au32
}

static u8 *DeflateBuf=NULL;
static u32 DeflateSize;

// ------------------------------------------------------------------------------------

bool Start(int FileHandle)
{
  fseek(FileHandle,0,SEEK_END);
  DeflateSize=ftell(FileHandle);
  fseek(FileHandle,0,SEEK_SET);
  
  DeflateBuf=(u8*)safemalloc(DeflateSize);
  if(DeflateBuf==NULL) return(false);
  fread(DeflateBuf,1,DeflateSize,FileHandle);
  
  if(LoadNSF(DeflateBuf,DeflateSize)==false){
    safefree(DeflateBuf); DeflateBuf=NULL;
//    ShowNSFInfo();
    _consolePrintf("NSF LoadError.\n");
    CPUHALT=True;
    return(false);
  }
  ShowNSFInfo();
  
  CPUHALT=False;
  
  fsize=(s32)(90*APU2A03_Renderfps); // 90sec.
  fofs=0;
  
  MMU_Init();
  MC6502_Init();
  APU2A03_Init();
  
  MMU_LoadFromNSF(&NSFInfo);
  MC6502_LoadFromNSF(&NSFInfo);
  APU2A03_LoadFromNSF(&NSFInfo);
  
  MC6502_SetREGs(0,NSFInfo.VSyncMode,0);
  MC6502_ExecRoutine(NSFInfo.InitAddress);
  
  _consolePrintSet(0,0);
  _consolePrintf("-- Init %dCycle.     \n",CPUCycle);
  
  return(true);
}

void Free(void)
{
  MMU_Free();
  MC6502_Free();
  APU2A03_Free();
  
  CPUHALT=True;
  
  if(DeflateBuf!=NULL){
    safefree(DeflateBuf); DeflateBuf=NULL;
  }
}

u32 Update(s16 *lbuf,s16 *rbuf)
{
  if(CPUHALT==True) return(0);
  
  MC6502_ExecRoutine(NSFInfo.PlayAddress);
  
  if(lbuf!=NULL){
    APU2A03_GenWave(lbuf,APU2A03_SampleSize);
    _consolePrintSet(0,1);
    _consolePrintf("-- Update %dCycle.    \n",CPUCycle);
    
    bool exists=false;
    
    s16 chkdata=lbuf[0];
    for(u32 idx=1;idx<APU2A03_SampleSize;idx++){
      if(chkdata!=lbuf[idx]) exists=true;
    }
    
    static int BlankCount=0;
    if(fofs==0) BlankCount=0;
    
    if(exists==true){
      BlankCount=(int)(APU2A03_Renderfps*3);
      }else{
      if(BlankCount!=0){
        BlankCount--;
        if(BlankCount==0) return(0);
      }
    }
    
  }
  
  fofs++;
  
  return(APU2A03_SampleSize);
}

s32 GetPosMax(void)
{
  return(fsize);
}

s32 GetPosOffset(void)
{
  return(fofs);
}

void SetPosOffset(s32 ofs)
{
  if(CPUHALT==True) return;
  
  if(ofs==fofs) return;
  
  if(ofs<fofs){
    fofs=0;
    
    CPUHALT=True;
    
    MMU_Free();
    MC6502_Free();
    APU2A03_Free();
    
    CPUHALT=False;
    
    MMU_Init();
    MC6502_Init();
    APU2A03_Init();
    
    MMU_LoadFromNSF(&NSFInfo);
    MC6502_LoadFromNSF(&NSFInfo);
    APU2A03_LoadFromNSF(&NSFInfo);
    
    MC6502_SetREGs(0,NSFInfo.VSyncMode,0);
    MC6502_ExecRoutine(NSFInfo.InitAddress);
  }
  
  u32 baseofs=fofs;
  
  MWin_ProgressShow("Seek...",ofs-baseofs);
  MWin_ProgressSetPos(fofs-baseofs);
  
  while(fofs<ofs){
    if((fofs&0xf)==0) MWin_ProgressSetPos(fofs-baseofs);
    if(Update(NULL,NULL)==0) break;
    if(CPUHALT==True) break;
  }
  MWin_ProgressHide();
}

u32 GetChannelCount(void)
{
  return(1);
}

u32 GetSampleRate(void)
{
  return(APU2A03_SampleRate);
}

u32 GetSamplePerFrame(void)
{
  return(APU2A03_SampleSize);
}

int GetInfoIndexCount(void)
{
  return(6);
}

bool GetInfoStrL(int idx,char *str,int len)
{
  switch(idx){
    case 0: snprintf(str,len,"Name=%s",NSFInfo.strName); return(true); break;
    case 1: snprintf(str,len,"Artist=%s",NSFInfo.strArtist); return(true); break;
    case 2: snprintf(str,len,"Copyright=%s",NSFInfo.strCopyright); return(true); break;
    case 3: {
/*
      u8 excf=NSFInfo.ExtraSoundChipFlags;
      if(excf==0) strcat(str,"not use extra sound chips.");
      if((excf&ESCF_VRCVI)!=0) strcat(str,"VRCVI,");
      if((excf&ESCF_VRCVII)!=0) strcat(str,"VRCVII,");
      if((excf&ESCF_FDS)!=0) strcat(str,"FDS,");
      if((excf&ESCF_MMC5)!=0) strcat(str,"MMC5,");
      if((excf&ESCF_Namco106)!=0) strcat(str,"Namco106,");
      if((excf&ESCF_SunsoftFME07)!=0) strcat(str,"SunsoftFME07,");
      if((excf&ESCF_Future6)!=0) strcat(str,"Future6,");
      if((excf&ESCF_Future7)!=0) strcat(str,"Future7,");
*/
      return(true);
    } break;
    case 4: snprintf(str,len,"Load0x%x Init0x%x Play0x%x",NSFInfo.LoadAddress,NSFInfo.InitAddress,NSFInfo.PlayAddress); return(true); break;
    case 5: snprintf(str,len,"NTSCTicks=%d PALTicks=%d",NSFInfo.NTSCTicksCount,NSFInfo.PALTicksCount); return(true); break;
  }
  return(false);
}

bool GetInfoStrW(int idx,UnicodeChar *str,int len)
{
  return(false);
}

bool GetInfoStrUTF8(int idx,char *str,int len)
{
  return(false);
}

// -----------------------------------------------------------

