
#include <stdio.h>

#include <NDS.h>

#include "_const.h"

#include "plugin.h"
#include "plugin_def.h"
#include "std.h"

#include "inifile.h"

#include "smidlib.h"
#include "rcplib.h"

// CallBack from plugin_dll.cpp

void cbLoadLibrary(void)
{
}

void cbFreeLibrary(void)
{
}

void cbQueryInterfaceLibrary(void)
{
}

// -----------------

static u8 *DeflateBuf=NULL;
static u32 DeflateSize;

static u32 SamplePerFrame;

static int TotalClock,CurrentClock;

static u32 ClockCur;

enum EFileFormat {EFF_MID,EFF_RCP};

static EFileFormat FileFormat;

#define MaxSampleRate (48000)

ALIGNED_VAR_IN_DTCM s16 DTCM_lbuf[(MaxSampleRate/120)+1],DTCM_rbuf[(MaxSampleRate/120)+1];

// --------------------------------------------------------------------

// ------------------------------------------------------------------------------------

void Free(void);

static void selSetParam(u8 *data,u32 SampleRate,u32 MaxChannelCount,u32 GenVolume)
{
  switch(FileFormat){
    case EFF_MID: smidlibSetParam(DeflateBuf,SampleRate,MaxChannelCount,GenVolume); break;
    case EFF_RCP: rcplibSetParam(DeflateBuf,SampleRate,MaxChannelCount,GenVolume); break;
  }
}

static bool selStart(void)
{
  switch(FileFormat){
    case EFF_MID: return(smidlibStart()); break;
    case EFF_RCP: return(rcplibStart()); break;
  }
  
  return(false);
}

static void selFree(void)
{
  switch(FileFormat){
    case EFF_MID: smidlibFree(); break;
    case EFF_RCP: rcplibFree(); break;
  }
}

static int selGetNearClock(void)
{
  switch(FileFormat){
    case EFF_MID: return(smidlibGetNearClock()); break;
    case EFF_RCP: return(rcplibGetNearClock()); break;
  }
  
  return(0);
}

static bool selNextClock(bool ShowEventMessage,bool EnableNote,int DecClock)
{
  switch(FileFormat){
    case EFF_MID: return(smidlibNextClock(ShowEventMessage,EnableNote,DecClock)); break;
    case EFF_RCP: return(rcplibNextClock(ShowEventMessage,EnableNote,DecClock)); break;
  }
  
  return(false);
}

static void selAllSoundOff(void)
{
  switch(FileFormat){
    case EFF_MID: smidlibAllSoundOff(); break;
    case EFF_RCP: rcplibAllSoundOff(); break;
  }
}

static bool sel_isAllTrackEOF(void)
{
  switch(FileFormat){
    case EFF_MID: return(SM_isAllTrackEOF()); break;
    case EFF_RCP: return(RCP_isAllTrackEOF()); break;
  }
  
  return(false);
}

static u32 sel_GetSamplePerClockFix16(void)
{
  switch(FileFormat){
    case EFF_MID: return(SM_GetSamplePerClockFix16()); break;
    case EFF_RCP: return(RCP_GetSamplePerClockFix16()); break;
  }
  
  return(0);
}

static void Start_smidlibDetectTotalClock()
{
  TSM_Track *pSM_Track=NULL;
  u32 trklen=0;
  
  {
    for(u32 idx=0;idx<StdMIDI.SM_Chank.Track;idx++){
      TSM_Track *pCurSM_Track=&StdMIDI.pSM_Tracks[idx];
      u32 ctrklen=(u32)pCurSM_Track->DataEnd-(u32)pCurSM_Track->Data;
      if(trklen<ctrklen){
        trklen=ctrklen;
        pSM_Track=pCurSM_Track;
      }
    }
  }
  
  if((pSM_Track==NULL)||(trklen==0)){
    _consolePrintf("Fatal error.\n");
    ShowLogHalt();
  }
  
  MWin_ProgressShow("Loading...",trklen);
  
  int LastClock=0;
  int PrevDiv=StdMIDI.SM_Chank.TimeRes*8;
  
  while(1){
    if(PrevDiv<(TotalClock-LastClock)){
      LastClock=TotalClock;
      u32 lastlen=(u32)pSM_Track->DataEnd-(u32)pSM_Track->Data;
      if(lastlen<trklen) MWin_ProgressSetPos(trklen-lastlen);
    }
    int DecClock=smidlibGetNearClock();
    if(smidlibNextClock(false,false,DecClock)==false) break;
    TotalClock+=DecClock;
  }
  
  MWin_ProgressHide();
}

static void Start_rcplibDetectTotalClock()
{
  TRCP_Track *pRCP_Track=NULL;
  u32 trklen=0;
  
  {
    for(u32 idx=0;idx<RCP.RCP_Chank.TrackCount;idx++){
      TRCP_Track *pCurRCP_Track=&RCP.RCP_Track[idx];
      u32 ctrklen=(u32)pCurRCP_Track->DataEnd-(u32)pCurRCP_Track->Data;
      if(trklen<ctrklen){
        trklen=ctrklen;
        pRCP_Track=pCurRCP_Track;
      }
    }
  }
  
  if((pRCP_Track==NULL)||(trklen==0)){
    _consolePrintf("Fatal error.\n");
    ShowLogHalt();
  }
  
  MWin_ProgressShow("Loading...",trklen);
  
  int LastClock=0;
  int PrevDiv=RCP.RCP_Chank.TimeRes*8;
  
  while(1){
    if(PrevDiv<(TotalClock-LastClock)){
      LastClock=TotalClock;
      u32 lastlen=(u32)pRCP_Track->DataEnd-(u32)pRCP_Track->Data;
      if(lastlen<trklen) MWin_ProgressSetPos(trklen-lastlen);
    }
    int DecClock=rcplibGetNearClock();
    if(rcplibNextClock(false,false,DecClock)==false) break;
    TotalClock+=DecClock;
  }
  
  MWin_ProgressHide();
}

bool Start(int FileHandle)
{
  InitINI();
  LoadINI(GetINIData(),GetINISize());
  
  {
    TiniMIDPlugin *MIDPlugin=&GlobalINI.MIDPlugin;
    if(MaxSampleRate<MIDPlugin->SampleRate) MIDPlugin->SampleRate=MaxSampleRate;
  }
  
  fseek(FileHandle,0,SEEK_END);
  DeflateSize=ftell(FileHandle);
  fseek(FileHandle,0,SEEK_SET);
  
  DeflateBuf=(u8*)safemalloc(DeflateSize);
  if(DeflateBuf==NULL) return(false);
  fread(DeflateBuf,1,DeflateSize,FileHandle);
  
  if(GetBINFileHandle()==0){
    _consolePrintf("not found SoundFont(smid.bin) file.\n");
    if(DeflateBuf!=NULL){
      safefree(DeflateBuf); DeflateBuf=NULL;
    }
    return(false);
  }
  
  PCH_SetProgramMap(GetBINFileHandle());
  
  {
    bool detect=false;
    
    if(strncmp((char*)DeflateBuf,"MThd",4)==0){
      _consolePrintf("Start Standard MIDI file.\n");
      detect=true;
      FileFormat=EFF_MID;
    }
    if(strncmp((char*)DeflateBuf,"RCM-PC98V2.0(C)COME ON MUSIC",28)==0){
      _consolePrintf("Start RecomposerV2.0 file.\n");
      detect=true;
      FileFormat=EFF_RCP;
    }
    
    if(detect==false){
      _consolePrintf("Unknown file format!!\n");
      PCH_FreeProgramMap();
      if(DeflateBuf!=NULL){
        safefree(DeflateBuf); DeflateBuf=NULL;
      }
      return(false);
    }
  }
  
  TiniMIDPlugin *MIDPlugin=&GlobalINI.MIDPlugin;
  
  SamplePerFrame=MIDPlugin->SampleRate/120; // 120fps
  
  selSetParam(DeflateBuf,MIDPlugin->SampleRate,MIDPlugin->MaxVoiceCount,MIDPlugin->GenVolume);
  if(selStart()==false){
    Free();
    return(false);
  }
  
  TotalClock=0;
  CurrentClock=0;
  
  switch(FileFormat){
    case EFF_MID: Start_smidlibDetectTotalClock(); break;
    case EFF_RCP: Start_rcplibDetectTotalClock(); break;
  }
  
  if(TotalClock==0){
    _consolePrintf("Detect TotalClock equal Zero.\n");
    Free();
    return(false);
  }
  
  selFree();
  selStart();
  
  selNextClock(MIDPlugin->ShowEventMessage,true,selGetNearClock());
  
  ClockCur=0;
  
  return(true);
}

void Free(void)
{
  selFree();
  
  PCH_FreeProgramMap();
  
  if(DeflateBuf!=NULL){
    safefree(DeflateBuf); DeflateBuf=NULL;
  }
}

u32 Update(s16 *lbuf,s16 *rbuf)
{
  if(sel_isAllTrackEOF()==true){
    MemSet16DMA3(0,lbuf,SamplePerFrame*2);
    MemSet16DMA3(0,rbuf,SamplePerFrame*2);
    return(0);
  }
  
  TiniMIDPlugin *MIDPlugin=&GlobalINI.MIDPlugin;
  
  int ProcClock=0;
  
  {
    u32 SamplePerFrameFix16=SamplePerFrame*0x10000;
    u32 SamplePerClockFix16=sel_GetSamplePerClockFix16();
    while(ClockCur<SamplePerFrameFix16){
      ProcClock++;
      ClockCur+=SamplePerClockFix16;
    }
    ClockCur-=SamplePerFrame*0x10000;
  }
  
  while(ProcClock!=0){
    int DecClock=selGetNearClock();
    if(ProcClock<DecClock) DecClock=ProcClock;
    if(selNextClock(MIDPlugin->ShowEventMessage,true,DecClock)==false) break;
    CurrentClock+=DecClock;
    ProcClock-=DecClock;
  }
  
  PCH_NextClock();
  
  if((lbuf!=NULL)&&(rbuf!=NULL)){
    TiniMIDPlugin *MIDPlugin=&GlobalINI.MIDPlugin;
    int Reverb=PCH_GetGlobalReverb()*MIDPlugin->ReverbFactor/150;
    if(100<Reverb) Reverb=100;
    
    s16 *dstlbuf=DTCM_lbuf;
    s16 *dstrbuf=DTCM_rbuf;
    
    if(Reverb==0){
      MemSet16CPU(0,dstlbuf,SamplePerFrame*2);
      MemSet16CPU(0,dstrbuf,SamplePerFrame*2);
      }else{
      for(u32 idx=0;idx<SamplePerFrame-1;idx++){
        s32 l,r;
        l=lbuf[idx+0]+lbuf[idx+1];
        dstrbuf[idx]=l*Reverb/(128*2);
        r=rbuf[idx+0]+rbuf[idx+1];
        dstlbuf[idx]=r*Reverb/(128*2);
      }
      {
        u32 idx=SamplePerFrame-1;
        s32 l,r;
        l=lbuf[idx];
        dstrbuf[idx]=l*Reverb/128;
        r=rbuf[idx];
        dstlbuf[idx]=r*Reverb/128;
      }
    }
    
    PCH_Render(dstlbuf,dstrbuf,SamplePerFrame);
    
    for(u32 idx=0;idx<SamplePerFrame;idx++){
      lbuf[idx]=dstlbuf[idx];
      rbuf[idx]=dstrbuf[idx];
    }
  }
  
  return(SamplePerFrame);
}

s32 GetPosMax(void)
{
  return(TotalClock);
}

s32 GetPosOffset(void)
{
  return(CurrentClock);
}

void SetPosOffset(s32 ofs)
{
  if(ofs<CurrentClock){
    selFree();
    selStart();
    CurrentClock=0;
    }else{
    selAllSoundOff();
  }
  
  while(CurrentClock<=ofs){
    int DecClock=selGetNearClock();
    if(selNextClock(false,false,DecClock)==false) break;
    CurrentClock+=DecClock;
  }
  
  ClockCur=0;
}

u32 GetChannelCount(void)
{
  return(2);
}

u32 GetSampleRate(void)
{
  TiniMIDPlugin *MIDPlugin=&GlobalINI.MIDPlugin;
  
  return(MIDPlugin->SampleRate);
}

u32 GetSamplePerFrame(void)
{
  return(SamplePerFrame);
}

int GetInfoIndexCount(void)
{
  if(GetBINSize()<1*1024*1024){
    return(5);
  }
  
  switch(FileFormat){
    case EFF_MID: return(5); break;
    case EFF_RCP: return(15); break;
  }
  
  return(0);
}

bool GetInfoStrL(int idx,char *str,int len)
{
  if(GetBINSize()<1*1024*1024){
    switch(idx){
      case 0: snprintf(str,len,"This sound font is very compact."); return(true); break;
      case 1: snprintf(str,len,"Install full sound font from web site."); return(true); break;
      case 2: snprintf(str,len,"̓RpNgȃTEhtHgłB"); return(true); break;
      case 3: snprintf(str,len,"EFuTCg犮SłDLĂB"); return(true); break;
      case 4: snprintf(str,len,"http://mdxonline.dyndns.org/"); return(true); break;
    }
    return(false);
  }
  
  if(idx==0){
    if(MemoryOverflowFlag==true){
      snprintf(str,len,"Memory Overflow!! removed text font?");
      return(true);
    }
  }
  
  switch(FileFormat){
    case EFF_MID: {
      TSM_Chank *_SM_Chank=&StdMIDI.SM_Chank;
      switch(idx){
        case 0: snprintf(str,len,"Format=%d Tracks=%d TimeRes=%d",_SM_Chank->Format,_SM_Chank->Track,_SM_Chank->TimeRes); return(true); break;
        case 1: snprintf(str,len,"Title=%s",(gME_Title!=NULL) ? gME_Title : ""); return(true); break;
        case 2: snprintf(str,len,"Copyright=%s",(gME_Copyright!=NULL) ? gME_Copyright : ""); return(true); break;
        case 3: snprintf(str,len,"Text=%s",(gME_Text!=NULL) ? gME_Text : ""); return(true); break;
        case 4: snprintf(str,len,"TotalClock=%d",TotalClock); return(true); break;
      }
    } break;
    case EFF_RCP: {
      TRCP_Chank *pRCP_Chank=&RCP.RCP_Chank;
      switch(idx){
        case 0: snprintf(str,len,"TimeRes=%d Tempo=%d PlayBias=%d",pRCP_Chank->TimeRes,pRCP_Chank->Tempo,pRCP_Chank->PlayBias); return(true); break;
        case 1: snprintf(str,len,"TrackCount=%d TotalClock=%d",pRCP_Chank->TrackCount,TotalClock); return(true); break;
        case 2: snprintf(str,len,"%0.64s",pRCP_Chank->Title); return(true); break;
        case 3: snprintf(str,len,"%0.28s",&pRCP_Chank->Memo[0*28]); return(true); break;
        case 4: snprintf(str,len,"%0.28s",&pRCP_Chank->Memo[1*28]); return(true); break;
        case 5: snprintf(str,len,"%0.28s",&pRCP_Chank->Memo[2*28]); return(true); break;
        case 6: snprintf(str,len,"%0.28s",&pRCP_Chank->Memo[3*28]); return(true); break;
        case 7: snprintf(str,len,"%0.28s",&pRCP_Chank->Memo[4*28]); return(true); break;
        case 8: snprintf(str,len,"%0.28s",&pRCP_Chank->Memo[5*28]); return(true); break;
        case 9: snprintf(str,len,"%0.28s",&pRCP_Chank->Memo[6*28]); return(true); break;
        case 10: snprintf(str,len,"%0.28s",&pRCP_Chank->Memo[7*28]); return(true); break;
        case 11: snprintf(str,len,"%0.28s",&pRCP_Chank->Memo[8*28]); return(true); break;
        case 12: snprintf(str,len,"%0.28s",&pRCP_Chank->Memo[9*28]); return(true); break;
        case 13: snprintf(str,len,"%0.28s",&pRCP_Chank->Memo[10*28]); return(true); break;
        case 14: snprintf(str,len,"%0.28s",&pRCP_Chank->Memo[11*28]); return(true); break;
      }
    } break;
  }
  
  return(false);
}

bool GetInfoStrW(int idx,UnicodeChar *str,int len)
{
  return(false);
}

bool GetInfoStrUTF8(int idx,char *str,int len)
{
  return(false);
}

// -----------------------------------------------------------

