
#include <stdio.h>

#include <nds.h>

#include "plugin.h"
#include "plugin_def.h"
#include "std.h"

#include "smidlib.h"

#include "powtbl_bin.h"
u16 *powtbl=(u16*)powtbl_bin;
#define POWRANGE (1024)

#include "sintbl_bin.h"
int *sintbl=(int*)sintbl_bin;
#define SINRANGE (120)

//////////////////////////////////////////////////////////////////////

enum EEnvState {EES_None,EES_Attack,EES_Decay,EES_Sustain,EES_Release3,EES_Release4,EES_Release5};

typedef struct {
  u32 SampleRate;
  u32 RootFreq; // fix16.16
  u32 MasterVolume;
  u32 FractionStart;
  u32 FractionEnd;
  u32 Length,LoopStart,LoopEnd;
  u32 LoopFlag,s16Flag;
  u32 MasterPanpot;
  u32 Note;
  u32 VibSweep;
  u32 VibRatio;
  int VibDepth;
  bool EnvelopeFlag;
  int EnvRate[6];
  u32 EnvOfs[6];
  u32 Data[];
} TProgram;

typedef struct {
  u32 PatchCount;
  TProgram *pPrg[16];
} TProgramPatch;

typedef struct {
  int FileHandle;
  u32 FileOffset;
  
  u32 DataTable[128*2];
  TProgramPatch *ppData[128];
} TProgramMap;

typedef struct {
  TProgramMap *pPrgMaps[128];
} TVariationMap;

ALIGNED_VAR_IN_DTCM TVariationMap VarMap_Tone;
ALIGNED_VAR_IN_DTCM TVariationMap VarMap_Drum;

// ----------------------------------------------------

typedef struct {
  bool Enabled;
  bool DrumMode;
  u32 Track;
  u32 Note,Vol,Exp,Vel,Reverb;
  u32 Freq;
  u32 MasterVolume;
  
  u32 OnClock,OffClock;
  
  u32 GT;
  
  bool Pedal;
  
  bool VibEnable;
  u32 VibSweepAdd,VibSweepCur;
  u32 VibAdd,VibCur;
  int VibDepth;
  int VibPhase;
  
  bool ModEnable;
  u32 ModAdd,ModCur;
  int ModDepth;
  int ModPhase;
  
  EEnvState EnvState;
  int EnvSpeed;
  u32 EnvCurLevel;
  u32 EnvEndLevel;
  u32 EnvRelLevel;
  
  TProgram *pPrg;
  u32 PrgPos;
  u32 PrgPanpot;
  u32 PrgMstVol,PrgVol;
  s32 PrgFraction;
  
  u32 FreqAddFix16;
  u32 FreqCurFix16;
  s32 CurSampleData,LastSampleData;
} TPCH;

static u32 PCHCount;
static TPCH *pPCH;

#define PCHCountMax (20)
ALIGNED_VAR_IN_DTCM TPCH DTCM_PCH[PCHCountMax];

// ----------------------------------------------------

static u32 Note2FreqTableFix16[128];

static u32 SampleRate;

static u32 GenVolume;

bool MemoryOverflowFlag;

// ----------------------------------------------------

static void PCH_SetProgramMap_Load(TProgramMap *pPrgMap,int FileHandle,u32 FileOffset)
{
  MemoryOverflowFlag=false;
  
  pPrgMap->FileHandle=FileHandle;
  pPrgMap->FileOffset=FileOffset;
  
  fseek(pPrgMap->FileHandle,pPrgMap->FileOffset,SEEK_SET);
  
  for(u32 idx=0;idx<128;idx++){
    fread(&pPrgMap->DataTable[idx*2+0],1,4,pPrgMap->FileHandle);
    fread(&pPrgMap->DataTable[idx*2+1],1,4,pPrgMap->FileHandle);
    pPrgMap->ppData[idx]=NULL;
  }
}

void PCH_SetProgramMap(int FileHandle)
{
  for(int prg=0;prg<128;prg++){
    VarMap_Tone.pPrgMaps[prg]=NULL;
    VarMap_Drum.pPrgMaps[prg]=NULL;
  }
  
  u32 ID;
  u32 ToneOffset[128];
  u32 DrumOffset[128];
  
  fseek(FileHandle,0x0,SEEK_SET);
  fread(&ID,1,4,FileHandle);
  
  for(int prg=0;prg<128;prg++){
    fread(&ToneOffset[prg],1,4,FileHandle);
  }
  for(int prg=0;prg<128;prg++){
    fread(&DrumOffset[prg],1,4,FileHandle);
  }
  
  for(int prg=0;prg<128;prg++){
    if(ToneOffset[prg]!=0){
      VarMap_Tone.pPrgMaps[prg]=(TProgramMap*)safemalloc(sizeof(TProgramMap));
      PCH_SetProgramMap_Load(VarMap_Tone.pPrgMaps[prg],FileHandle,ToneOffset[prg]);
    }
  }
  
  for(int prg=0;prg<128;prg++){
    if(DrumOffset[prg]!=0){
      VarMap_Drum.pPrgMaps[prg]=(TProgramMap*)safemalloc(sizeof(TProgramMap));
      PCH_SetProgramMap_Load(VarMap_Drum.pPrgMaps[prg],FileHandle,DrumOffset[prg]);
    }
  }
}

static void PCH_FreeProgramMap_Free_Patch(TProgramPatch *pPrgPatch)
{
  for(u32 patidx=0;patidx<pPrgPatch->PatchCount;patidx++){
    safefree(pPrgPatch->pPrg[patidx]); pPrgPatch->pPrg[patidx]=NULL;
  }
}

static void PCH_FreeProgramMap_Free(TProgramMap *pPrgMap)
{
  for(u32 idx=0;idx<128;idx++){
    if(pPrgMap->ppData[idx]!=NULL){
      PCH_FreeProgramMap_Free_Patch(pPrgMap->ppData[idx]);
      safefree(pPrgMap->ppData[idx]); pPrgMap->ppData[idx]=NULL;
    }
  }
}

void PCH_FreeProgramMap(void)
{
  for(u32 idx=0;idx<128;idx++){
    if(VarMap_Tone.pPrgMaps[idx]!=NULL){
      PCH_FreeProgramMap_Free(VarMap_Tone.pPrgMaps[idx]);
      safefree(VarMap_Tone.pPrgMaps[idx]); VarMap_Tone.pPrgMaps[idx]=NULL;
    }
    if(VarMap_Drum.pPrgMaps[idx]!=NULL){
      PCH_FreeProgramMap_Free(VarMap_Drum.pPrgMaps[idx]);
      safefree(VarMap_Drum.pPrgMaps[idx]); VarMap_Drum.pPrgMaps[idx]=NULL;
    }
  }
}

static u32 GetVariationNumber(TVariationMap *pVarMap,u32 VarNum,u32 PrgNum,bool DrumMode)
{
  if(128<=VarNum) return((u32)-1);
  if(128<=PrgNum) return((u32)-1);
  
  for(;VarNum!=0;VarNum--){
    if(pVarMap->pPrgMaps[VarNum]!=NULL){
      TProgramMap *pPrgMap=pVarMap->pPrgMaps[VarNum];
      if(pPrgMap!=NULL){
        u32 ofs=pPrgMap->DataTable[PrgNum*2+0];
        if(ofs!=0) return(VarNum);
      }
      break;
    }
    if(DrumMode==true) break;
  }
  
  VarNum=0;
  
  {
    TProgramMap *pPrgMap=pVarMap->pPrgMaps[VarNum];
    if(pPrgMap!=NULL){
      if(DrumMode==false){
        return(VarNum);
        }else{
        if(pPrgMap->DataTable[PrgNum*2+0]!=0) return(VarNum);
      }
    }
  }
  
  return((u32)-1);
}

static TProgramPatch *GetProgramPatchPtr(TVariationMap *pVarMap,u32 VarNum,u32 PrgNum,bool DrumMode)
{
  if(128<=VarNum) return(NULL);
  if(128<=PrgNum) return(NULL);
  
  for(;VarNum!=0;VarNum--){
    if(pVarMap->pPrgMaps[VarNum]!=NULL){
      TProgramMap *pPrgMap=pVarMap->pPrgMaps[VarNum];
      if(pPrgMap!=NULL){
        if(pPrgMap->ppData[PrgNum]!=NULL) return(pPrgMap->ppData[PrgNum]);
      }
      break;
    }
    if(DrumMode==true) break;
  }
  
  VarNum=0;
  
  {
    TProgramMap *pPrgMap=pVarMap->pPrgMaps[VarNum];
    if(pPrgMap!=NULL){
      if(DrumMode==false){
        if(pPrgMap->ppData[PrgNum]!=NULL) return(pPrgMap->ppData[PrgNum]);
        PrgNum&=~7; // find captital tone
        if(pPrgMap->ppData[PrgNum]!=NULL) return(pPrgMap->ppData[PrgNum]);
        }else{
        if(pPrgMap->ppData[PrgNum]!=NULL) return(pPrgMap->ppData[PrgNum]);
      }
    }
  }
  
  return(NULL);
}

static TProgram *GetProgramFromPatchPtr(TProgramPatch *pPrgPatch,s32 Note,bool DrumMode)
{
//  return(pPrgPatch->pPrg[0]);
  
  if(pPrgPatch==NULL) return(NULL);
  if(128<=Note) return(NULL);
  
  if(DrumMode==true) return(pPrgPatch->pPrg[0]);
  
  if(pPrgPatch->PatchCount==0) return(NULL);
  if(pPrgPatch->PatchCount==1) return(pPrgPatch->pPrg[0]);
  
  u32 NoteFreq=Note2FreqTableFix16[Note];
  
  u32 NearPatchIndex=0x100;
  u32 NearFreq=0xffffffff;
  
  for(u32 patidx=0;patidx<pPrgPatch->PatchCount;patidx++){
    u32 RootFreq=pPrgPatch->pPrg[patidx]->RootFreq;
    
    u32 sub;
    if(NoteFreq<RootFreq){
      sub=RootFreq-NoteFreq;
      }else{
      sub=NoteFreq-RootFreq;
    }
    
    if(sub<=NearFreq){
      NearFreq=sub;
      NearPatchIndex=patidx;
    }
    
//    _consolePrintf("%d,%d,%d,%d,%d\n",patidx,NoteFreq>>16,NearFreq>>16,RootFreq>>16,NearPatchIndex);
  }
  
  if(NearPatchIndex==0x100){
    _consolePrintf("find near root freq from patch list??\n");
    return(NULL);
  }
  
//  _consolePrintf("near %4d,%4d,%d\n",NoteFreq>>16,NearFreq>>16,NearPatchIndex);
  
  return(pPrgPatch->pPrg[NearPatchIndex]);
}

bool PCH_LoadProgram(s32 Note,u32 var,u32 prg,bool DrumMode)
{
  {
    void *pchk=safemalloc(128*1024); // 128kb͕Kc
    if(pchk==NULL){
      MemoryOverflowFlag=true;
      return(false);
    }
    safefree(pchk); pchk=NULL;
  }
  
  if(Note==0) return(false);
  
  TVariationMap *pVarMap;
  u32 VarNum,PrgNum;
  
  if(DrumMode==true){
    pVarMap=&VarMap_Drum;
    VarNum=prg;
    PrgNum=Note;
    }else{
    pVarMap=&VarMap_Tone;
    VarNum=var;
    PrgNum=prg;
  }
  
  VarNum=GetVariationNumber(pVarMap,VarNum,PrgNum,DrumMode);
  if(VarNum==(u32)-1) return(false);
  
  TProgramMap *pPrgMap=pVarMap->pPrgMaps[VarNum];
  if(pPrgMap==NULL) return(false);
  
  if(pPrgMap->ppData[PrgNum]!=NULL) return(true); // already loaded.
  
  u32 ofs=pPrgMap->DataTable[PrgNum*2+0];
  u32 size=pPrgMap->DataTable[PrgNum*2+1];
  
  if((ofs==0)||(size==0)) return(false);
  
  if(DrumMode==true){
    _consolePrintf("LoadDrum:%d,%d->%d,%d\n",prg,Note,VarNum,PrgNum);
    }else{
    _consolePrintf("LoadTone:%d,%d->%d,%d\n",var,prg,VarNum,PrgNum);
  }
  
  pPrgMap->ppData[PrgNum]=(TProgramPatch*)safemalloc(sizeof(TProgramPatch));
  if(pPrgMap->ppData[PrgNum]==NULL){
    _consolePrintf("malloc(%d); out of memory!!\n",size);
    return(false);
  }
  
  TProgramPatch *pPrgPatch=pPrgMap->ppData[PrgNum];
  
  u32 patcount;
  u32 patsize[16];
  
  for(u32 patidx=0;patidx<16;patidx++){
    patsize[patidx]=0;
    pPrgPatch->pPrg[patidx]=NULL;
  }
  
  fseek(pPrgMap->FileHandle,pPrgMap->FileOffset+ofs,SEEK_SET);
  
  fread(&patcount,4,1,pPrgMap->FileHandle);
  pPrgPatch->PatchCount=patcount;
  
  fread(&patsize[0],4,patcount,pPrgMap->FileHandle);
  
  for(u32 patidx=0;patidx<patcount;patidx++){
    pPrgPatch->pPrg[patidx]=(TProgram*)safemalloc(patsize[patidx]);
    if(pPrgPatch->pPrg[patidx]==NULL){
      MemoryOverflowFlag=true;
      _consolePrintf("Patch%d malloc(%d); out of memory!!\n",patidx,patsize[patidx]);
      if(patidx!=0){
        for(u32 idx=0;idx<patidx;idx++){
          free(pPrgPatch->pPrg[idx]); pPrgPatch->pPrg[idx]=NULL;
        }
      }
      free(pPrgMap->ppData[PrgNum]); pPrgMap->ppData[PrgNum]=NULL;
      // ōĂH
      return(false);
    }
    fread(pPrgPatch->pPrg[patidx],1,patsize[patidx],pPrgMap->FileHandle);
  }
  
  for(u32 patidx=0;patidx<patcount;patidx++){
    TProgram *pPrg=pPrgPatch->pPrg[patidx];
    
    if(pPrg->s16Flag==1){
      pPrg->Length/=2;
      pPrg->LoopStart/=2;
      pPrg->LoopEnd/=2;
    }
    
    pPrg->VibSweep=pPrg->VibSweep/120;
  
/*
    _consolePrintf("EnvRate=");
    for(int idx=0;idx<6;idx++){
      _consolePrintf("%02x,",pPrg->EnvRate[idx]);
    }
    _consolePrintf("\n");
    
    _consolePrintf("EnvOfs =");
    for(int idx=0;idx<6;idx++){
      _consolePrintf("%02x,",pPrg->EnvOfs[idx]);
    }
    _consolePrintf("\n");
*/
    
    for(int idx=0;idx<6;idx++){
      pPrg->EnvRate[idx]*=96;
    }
  }
  
  return(true);
}

static void PCH_ChInit(u32 ch)
{
  TPCH *_pPCH=&pPCH[ch];
  
  u32 *ptr=(u32*)_pPCH;
  
  for(u32 idx=sizeof(TPCH)/4;idx!=0;idx--){
    *ptr++=0;
  }
  
  _pPCH->Track=(u32)-1;
  _pPCH->EnvState=EES_None;
  
/*
  _pPCH->Enabled=false;
  _pPCH->DrumMode=false;
  
  _pPCH->Track=(u32)-1;
  _pPCH->Note=0;
  _pPCH->Vol=0;
  _pPCH->Exp=0;
  _pPCH->Vel=0;
  _pPCH->Reverb=0;
  _pPCH->Freq=0;
  
  _pPCH->OnClock=0;
  _pPCH->OffClock=0;
  
  _pPCH->GT=0;
  
  _pPCH->Pedal=false;
  
  _pPCH->VibEnable=false;
  _pPCH->VibSweepAdd=0;
  _pPCH->VibSweepCur=0;
  _pPCH->VibAdd=0;
  _pPCH->VibCur=0;
  _pPCH->VibDepth=0;
  _pPCH->VibPhase=0;
  
  _pPCH->ModEnable=false;
  _pPCH->ModAdd=0;
  _pPCH->ModCur=0;
  _pPCH->ModDepth=0;
  _pPCH->ModPhase=0;
  
  _pPCH->EnvState=EES_None;
  _pPCH->EnvSpeed=0;
  _pPCH->EnvCurLevel=0;
  _pPCH->EnvEndLevel=0;
  _pPCH->EnvRelLevel=0;
  
  _pPCH->pPrg=NULL;
  _pPCH->PrgPos=0;
  _pPCH->PrgPanpot=0;
  _pPCH->PrgMstVol=0;
  _pPCH->PrgVol=0;
  _pPCH->PrgFraction=0;
  
  _pPCH->FreqAddFix16=0;
  _pPCH->FreqCurFix16=0;
  
  _pPCH->CurSampleData=0;
  _pPCH->LastSampleData=0;
*/
}

bool PCH_Init(u32 _SampleRate,u32 MaxChannelCount,u32 _GenVolume)
{
  SampleRate=_SampleRate;
  GenVolume=_GenVolume;
  
  PCHCount=MaxChannelCount;
  if(PCHCountMax<PCHCount) PCHCount=PCHCountMax;
  
//  pPCH=(TPCH*)safemalloc(PCHCount*sizeof(TPCH));
//  if(pPCH==NULL) return(false);
  
  MemSet32CPU(0,&DTCM_PCH,PCHCountMax*sizeof(TPCH));
  pPCH=DTCM_PCH;
  
  for(u32 ch=0;ch<PCHCount;ch++){
    PCH_ChInit(ch);
  }
  
  for(int idx=0;idx<128;idx++){
    s32 Note=idx-1;
    
    Note-=68; // NoteA4
    
    u32 basefreq=440;
    
    while(Note<0){
      basefreq/=2;
      Note+=12;
    }
    
    while(12<=Note){
      basefreq*=2;
      Note-=12;
    }
    
    Note2FreqTableFix16[idx]=(u32)(basefreq*(0x10000+(u32)powtbl[(Note*POWRANGE)]));
  }
  
  return(true);
}

void PCH_Free(void)
{
  SampleRate=0;
  
/*
  if(pPCH!=NULL){
    safefree(pPCH); pPCH=NULL;
  }
*/
}

void PCH_AllSoundOff(void)
{
  for(u32 ch=0;ch<PCHCount;ch++){
    PCH_ChInit(ch);
  }
}

void PCH_AllNoteOff(u32 trk)
{
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    if(_pPCH->Track==trk) _pPCH->Enabled=false;
  }
}

void PCH_NextClock(void)
{
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    if(_pPCH->Enabled==true){
      _pPCH->OnClock++;
      if(_pPCH->DrumMode==true){
        if((120*1.5)==_pPCH->OnClock){
          if(_pPCH->EnvState!=EES_Release5){
            _pPCH->EnvRelLevel=_pPCH->EnvCurLevel;
            _pPCH->EnvState=EES_Release5;
            _pPCH->EnvSpeed=-0x100;//_pPCH->pPrg->EnvRate[5];
            _pPCH->EnvEndLevel=0;
          }
        }
      }
      }else{
      if(_pPCH->Pedal==true){
        }else{
        _pPCH->OffClock++;
        if((120*2)==_pPCH->OffClock) PCH_ChInit(ch); // exclusive stop when 2sec over.
      }
    }
  }
  
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    if(_pPCH->VibEnable==true){
      _pPCH->VibCur+=_pPCH->VibAdd;
      while((120*0x10000)<=_pPCH->VibCur){
        _pPCH->VibCur-=120*0x10000;
      }
      _pPCH->VibPhase=sintbl[_pPCH->VibCur>>16]*_pPCH->VibDepth/0x10000;
      if(_pPCH->VibSweepCur!=0x10000){
        _pPCH->VibSweepCur+=_pPCH->VibSweepAdd;
        if(_pPCH->VibSweepCur<0x10000){
          s32 Factor=_pPCH->VibSweepCur>>8;
          _pPCH->VibPhase=(_pPCH->VibPhase*Factor)/0x100;
          }else{
          _pPCH->VibSweepCur=0x10000;
        }
      }
    }
  }
  
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    if(_pPCH->ModEnable==true){
      _pPCH->ModCur+=_pPCH->ModAdd;
      while((120*0x10000)<=_pPCH->ModCur){
        _pPCH->ModCur-=120*0x10000;
      }
      _pPCH->ModPhase=sintbl[_pPCH->ModCur>>16]*_pPCH->ModDepth/0x10000;
    }
  }
  
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    switch(_pPCH->EnvState){
      case EES_None: {
      } break;
      case EES_Attack: {
        bool next=false;
//        _consolePrintf("a%x\n",_pPCH->EnvCurLevel);
        
        if(_pPCH->EnvSpeed==0){
          next=true;
          }else{
          int nextlevel=(int)_pPCH->EnvCurLevel+_pPCH->EnvSpeed;
          
          if((_pPCH->EnvSpeed<0)&&(nextlevel<(int)_pPCH->EnvEndLevel)) next=true;
          if((0<_pPCH->EnvSpeed)&&((int)_pPCH->EnvEndLevel<nextlevel)) next=true;
          
          _pPCH->EnvCurLevel=(u32)nextlevel;
        }
        if(next==true){
          _pPCH->EnvState=EES_Decay;
          _pPCH->EnvCurLevel=_pPCH->EnvEndLevel;
          _pPCH->EnvSpeed=_pPCH->pPrg->EnvRate[1];
          _pPCH->EnvEndLevel=_pPCH->pPrg->EnvOfs[1];
          if(_pPCH->EnvEndLevel<_pPCH->EnvCurLevel) _pPCH->EnvSpeed=-_pPCH->EnvSpeed;
        }
        _pPCH->PrgVol=_pPCH->PrgMstVol*_pPCH->EnvCurLevel/0x10000;
      } break;
      case EES_Decay: {
        bool next=false;
//        _consolePrintf("d%x\n",_pPCH->EnvCurLevel);
        
        if(_pPCH->EnvSpeed==0){
          next=true;
          }else{
          int nextlevel=(int)_pPCH->EnvCurLevel+_pPCH->EnvSpeed;
          
          if((_pPCH->EnvSpeed<0)&&(nextlevel<(int)_pPCH->EnvEndLevel)) next=true;
          if((0<_pPCH->EnvSpeed)&&((int)_pPCH->EnvEndLevel<nextlevel)) next=true;
          
          _pPCH->EnvCurLevel=(u32)nextlevel;
        }
        if(next==true){
          if(true){ // if(_pPCH->DrumMode==false){
            _pPCH->EnvState=EES_Sustain;
            _pPCH->EnvCurLevel=_pPCH->EnvEndLevel;
            _pPCH->EnvSpeed=_pPCH->pPrg->EnvRate[2];
            _pPCH->EnvEndLevel=_pPCH->pPrg->EnvOfs[2];
            if(_pPCH->EnvEndLevel<_pPCH->EnvCurLevel) _pPCH->EnvSpeed=-_pPCH->EnvSpeed;
            }else{
            _pPCH->EnvRelLevel=_pPCH->EnvCurLevel;
            _pPCH->EnvState=EES_Release3;
            _pPCH->EnvSpeed=-_pPCH->pPrg->EnvRate[3];
            _pPCH->EnvEndLevel=(_pPCH->pPrg->EnvOfs[3]*_pPCH->EnvRelLevel)>>16;
          }
        }
        _pPCH->PrgVol=_pPCH->PrgMstVol*_pPCH->EnvCurLevel/0x10000;
      } break;
      case EES_Sustain: {
//        _consolePrintf("s%x\n",_pPCH->EnvCurLevel);
        bool next=false;
        
        if(_pPCH->EnvSpeed==0){
          next=true;
          }else{
          int nextlevel=(int)_pPCH->EnvCurLevel+_pPCH->EnvSpeed;
          
          if((_pPCH->EnvSpeed<0)&&(nextlevel<(int)_pPCH->EnvEndLevel)) next=true;
          if((0<_pPCH->EnvSpeed)&&((int)_pPCH->EnvEndLevel<nextlevel)) next=true;
          
          _pPCH->EnvCurLevel=(u32)nextlevel;
        }
        if(next==true){
          _pPCH->EnvState=EES_None;
          _pPCH->EnvCurLevel=_pPCH->EnvEndLevel;
          _pPCH->EnvSpeed=0;
          _pPCH->EnvEndLevel=0;
        }
        _pPCH->PrgVol=_pPCH->PrgMstVol*_pPCH->EnvCurLevel/0x10000;
      } break;
      case EES_Release3: {
        bool next=false;
//        _consolePrintf("d%x\n",_pPCH->EnvCurLevel);
        
        if(_pPCH->EnvSpeed==0){
          next=true;
          }else{
          int nextlevel=(int)_pPCH->EnvCurLevel+_pPCH->EnvSpeed;
          
          if(nextlevel<(int)_pPCH->EnvEndLevel) next=true;
          
          _pPCH->EnvCurLevel=(u32)nextlevel;
        }
        if(next==true){
          _pPCH->EnvState=EES_Release4;
          _pPCH->EnvCurLevel=_pPCH->EnvEndLevel;
          _pPCH->EnvSpeed=-_pPCH->pPrg->EnvRate[4];
          _pPCH->EnvEndLevel=(_pPCH->pPrg->EnvOfs[4]*_pPCH->EnvRelLevel)>>16;
        }
        _pPCH->PrgVol=_pPCH->PrgMstVol*_pPCH->EnvCurLevel/0x10000;
      } break;
      case EES_Release4: {
        bool next=false;
//        _consolePrintf("d%x\n",_pPCH->EnvCurLevel);
        
        if(_pPCH->EnvSpeed==0){
          next=true;
          }else{
          int nextlevel=(int)_pPCH->EnvCurLevel+_pPCH->EnvSpeed;
          
          if(nextlevel<(int)_pPCH->EnvEndLevel) next=true;
          
          _pPCH->EnvCurLevel=(u32)nextlevel;
        }
        if(next==true){
          _pPCH->EnvState=EES_Release5;
          _pPCH->EnvCurLevel=_pPCH->EnvEndLevel;
          _pPCH->EnvSpeed=-_pPCH->pPrg->EnvRate[5];
//          _pPCH->EnvEndLevel=(_pPCH->pPrg->EnvOfs[5]*_pPCH->EnvRelLevel)>>16;
          _pPCH->EnvEndLevel=0;
        }
        _pPCH->PrgVol=_pPCH->PrgMstVol*_pPCH->EnvCurLevel/0x10000;
      } break;
      case EES_Release5: {
        bool next=false;
//        _consolePrintf("r%x\n",_pPCH->EnvCurLevel);
        
        if(_pPCH->EnvSpeed==0){
          next=true;
          }else{
          int nextlevel=(int)_pPCH->EnvCurLevel+_pPCH->EnvSpeed;
          
          if(nextlevel<(int)_pPCH->EnvEndLevel) next=true;
          
          _pPCH->EnvCurLevel=(u32)nextlevel;
        }
        if(next==true){
          PCH_ChInit(ch);
          }else{
          _pPCH->PrgVol=_pPCH->PrgMstVol*_pPCH->EnvCurLevel/0x10000;
        }
      } break;
    }
  }
  
}

void PCH_ChangeVolume(u32 trk,u32 v)
{
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    if((_pPCH->Enabled==true)&&(_pPCH->Track==trk)){
      _pPCH->Vol=v;
      
      u32 vol;
      
      vol=(_pPCH->Vol*_pPCH->Exp*_pPCH->Vel*_pPCH->MasterVolume)/(128*128*128*128/128);
      vol=vol*GenVolume/128;
      _pPCH->PrgMstVol=vol;
      if(_pPCH->pPrg->EnvelopeFlag==false){
        _pPCH->PrgVol=vol;
        }else{
        _pPCH->PrgVol=vol*_pPCH->EnvCurLevel/0x10000;
      }
    }
  }
}

void PCH_ChangeExpression(u32 trk,u32 e)
{
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    if((_pPCH->Enabled==true)&&(_pPCH->Track==trk)){
      _pPCH->Exp=e;
      
      u32 vol;
      
      vol=(_pPCH->Vol*_pPCH->Exp*_pPCH->Vel*_pPCH->MasterVolume)/(128*128*128*128/128);
      vol=vol*GenVolume/128;
      _pPCH->PrgMstVol=vol;
      if(_pPCH->pPrg->EnvelopeFlag==false){
        _pPCH->PrgVol=vol;
        }else{
        _pPCH->PrgVol=vol*_pPCH->EnvCurLevel/0x10000;
      }
    }
  }
}

void PCH_ChangePitchBend(u32 trk,s32 Pitch)
{
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    if((_pPCH->Enabled==true)&&(_pPCH->Track==trk)){
      TProgram *pPrg=_pPCH->pPrg;
      s32 Note=_pPCH->Note;
      
      Note-=1;
      
      Note-=68; // NoteA4
      
      while(Pitch<0){
        Note--;
        Pitch+=POWRANGE;
      }
      
      while(POWRANGE<=Pitch){
        Note++;
        Pitch-=POWRANGE;
      }
      
      u32 basefreq=440;
      
      while(Note<0){
        basefreq/=2;
        Note+=12;
      }
      
      while(12<=Note){
        basefreq*=2;
        Note-=12;
      }
      
      u64 f=basefreq*(0x10000+(u32)powtbl[(Note*POWRANGE)+Pitch]);
      f=(f*pPrg->SampleRate)*0x10000/pPrg->RootFreq;
      _pPCH->FreqAddFix16=(u32)(f/SampleRate);
    }
  }
}

void PCH_ChangePanpot(u32 trk,u32 p)
{
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    if((_pPCH->Enabled==true)&&(_pPCH->Track==trk)){
      _pPCH->PrgPanpot=p;
    }
  }
}

void PCH_ChangeModLevel(u32 trk,u32 ModLevel)
{
  if(ModLevel==0){
    for(u32 ch=0;ch<PCHCount;ch++){
      TPCH *_pPCH=&pPCH[ch];
      if((_pPCH->Enabled==true)&&(_pPCH->Track==trk)){
        _pPCH->ModEnable=false;
      }
    }
    }else{
    for(u32 ch=0;ch<PCHCount;ch++){
      TPCH *_pPCH=&pPCH[ch];
      if((_pPCH->Enabled==true)&&(_pPCH->Track==trk)){
        _pPCH->ModEnable=true;
        _pPCH->ModDepth=ModLevel*10; // (128*10)/0x10000 per
        _pPCH->ModDepth=(int)(_pPCH->FreqAddFix16*_pPCH->ModDepth/0x10000);
      }
    }
  }
}

static int PCH_FindEqualChannel(u32 trk,u32 Note)
{
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    if(_pPCH->Enabled==true){
      if((_pPCH->Track==trk)&&(_pPCH->Note==Note)){
        return(ch);
      }
    }
  }
  
  return(-1);
}

static int PCH_FindDisableChannel(void)
{
  bool FindedDisabledCh=false;
  
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    if(_pPCH->Enabled==false){
      FindedDisabledCh=true;
    }
  }
  
  if(FindedDisabledCh==true){
    u32 MaxClock=0;
    for(u32 ch=0;ch<PCHCount;ch++){
      TPCH *_pPCH=&pPCH[ch];
      if(_pPCH->Enabled==false){
        if(MaxClock<_pPCH->OffClock) MaxClock=_pPCH->OffClock;
      }
    }
    for(u32 ch=0;ch<PCHCount;ch++){
      TPCH *_pPCH=&pPCH[ch];
      if(_pPCH->Enabled==false){
        if(MaxClock==_pPCH->OffClock) return(ch);
      }
    }
    return(-1);
    
    }else{
    
    u32 MaxClock=0;
    for(u32 ch=0;ch<PCHCount;ch++){
      TPCH *_pPCH=&pPCH[ch];
      if(_pPCH->Enabled==true){
        if(MaxClock<_pPCH->OnClock) MaxClock=_pPCH->OnClock;
      }
    }
    for(u32 ch=0;ch<PCHCount;ch++){
      TPCH *_pPCH=&pPCH[ch];
      if(_pPCH->Enabled==true){
        if(MaxClock==_pPCH->OnClock) return(ch);
      }
    }
    return(-1);
  }
}

void PCH_NoteOn(u32 trk,u32 GT,s32 Note,s32 Pitch,u32 Vol,u32 Exp,u32 Vel,u32 var,u32 prg,u32 panpot,u32 reverb,bool DrumMode,u32 ModLevel)
{
//  if(GT!=0) GT++;
  
  {
    int EqualCh=PCH_FindEqualChannel(trk,Note);
    if(EqualCh!=-1){
      TPCH *_pPCH=&pPCH[EqualCh];
      _pPCH->GT=GT;
      return;
    }
  }
  
  TVariationMap *pVarMap;
  u32 VarNum,PrgNum;
  
  if(DrumMode==true){
    pVarMap=&VarMap_Drum;
    VarNum=prg;
    PrgNum=Note;
    }else{
    pVarMap=&VarMap_Tone;
    VarNum=var;
    PrgNum=prg;
  }
  
  TProgram *pPrg;
  
  {
    TProgramPatch *pPrgPatch=GetProgramPatchPtr(pVarMap,VarNum,PrgNum,DrumMode);
    if(pPrgPatch==NULL) return;
    pPrg=GetProgramFromPatchPtr(pPrgPatch,Note,DrumMode);
    if(pPrg==NULL) return;
  }
  
  if(DrumMode==true) reverb=0;
  
  int ch=PCH_FindDisableChannel();
  
  if(ch==-1){
    _consolePrintf("PCHON %d,%d Error ChannelEmpty.\n",trk,Note);
    return;
  }
  
  PCH_ChInit(ch);
  
  TPCH *_pPCH=&pPCH[ch];
  
  _pPCH->Enabled=true;
  _pPCH->GT=GT;
  _pPCH->DrumMode=DrumMode;
  _pPCH->Track=trk;
  _pPCH->Note=Note;
  _pPCH->Vol=Vol;
  _pPCH->Exp=Exp;
  _pPCH->Vel=Vel;
  _pPCH->Reverb=reverb;
  _pPCH->MasterVolume=pPrg->MasterVolume;
  _pPCH->pPrg=pPrg;
  _pPCH->PrgPos=0;
  
  pPrg->VibRatio=0;
  
  if((pPrg->VibRatio==0)||(pPrg->VibDepth==0)){
    _pPCH->VibEnable=false;
    }else{
    _pPCH->VibEnable=true;
    if(pPrg->VibSweep==0){
      _pPCH->VibSweepAdd=0;
      _pPCH->VibSweepCur=0x10000;
      }else{
      _pPCH->VibSweepAdd=pPrg->VibSweep;
      _pPCH->VibSweepCur=0;
    }
    _pPCH->VibAdd=pPrg->VibRatio;
    _pPCH->VibCur=0;
    _pPCH->VibDepth=0;
    _pPCH->VibPhase=0;
  }
  
  if(ModLevel==0){
    _pPCH->ModEnable=false;
    }else{
    _pPCH->ModEnable=true;
  }
  _pPCH->ModAdd=0x10000*8; // 1/8sec
  _pPCH->ModCur=0;
  _pPCH->ModDepth=ModLevel*8; // (128*8)/0x10000 per
  _pPCH->ModPhase=0;
  
  if(pPrg->EnvelopeFlag==false){
    _pPCH->EnvState=EES_None;
    }else{
    _pPCH->EnvState=EES_Attack;
    _pPCH->EnvSpeed=pPrg->EnvRate[0];
    _pPCH->EnvCurLevel=0;
    _pPCH->EnvEndLevel=pPrg->EnvOfs[0];
  }
  
  {
    u32 vol;
    
    vol=(_pPCH->Vol*_pPCH->Exp*_pPCH->Vel*_pPCH->MasterVolume)/(128*128*128*128/128);
    vol=vol*GenVolume/128;
    _pPCH->PrgMstVol=vol;
    if(pPrg->EnvelopeFlag==false){
      _pPCH->PrgVol=vol;
      }else{
      _pPCH->PrgVol=vol*_pPCH->EnvCurLevel/0x10000;
    }
    _pPCH->PrgFraction=((s32)pPrg->FractionStart)-((s32)pPrg->FractionEnd);
    _pPCH->PrgFraction<<=12;
  }
  
  if(DrumMode==false){
    _pPCH->PrgPanpot=panpot;
    }else{
    _pPCH->PrgPanpot=pPrg->MasterPanpot;
    _pPCH->Note=pPrg->Note;
    Note=_pPCH->Note;
  }
  
  if(_pPCH->Note==0){
    _pPCH->FreqAddFix16=(u32)((float)pPrg->SampleRate/SampleRate*0x10000);
    _pPCH->FreqCurFix16=0;
    }else{
    Note-=1;
    
    Note-=68; // NoteA4
    
    while(Pitch<0){
      Note--;
      Pitch+=POWRANGE;
    }
    
    while(POWRANGE<=Pitch){
      Note++;
      Pitch-=POWRANGE;
    }
    
    u32 basefreq=440;
    
    while(Note<0){
      basefreq/=2;
      Note+=12;
    }
    
    while(12<=Note){
      basefreq*=2;
      Note-=12;
    }
    
    u64 f=basefreq*(0x10000+(u32)powtbl[(Note*POWRANGE)+Pitch]);
    f=(f*pPrg->SampleRate)*0x10000/pPrg->RootFreq;
    _pPCH->FreqAddFix16=(u32)(f/SampleRate);
    
    _pPCH->PrgPanpot=panpot;
    _pPCH->FreqCurFix16=0;
    
  }
  
  if(_pPCH->VibEnable==true){
    _pPCH->VibDepth=(int)(_pPCH->FreqAddFix16*pPrg->VibDepth/0x10000);
  }
  
  if(_pPCH->ModEnable==true){
    _pPCH->ModDepth=(int)(_pPCH->FreqAddFix16*_pPCH->ModDepth/0x10000);
  }
}

void PCH_NoteOff(u32 trk,u32 Note,bool DrumMode,bool Pedal)
{
  if(DrumMode==true) return;
  
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    if(_pPCH->Enabled==true){
      if((_pPCH->Track==trk)&&(_pPCH->Note==Note)){
        if(Pedal==true){
          _pPCH->Pedal=true;
          }else{
          _pPCH->Enabled=false;
          if((_pPCH->pPrg->EnvelopeFlag==false)||(_pPCH->EnvCurLevel==0)){
            PCH_ChInit(ch);
            }else{
            _pPCH->EnvRelLevel=_pPCH->EnvCurLevel;
            _pPCH->EnvState=EES_Release3;
            _pPCH->EnvSpeed=-_pPCH->pPrg->EnvRate[3];
            _pPCH->EnvEndLevel=(_pPCH->pPrg->EnvOfs[3]*_pPCH->EnvRelLevel)>>16;
          }
        }
      }
    }
  }
}

void PCH_PedalOff(u32 trk)
{
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    if(_pPCH->Enabled==true){
      if(_pPCH->Track==trk){
        if(_pPCH->Pedal==true){
          _pPCH->Pedal=false;
          _pPCH->Enabled=false;
          if((_pPCH->pPrg->EnvelopeFlag==false)||(_pPCH->EnvCurLevel==0)){
            PCH_ChInit(ch);
            }else{
            _pPCH->EnvRelLevel=_pPCH->EnvCurLevel;
            _pPCH->EnvState=EES_Release3;
            _pPCH->EnvSpeed=-_pPCH->pPrg->EnvRate[3];
            _pPCH->EnvEndLevel=(_pPCH->pPrg->EnvOfs[3]*_pPCH->EnvRelLevel)>>16;
          }
        }
      }
    }
  }
}

static void PCH_Render_Loop(int ch,u32 FreqAddFix16,s16 *lbuf,s16 *rbuf,int SampleCount)
{
  TPCH *_pPCH=&pPCH[ch];
  TProgram *pPrg=_pPCH->pPrg;
  
  u32 PrgPos=_pPCH->PrgPos;
  u32 FreqCurFix16=_pPCH->FreqCurFix16;
  
  s32 CurSampleData=_pPCH->CurSampleData;
  s32 LastSampleData=_pPCH->LastSampleData;
  
  s32 lvol=_pPCH->PrgVol*_pPCH->PrgPanpot/0x80;
  s32 rvol=_pPCH->PrgVol*(0x80-_pPCH->PrgPanpot)/0x80;
  
  u32 EndPoint;
  
  if(pPrg->LoopFlag==false){
    EndPoint=pPrg->Length;
    }else{
    EndPoint=pPrg->LoopEnd;
  }
  
  if(pPrg->s16Flag==false){
    s8 *Data8=(s8*)&pPrg->Data[0];
    for(u32 idx=SampleCount;idx!=0;idx--){
      s32 sample;
      
      u32 samplediv=PrgPos&0xffff;
      sample=LastSampleData*(0x10000-samplediv); // << 8
      sample+=CurSampleData*samplediv; // << 8
      sample>>=16;
      
      *lbuf+=(s16)(sample*lvol); // >> 8
      *rbuf+=(s16)(sample*rvol); // >> 8
      lbuf++;
      rbuf++;
      
      FreqCurFix16+=FreqAddFix16;
      
      if(0x10000<=FreqCurFix16){
        while(0x10000<=FreqCurFix16){
          FreqCurFix16-=0x10000;
          PrgPos++;
          if(PrgPos==EndPoint){
            if(pPrg->LoopFlag==false){
              PCH_ChInit(ch);
              return;
              }else{
              PrgPos=pPrg->LoopStart;
              s32 Adj=_pPCH->PrgFraction;
              if(Adj<0){
                PrgPos--;
                Adj+=0x10<<12;
              }
              FreqCurFix16+=(u32)Adj;
            }
          }
        }
        LastSampleData=CurSampleData;
        CurSampleData=Data8[PrgPos];
      }
    }
    }else{
    s16 *Data16=(s16*)&pPrg->Data[0];
    for(u32 idx=SampleCount;idx!=0;idx--){
      s32 sample;
      
      u32 samplediv=PrgPos&0xffff;
      sample=LastSampleData*(0x10000-samplediv);
      sample+=CurSampleData*samplediv;
      sample>>=16;
      
      *lbuf+=(s16)(sample*lvol/0x100);
      *rbuf+=(s16)(sample*rvol/0x100);
      lbuf++;
      rbuf++;
      
      FreqCurFix16+=FreqAddFix16;
      
      if(0x10000<=FreqCurFix16){
        while(0x10000<=FreqCurFix16){
          FreqCurFix16-=0x10000;
          PrgPos++;
          if(PrgPos==EndPoint){
            if(pPrg->LoopFlag==false){
              PCH_ChInit(ch);
              return;
              }else{
              PrgPos=pPrg->LoopStart;
              s32 Adj=_pPCH->PrgFraction;
              if(Adj<0){
                PrgPos--;
                Adj+=0x10<<12;
              }
              FreqCurFix16+=(u32)Adj;
            }
          }
        }
        LastSampleData=CurSampleData;
        CurSampleData=Data16[PrgPos];
      }
    }
  }
  
  _pPCH->CurSampleData=CurSampleData;
  _pPCH->LastSampleData=LastSampleData;
  
  _pPCH->FreqCurFix16=FreqCurFix16;
  _pPCH->PrgPos=PrgPos;
}

static void PCH_Render_Bulk(int ch,u32 FreqAddFix16,s16 *lbuf,s16 *rbuf,int SampleCount)
{
  TPCH *_pPCH=&pPCH[ch];
  TProgram *pPrg=_pPCH->pPrg;
  
  u32 PrgPos=_pPCH->PrgPos;
  u32 FreqCurFix16=_pPCH->FreqCurFix16;
  
  s32 CurSampleData=_pPCH->CurSampleData;
  s32 LastSampleData=_pPCH->LastSampleData;
  
  s32 lvol=_pPCH->PrgVol*_pPCH->PrgPanpot/0x80;
  s32 rvol=_pPCH->PrgVol*(0x80-_pPCH->PrgPanpot)/0x80;
  
  if(pPrg->s16Flag==false){
    s8 *Data8=(s8*)&pPrg->Data[0];
    for(u32 idx=SampleCount;idx!=0;idx--){
      s32 sample;
      
      u32 samplediv=PrgPos&0xffff;
      sample=LastSampleData*(0x10000-samplediv); // << 8
      sample+=CurSampleData*samplediv; // << 8
      sample>>=16;
      
      *lbuf+=(s16)(sample*lvol); // >> 8
      *rbuf+=(s16)(sample*rvol); // >> 8
      lbuf++;
      rbuf++;
      
      FreqCurFix16+=FreqAddFix16;
      
      if(0x10000<=FreqCurFix16){
        PrgPos+=FreqCurFix16>>16;
        FreqCurFix16&=0xffff;
//        LastSampleData=Data8[PrgPos-1];
//        CurSampleData=Data8[PrgPos];
        LastSampleData=CurSampleData;
        CurSampleData=Data8[PrgPos];
      }
    }
    }else{
    s16 *Data16=(s16*)&pPrg->Data[0];
    for(u32 idx=SampleCount;idx!=0;idx--){
      s32 sample;
      
      s32 samplediv=PrgPos&0xffff;
      sample=LastSampleData*(0x10000-samplediv);
      sample+=CurSampleData*samplediv;
      sample>>=16;
      
      *lbuf+=(s16)(sample*lvol/0x100);
      *rbuf+=(s16)(sample*rvol/0x100);
      lbuf++;
      rbuf++;
      
      FreqCurFix16+=FreqAddFix16;
      
      if(0x10000<=FreqCurFix16){
        PrgPos+=FreqCurFix16>>16;
        FreqCurFix16&=0xffff;
//        LastSampleData=Data16[PrgPos-1];
//        CurSampleData=Data16[PrgPos];
        LastSampleData=CurSampleData;
        CurSampleData=Data16[PrgPos];
      }
    }
  }
  
  _pPCH->CurSampleData=CurSampleData;
  _pPCH->LastSampleData=LastSampleData;
  
  _pPCH->FreqCurFix16=FreqCurFix16;
  _pPCH->PrgPos=PrgPos;
}

void PCH_Render(s16 *lbuf,s16 *rbuf,int SampleCount)
{
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    
    if((_pPCH->pPrg!=NULL)&&(_pPCH->FreqAddFix16!=0)){
      TProgram *pPrg=_pPCH->pPrg;
      
      u32 FreqAddFix16=_pPCH->FreqAddFix16;
      
      if(_pPCH->VibEnable==true) FreqAddFix16=(u32)((int)FreqAddFix16+_pPCH->VibPhase);
      if(_pPCH->ModEnable==true) FreqAddFix16=(u32)((int)FreqAddFix16+_pPCH->ModPhase);
      
      u32 EndPos=_pPCH->PrgPos+((FreqAddFix16*SampleCount)>>16)+1;
      
      bool BulkFlag=true;
      
      if(pPrg->LoopFlag==false){
        if(pPrg->Length<=EndPos) BulkFlag=false;
        }else{
        if(pPrg->LoopEnd<=EndPos) BulkFlag=false;
      }
      
      if(BulkFlag==true){
        PCH_Render_Bulk(ch,FreqAddFix16,lbuf,rbuf,SampleCount);
        }else{
        PCH_Render_Loop(ch,FreqAddFix16,lbuf,rbuf,SampleCount);
      }
    }
  }
  
}

int PCH_GetGlobalReverb(void)
{
  u32 res=0;
  
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    
    if(_pPCH->Enabled==true){
      if(res<_pPCH->Reverb) res=_pPCH->Reverb;
    }
  }
  
  return((int)res);
}

int PCH_GT_GetNearClock(void)
{
  int NearClock=0x7fffffff;
  
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    
    if(_pPCH->Enabled==true){
      if(_pPCH->GT!=0){
        if((s32)_pPCH->GT<NearClock) NearClock=(s32)_pPCH->GT;
      }
    }
  }
  
  if(NearClock==0x7fffffff) NearClock=0;
  
  return(NearClock);
}

void PCH_GT_DecClock(u32 clk)
{
  for(u32 ch=0;ch<PCHCount;ch++){
    TPCH *_pPCH=&pPCH[ch];
    
    if(_pPCH->Enabled==true){
      if(_pPCH->GT!=0){
        _pPCH->GT-=clk;
        if(_pPCH->GT==0){
          PCH_NoteOff(_pPCH->Track,_pPCH->Note,_pPCH->DrumMode,false);
        }
      }
    }
  }
}

