
#include <stdio.h>
#include <math.h>

#include <NDS.h>

#include "_const.h"

#include "plugin.h"
#include "plugin_def.h"

#include "inifile.h"


// CallBack from plugin_dll.cpp

void cbLoadLibrary(void)
{
}

void cbFreeLibrary(void)
{
}

void cbQueryInterfaceLibrary(void)
{
}

// -----------------

typedef struct {
  int w,h;
  bool trans;
  u16 *pbmbuf;
} TBM;

typedef struct {
  int SrcWidth,SrcHeight;
  int DstWidth,DstHeight;
  int NumStart,NumCount;
  bool trans;
  u16 *pbmbuf;
} TFont;

TBM bmDate,bmBG,bmDigiBig,bmDigiSmall,bmAMPM,bmTemperature;
TFont fontDate,fontDigiBig,fontDigiSmall,fontAMPM,fontTemperature;

// ------------------------------------------------------------------------------------

void LoadB15(int fh,u32 ofs,TBM *pBM)
{
  pBM->w=0;
  pBM->h=0;
  pBM->trans=0;
  pBM->pbmbuf=NULL;
  
  if(ofs==0) return;
  
  fseek(fh,ofs,SEEK_SET);
  
  fread(&pBM->w,2,1,fh);
  fread(&pBM->h,2,1,fh);
  fread(&pBM->trans,2,1,fh);
  
  int size=pBM->h*pBM->w;
  
  if(size!=0){
    pBM->pbmbuf=(u16*)safemalloc(size*2);
    fread(&pBM->pbmbuf[0],2,size,fh);
  }
}

void FreeB15(TBM *pBM)
{
  pBM->w=0;
  pBM->h=0;
  pBM->trans=false;
  
  if(pBM->pbmbuf!=NULL){
    safefree(pBM->pbmbuf); pBM->pbmbuf=NULL;
  }
}

void TransBMFull(TBM *pBM,u16 *pbmbuf)
{
  if(pBM->pbmbuf!=NULL){
    MemCopy16DMA3(pBM->pbmbuf,pbmbuf,ScreenHeight*ScreenWidth*2);
    }else{
    TiniClockPlugin *ClockPlugin=&GlobalINI.ClockPlugin;
    MemSet16DMA3(ClockPlugin->BGColor | BIT15,pbmbuf,ScreenHeight*ScreenWidth*2);
  }
}

void LoadFont(TBM *pBM,TFont *pFont,int NumStart,int NumCount,float PitchWidth,float PitchHeight)
{
  pFont->SrcWidth=pBM->w;
  pFont->SrcHeight=pBM->h/NumCount;
  pFont->DstWidth=(int)((float)pFont->SrcWidth/PitchWidth);
  pFont->DstHeight=(int)((float)pFont->SrcHeight/PitchHeight);
  pFont->NumStart=NumStart;
  pFont->NumCount=NumCount;
  pFont->trans=pBM->trans;
  pFont->pbmbuf=pBM->pbmbuf;
  
//  _consolePrintf("fontload:%d,%d %d,%d,%d,%d\n",pFont->NumStart,pFont->NumCount,pFont->SrcWidth,pFont->SrcHeight,pFont->DstWidth,pFont->DstHeight);
}

void FreeFont(TFont *pFont)
{
  pFont->SrcWidth=0;
  pFont->SrcHeight=0;
  pFont->DstWidth=0;
  pFont->DstHeight=0;
  pFont->NumStart=0;
  pFont->NumCount=0;
  pFont->trans=false;
  pFont->pbmbuf=NULL;
}

void DrawFontChar(TFont *pFont,char c,u16 *pbmbuf,int x,int y)
{
  int ic=(int)c;
  
  u32 ScreenOffset=x+(y*ScreenWidth);
  
  pbmbuf=&pbmbuf[ScreenOffset];
  
  u16 *pbgbuf=&bmBG.pbmbuf[ScreenOffset];
  u16 *psrcbuf;
  
  ic-=pFont->NumStart;
  if((ic<0)||(pFont->NumCount<=ic)){
    psrcbuf=NULL;
    }else{
    psrcbuf=&pFont->pbmbuf[ic*pFont->SrcHeight*pFont->SrcWidth];
  }
  
  if(psrcbuf==NULL){
    psrcbuf=bmBG.pbmbuf;
    if(psrcbuf!=NULL){
      psrcbuf=&psrcbuf[ScreenOffset];
      for(u32 y=(u32)pFont->SrcHeight;y!=0;y--){
        MemCopy16DMA3(psrcbuf,pbmbuf,pFont->SrcWidth*2);
        pbmbuf+=ScreenWidth;
        psrcbuf+=ScreenWidth;
      }
      }else{
      TiniClockPlugin *ClockPlugin=&GlobalINI.ClockPlugin;
      for(u32 y=(u32)pFont->SrcHeight;y!=0;y--){
        MemSet16DMA3(ClockPlugin->BGColor | BIT15,pbmbuf,pFont->SrcWidth*2);
        pbmbuf+=ScreenWidth;
      }
    }
    return;
  }
  
  if(pFont->trans==false){
    for(u32 y=(u32)pFont->SrcHeight;y!=0;y--){
      MemCopy16DMA3(psrcbuf,pbmbuf,pFont->SrcWidth*2);
      pbmbuf+=ScreenWidth;
      psrcbuf+=pFont->SrcWidth;
    }
    }else{
    for(u32 y=(u32)pFont->SrcHeight;y!=0;y--){
      for(int x=0;x<pFont->SrcWidth;x++){
        if(psrcbuf[x]==0){
          pbmbuf[x]=pbgbuf[x];
          }else{
          pbmbuf[x]=psrcbuf[x];
        }
      }
      pbmbuf+=ScreenWidth;
      pbgbuf+=ScreenWidth;
      psrcbuf+=pFont->SrcWidth;
    }
  }
}

void DrawFontStr(TFont *pFont,char *pstr,u16 *pbmbuf,int x,int y)
{
  if(pstr==NULL) return;
  
  while(*pstr!=0){
    DrawFontChar(pFont,*pstr,pbmbuf,x,y);
    x+=pFont->DstWidth;
    pstr++;
  }
}

// ------------------------------------------------------------------------------------

// ------------------------------------------------------------------------------------

#define PI (3.141592653582)

// ------------------------------------------------------------------------------------

typedef struct {
  int x,y;
  float rot;
  int len;
  int vx,vy;
  int ex,ey;
} TWire;

#define WireCount (3)

enum EWireID {EWI_Hour=0,EWI_Minuts=1,EWI_Second=2};

static TWire Wire[WireCount],WireLast[WireCount];

static u8 WireMap[ScreenHeight*ScreenWidth];

typedef struct {
  bool Enabled;
  u32 life;
  int lx,ly;
  int x,y;
  int sx,sy;
} TDot;

#define DotCount (128)

static TDot Dot[DotCount];

static u16 BGColor,HourWireColor,MinutsWireColor,SecondWireColor,DotColor;

void ColorInit(void)
{
  TiniClockPlugin *ClockPlugin=&GlobalINI.ClockPlugin;
  
  BGColor=ClockPlugin->BGColor | BIT15;
  HourWireColor=ClockPlugin->HourWireColor | BIT15;
  MinutsWireColor=ClockPlugin->MinutsWireColor | BIT15;
  SecondWireColor=ClockPlugin->SecondWireColor | BIT15;
  DotColor=ClockPlugin->DotColor | BIT15;
}

void wireInit(void)
{
  for(u32 idx=0;idx<WireCount;idx++){
    TWire *pw=&Wire[idx];
    pw->len=0;
  }
}

void wireSet(EWireID EWI,float rot,int len)
{
  const int centx=192/2;
  const int centy=192/2;
  
  TWire *pw=&Wire[(u32)EWI];
  
  pw->x=centx*0x100;
  pw->y=centy*0x100;
  pw->rot=rot-PI/2;
  pw->len=(int)(len*0x100);
  pw->vx=(int)(cos(pw->rot)*0x100);
  pw->vy=(int)(sin(pw->rot)*0x100);
  pw->ex=pw->vx*len/0x100;
  pw->ey=pw->vy*len/0x100;
}

void wiremapClear(void)
{
  MemSet32DMA3(0,&WireMap[0],ScreenHeight*ScreenWidth);
}

void wiremapDraw(u16 *pbmbuf)
{
  u16 _BGColor=BGColor;
  
  u8 *pwm=&WireMap[0];
  
  for(u32 idx=0;idx<WireCount;idx++){
    TWire *pw=&WireLast[idx];
    
    int px=pw->x;
    int py=pw->y;
    int vx=pw->vx;
    int vy=pw->vy;
    
    for(u32 lp=0;lp<(u32)pw->len/0x100;lp++){
      u32 ofs=(px/0x100)+((py/0x100)*ScreenWidth);
      pwm[ofs]=idx+1;
      pbmbuf[ofs]=_BGColor;
      px+=vx;
      py+=vy;
    }
  }
  
  for(u32 idx=0;idx<WireCount;idx++){
    WireLast[idx]=Wire[idx];
  }
  
  for(u32 idx=0;idx<WireCount;idx++){
    u16 col;
    switch((EWireID)idx){
      case EWI_Hour: col=HourWireColor; break;
      case EWI_Minuts: col=MinutsWireColor; break;
      case EWI_Second: col=SecondWireColor; break;
      default: col=RGB15(0,0,0) | BIT15; break;
    }
    
    TWire *pw=&Wire[idx];
    
    int px=pw->x;
    int py=pw->y;
    int vx=pw->vx;
    int vy=pw->vy;
    
    for(u32 lp=0;lp<(u32)pw->len/0x100;lp++){
      u32 ofs=(px/0x100)+((py/0x100)*ScreenWidth);
      pwm[ofs]=idx+1;
      pbmbuf[ofs]=col;
      px+=vx;
      py+=vy;
    }
  }
  
}

void dotInit(void)
{
  for(u32 idx=0;idx<DotCount;idx++){
    Dot[idx].Enabled=false;
  }
}

void dotProc(u16 *pbmbuf,u32 VsyncCount)
{
  u16 _BGColor=BGColor;
  u16 _DotColor=DotColor;
  
  u8 *pwm=&WireMap[0];
  
  for(u32 idx=0;idx<DotCount;idx++){
    TDot *pdot=&Dot[idx];
    if(pdot->Enabled==true){
      pdot->life--;
      pdot->sx=(pdot->sx*127)/128;
      pdot->sy+=2;
      pdot->lx=pdot->x;
      pdot->ly=pdot->y;
      pdot->x+=pdot->sx*VsyncCount;
      pdot->y+=pdot->sy*VsyncCount;
      
      {
        u32 hitchk=0;
        
        {
          int x0=pdot->lx/0x100;
          int x1=pdot->x/0x100;
          int y0=pdot->ly/0x100;
          int y1=pdot->y/0x100;
          
          if(x1<x0){
            int x=x1;
            x1=x0;
            x0=x;
          }
          if(y1<y0){
            int y=y1;
            y1=y0;
            y0=y;
          }
          
          u8 *pwm2=&pwm[x0+(y0*ScreenWidth)];
          
          x1-=x0;
          y1-=y0;
          
          for(int y=0;y<=y1;y++){
            for(int x=0;x<=x1;x++){
              u8 data=pwm2[x];
              if(data!=0) hitchk=data;
            }
            pwm2+=ScreenWidth;
          }
        }
        
        if(hitchk!=0){
          pdot->x=pdot->lx;
          pdot->y=pdot->ly;
          
          float wirerot=Wire[hitchk-1].rot;
          
          float len=sqrt(((float)(pdot->sx*pdot->sx)/0x10000)+((float)(pdot->sy*pdot->sy)/0x10000));
          float rot=atan(((float)pdot->sy)/((float)pdot->sx));
          
          if(pdot->sx<0) rot=rot-PI;
          rot=wirerot+(wirerot-rot);
          len*=0.9;
          len*=0x100;
          
          pdot->sx=(int)(cos(rot)*len);
          pdot->sy=(int)(sin(rot)*len);
          pdot->x+=pdot->sx*VsyncCount;
          pdot->y+=pdot->sy*VsyncCount;
          
          pdot->life=(pdot->life*3)/4;
        }
      }
      
      u32 ofs=(pdot->lx/0x100)+((pdot->ly/0x100)*ScreenWidth);
      pbmbuf[ofs]=_BGColor;
      
      if(pdot->life==0) pdot->Enabled=false;
      if((pdot->x<0)||((ScreenWidth*0x100)<=pdot->x)) pdot->Enabled=false;
      if((pdot->y<0)||(((ScreenHeight-12)*0x100)<=pdot->y)) pdot->Enabled=false;
      
      if(pdot->Enabled==true){
        u32 ofs=(pdot->x/0x100)+((pdot->y/0x100)*ScreenWidth);
        pbmbuf[ofs]=_DotColor;
      }
    }
  }
  
  for(u32 idx=0;idx<DotCount;idx++){
    TDot *pdot=&Dot[idx];
    if(pdot->Enabled==false){
      pdot->Enabled=true;
      pdot->life=256;
      pdot->lx=0*0x100;
      pdot->ly=0*0x100;
      pdot->x=rand()%(ScreenWidth*0x100);
      pdot->y=rand()%0x10;
      pdot->sx=(rand()%0x200)-0x100;
      pdot->sx=-0x100;
      pdot->sy=0x80;
      
      return;
    }
  }
}

// ------------------------------------------------------------------------------------

#define FileCount (6)

enum EFileID {EFI_AMPM=0,EFI_BG=1,EFI_Date=2,EFI_DigiBig=3,EFI_DigiSmall=4,EFI_Temperature=5};

bool Start(int _FileHandle)
{
  _FileHandle=0; // not use FileHandle.
  
  InitINI();
  LoadINI(GetINIData(),GetINISize());
  
  int binfh=GetBINFileHandle();
  
  if(binfh==0){
    _consolePrintf("ClockPlugin:can't found clk_???.bin file.\n");
    ShowLogHalt();
  }
  
  u32 FileOfs[FileCount];
  fread(&FileOfs[0],4,FileCount,binfh);
  
  LoadB15(binfh,FileOfs[EFI_AMPM],&bmAMPM);
  LoadB15(binfh,FileOfs[EFI_BG],&bmBG);
  LoadB15(binfh,FileOfs[EFI_Date],&bmDate);
  LoadB15(binfh,FileOfs[EFI_DigiBig],&bmDigiBig);
  LoadB15(binfh,FileOfs[EFI_DigiSmall],&bmDigiSmall);
  LoadB15(binfh,FileOfs[EFI_Temperature],&bmTemperature);
  
  LoadFont(&bmDate,&fontDate,int(','),14,1,1);
  LoadFont(&bmDigiBig,&fontDigiBig,int('0'),11,1,1);
  LoadFont(&bmDigiSmall,&fontDigiSmall,int('0'),11,1,1);
  LoadFont(&bmAMPM,&fontAMPM,int('0'),2,1,1);
  LoadFont(&bmTemperature,&fontTemperature,int('.'),13,1,1);
  
  ColorInit();
  wireInit();
  dotInit();
  
  return(true);
}

void Free(void)
{
  FreeFont(&fontAMPM);
  FreeFont(&fontDate);
  FreeFont(&fontDigiBig);
  FreeFont(&fontDigiSmall);
  FreeFont(&fontTemperature);
  
  FreeB15(&bmAMPM);
  FreeB15(&bmBG);
  FreeB15(&bmDate);
  FreeB15(&bmDigiBig);
  FreeB15(&bmDigiSmall);
  FreeB15(&bmTemperature);
}

u16 ProcKeys(u16 keys)
{
 return(keys); // ignored keys.
}

static TPlugin_ClockLib_Time LastTime;

void UpdateVsync(u16 *pbmbuf,int VsyncPassedCount,bool RunningFile,bool RequestRefresh,TPlugin_ClockLib_Time Time)
{
  if(pbmbuf==NULL) return;
  if(VsyncPassedCount==0) return;
  
  if(RequestRefresh==true){
    TransBMFull(&bmBG,pbmbuf);
    
    LastTime.Year=(u32)-1;
    LastTime.Month=(u32)-1;
    LastTime.Day=(u32)-1;
    LastTime.Hour=(u32)-1;
    LastTime.Minuts=(u32)-1;
    LastTime.Second=(u32)-1;
    LastTime.SystemRTC24Hour=Time.SystemRTC24Hour;
    LastTime.Temperature12=(u32)-1;
  }
  
  TiniClockPlugin *ClockPlugin=&GlobalINI.ClockPlugin;
  
  bool WireFlag=false;
  
  if(ClockPlugin->HourMinutsShow==true){
    if((LastTime.Hour!=Time.Hour)||(LastTime.Minuts!=Time.Minuts)){
      LastTime.Hour=Time.Hour;
      LastTime.Minuts=Time.Minuts;
      wireSet(EWI_Hour,(((float)Time.Hour/12)+((float)Time.Minuts/60/12))*PI*2,48);
      wireSet(EWI_Minuts,(float)Time.Minuts/60*PI*2,72);
      WireFlag=true;
      if(Time.SystemRTC24Hour==true){
        char str[256];
        snprintf(str,256,ClockPlugin->HourMinutsFormat,Time.Hour,Time.Minuts);
        DrawFontStr(&fontDigiBig,str,pbmbuf,ClockPlugin->HourMinutsOfsX,ClockPlugin->HourMinutsOfsY);
        }else{
        u32 h=Time.Hour;
        bool nowpm=false;
        
        if(12<=h){
          h-=12;
          nowpm=true;
        }
        
        char str[256];
        snprintf(str,256,ClockPlugin->HourMinutsFormat,h,Time.Minuts);
        DrawFontStr(&fontDigiBig,str,pbmbuf,ClockPlugin->HourMinutsOfsX,ClockPlugin->HourMinutsOfsY);
        
        if(ClockPlugin->AMPMShow==true){
          if(nowpm==false){
            DrawFontStr(&fontAMPM,"0",pbmbuf,ClockPlugin->AMPMOfsX,ClockPlugin->AMPMOfsY);
            }else{
            DrawFontStr(&fontAMPM,"1",pbmbuf,ClockPlugin->AMPMOfsX,ClockPlugin->AMPMOfsY);
          }
        }
      }
    }
  }
  
  if(ClockPlugin->SecondShow==true){
    if(LastTime.Second!=Time.Second){
      LastTime.Second=Time.Second;
      wireSet(EWI_Second,(float)Time.Second/60*PI*2,84);
      WireFlag=true;
      char str[256];
      snprintf(str,256,ClockPlugin->SecondFormat,Time.Second);
      DrawFontStr(&fontDigiSmall,str,pbmbuf,ClockPlugin->SecondOfsX,ClockPlugin->SecondOfsY);
    }
  }
  
  if(ClockPlugin->DateShow==true){
    if((LastTime.Year!=Time.Year)||(LastTime.Month!=Time.Month)||(LastTime.Day!=Time.Day)){
      LastTime.Year=Time.Year;
      LastTime.Month=Time.Month;
      LastTime.Day=Time.Day;
      char str[256+1];
      formdt_FormatDate(str,256,Time.Year,Time.Month,Time.Day);
      DrawFontStr(&fontDate,str,pbmbuf,ClockPlugin->DateOfsX,ClockPlugin->DateOfsY);
    }
  }
  
  if(ClockPlugin->TemperatureShow==true){
    if((VsyncPassedCount!=0)&&(VsyncPassedCount<6)){
      if(LastTime.Temperature12!=Time.Temperature12){
        LastTime.Temperature12=Time.Temperature12;
        char str[256];
        snprintf(str,256,ClockPlugin->TemperatureFormat,((float)Time.Temperature12) / 0x1000);
        DrawFontStr(&fontTemperature,str,pbmbuf,ClockPlugin->TemperatureOfsX,ClockPlugin->TemperatureOfsY);
      }
    }
    
  }
  
  if(WireFlag==true){
    wiremapClear();
    wiremapDraw(pbmbuf);
  }
  
  dotProc(pbmbuf,VsyncPassedCount);
}
