
#include <stdio.h>
#include <math.h>

#include <NDS.h>

#include "_const.h"

#include "plugin.h"
#include "plugin_def.h"

#include "inifile.h"


// CallBack from plugin_dll.cpp

void cbLoadLibrary(void)
{
}

void cbFreeLibrary(void)
{
}

void cbQueryInterfaceLibrary(void)
{
}

// -----------------

typedef struct {
  int w,h;
  bool trans;
  u16 *pbmbuf;
} TBM;

typedef struct {
  int SrcWidth,SrcHeight;
  int DstWidth,DstHeight;
  int NumStart,NumCount;
  bool trans;
  u16 *pbmbuf;
} TFont;

TBM bmDate,bmBG,bmDigiBig,bmDigiSmall,bmAMPM,bmTemperature;
TFont fontDate,fontDigiBig,fontDigiSmall,fontAMPM,fontTemperature;

// ------------------------------------------------------------------------------------

void LoadB15(int fh,u32 ofs,TBM *pBM)
{
  pBM->w=0;
  pBM->h=0;
  pBM->trans=0;
  pBM->pbmbuf=NULL;
  
  if(ofs==0) return;
  
  fseek(fh,ofs,SEEK_SET);
  
  fread(&pBM->w,2,1,fh);
  fread(&pBM->h,2,1,fh);
  fread(&pBM->trans,2,1,fh);
  
  int size=pBM->h*pBM->w;
  
  if(size!=0){
    pBM->pbmbuf=(u16*)safemalloc(size*2);
    fread(&pBM->pbmbuf[0],2,size,fh);
  }
}

void FreeB15(TBM *pBM)
{
  pBM->w=0;
  pBM->h=0;
  pBM->trans=false;
  
  if(pBM->pbmbuf!=NULL){
    safefree(pBM->pbmbuf); pBM->pbmbuf=NULL;
  }
}

void TransBMFull(TBM *pBM,u16 *pbmbuf)
{
  if(pBM->pbmbuf!=NULL){
    MemCopy16DMA3(pBM->pbmbuf,pbmbuf,ScreenHeight*ScreenWidth*2);
    }else{
    TiniClockPlugin *ClockPlugin=&GlobalINI.ClockPlugin;
    MemSet16DMA3(ClockPlugin->BGColor | BIT15,pbmbuf,ScreenHeight*ScreenWidth*2);
  }
}

void LoadFont(TBM *pBM,TFont *pFont,int NumStart,int NumCount,float PitchWidth,float PitchHeight)
{
  pFont->SrcWidth=pBM->w;
  pFont->SrcHeight=pBM->h/NumCount;
  pFont->DstWidth=(int)((float)pFont->SrcWidth/PitchWidth);
  pFont->DstHeight=(int)((float)pFont->SrcHeight/PitchHeight);
  pFont->NumStart=NumStart;
  pFont->NumCount=NumCount;
  pFont->trans=pBM->trans;
  pFont->pbmbuf=pBM->pbmbuf;
  
//  _consolePrintf("fontload:%d,%d %d,%d,%d,%d\n",pFont->NumStart,pFont->NumCount,pFont->SrcWidth,pFont->SrcHeight,pFont->DstWidth,pFont->DstHeight);
}

void FreeFont(TFont *pFont)
{
  pFont->SrcWidth=0;
  pFont->SrcHeight=0;
  pFont->DstWidth=0;
  pFont->DstHeight=0;
  pFont->NumStart=0;
  pFont->NumCount=0;
  pFont->trans=false;
  pFont->pbmbuf=NULL;
}

void DrawFontChar(TFont *pFont,char c,u16 *pbmbuf,int x,int y)
{
  int ic=(int)c;
  
  u32 ScreenOffset=x+(y*ScreenWidth);
  
  pbmbuf=&pbmbuf[ScreenOffset];
  
  u16 *pbgbuf=&bmBG.pbmbuf[ScreenOffset];
  u16 *psrcbuf;
  
  ic-=pFont->NumStart;
  if((ic<0)||(pFont->NumCount<=ic)){
    psrcbuf=NULL;
    }else{
    psrcbuf=&pFont->pbmbuf[ic*pFont->SrcHeight*pFont->SrcWidth];
  }
  
  if(psrcbuf==NULL){
    psrcbuf=bmBG.pbmbuf;
    if(psrcbuf!=NULL){
      psrcbuf=&psrcbuf[ScreenOffset];
      for(u32 y=(u32)pFont->SrcHeight;y!=0;y--){
        MemCopy16DMA3(psrcbuf,pbmbuf,pFont->SrcWidth*2);
        pbmbuf+=ScreenWidth;
        psrcbuf+=ScreenWidth;
      }
      }else{
      TiniClockPlugin *ClockPlugin=&GlobalINI.ClockPlugin;
      for(u32 y=(u32)pFont->SrcHeight;y!=0;y--){
        MemSet16DMA3(ClockPlugin->BGColor | BIT15,pbmbuf,pFont->SrcWidth*2);
        pbmbuf+=ScreenWidth;
      }
    }
    return;
  }
  
  if(pFont->trans==false){
    for(u32 y=(u32)pFont->SrcHeight;y!=0;y--){
      MemCopy16DMA3(psrcbuf,pbmbuf,pFont->SrcWidth*2);
      pbmbuf+=ScreenWidth;
      psrcbuf+=pFont->SrcWidth;
    }
    }else{
    for(u32 y=(u32)pFont->SrcHeight;y!=0;y--){
      for(int x=0;x<pFont->SrcWidth;x++){
        if(psrcbuf[x]==0){
          pbmbuf[x]=pbgbuf[x];
          }else{
          pbmbuf[x]=psrcbuf[x];
        }
      }
      pbmbuf+=ScreenWidth;
      pbgbuf+=ScreenWidth;
      psrcbuf+=pFont->SrcWidth;
    }
  }
}

void DrawFontStr(TFont *pFont,char *pstr,u16 *pbmbuf,int x,int y)
{
  if(pstr==NULL) return;
  
  while(*pstr!=0){
    DrawFontChar(pFont,*pstr,pbmbuf,x,y);
    x+=pFont->DstWidth;
    pstr++;
  }
}

// ------------------------------------------------------------------------------------

#include "ledon_b15_bin.h"
#include "ledoff_b15_bin.h"

TBM bmLEDOn,bmLEDOff;

// ------------------------------------------------------------------------------------

#define LEDSIZE (12)

enum ELEDMODE {ELM_Clear=0,ELM_PowOff=1,ELM_PowOn=2};

void LoadMemB15(u16 *data,TBM *pBM)
{
  pBM->w=*data++;
  pBM->h=*data++;
  pBM->trans=*data++;
  pBM->pbmbuf=data;
}

void ledInit(void)
{
  LoadMemB15((u16*)ledon_b15_bin,&bmLEDOn);
  LoadMemB15((u16*)ledoff_b15_bin,&bmLEDOff);
}

void ledDraw(u16 *pbmbuf,int x,int y,ELEDMODE ELM)
{
  u16 *psrcbuf=NULL;
  
  switch(ELM){
    case ELM_Clear: psrcbuf=NULL; break;
    case ELM_PowOff: psrcbuf=bmLEDOff.pbmbuf; break;
    case ELM_PowOn: psrcbuf=bmLEDOn.pbmbuf; break;
  }
  
  u32 ScreenOffset=x+(y*ScreenWidth);
  
  pbmbuf=&pbmbuf[ScreenOffset];
  
  if(psrcbuf!=NULL){
    for(u32 y=LEDSIZE;y!=0;y--){
      MemCopy16DMA3(psrcbuf,pbmbuf,LEDSIZE*2);
      pbmbuf+=ScreenWidth;
      psrcbuf+=LEDSIZE;
    }
    }else{
    TiniClockPlugin *ClockPlugin=&GlobalINI.ClockPlugin;
    for(u32 y=LEDSIZE;y!=0;y--){
      MemSet16DMA3(ClockPlugin->BGColor | BIT15,pbmbuf,LEDSIZE*2);
      pbmbuf+=ScreenWidth;
    }
  }
  
}

#define PI (3.141592653582)

void ledDrawRot(u16 *pbmbuf,float rot,float len,ELEDMODE ELM)
{
  const int centx=192/2;
  const int centy=192/2;
  
  float px,py;
  
  px=centx+(cos(rot-PI/2)*len)-(LEDSIZE/2);
  py=centy+(sin(rot-PI/2)*len)-(LEDSIZE/2);
  
  ledDraw(pbmbuf,(int)px,(int)py,ELM);
}

// ------------------------------------------------------------------------------------

#define FileCount (6)

enum EFileID {EFI_AMPM=0,EFI_BG=1,EFI_Date=2,EFI_DigiBig=3,EFI_DigiSmall=4,EFI_Temperature=5};

bool Start(int _FileHandle)
{
  _FileHandle=0; // not use FileHandle.
  
  InitINI();
  LoadINI(GetINIData(),GetINISize());
  
  int binfh=GetBINFileHandle();
  
  if(binfh==0){
    _consolePrintf("ClockPlugin:can't found clk_???.bin file.\n");
    ShowLogHalt();
  }
  
  u32 FileOfs[FileCount];
  fread(&FileOfs[0],4,FileCount,binfh);
  
  LoadB15(binfh,FileOfs[EFI_AMPM],&bmAMPM);
  LoadB15(binfh,FileOfs[EFI_BG],&bmBG);
  LoadB15(binfh,FileOfs[EFI_Date],&bmDate);
  LoadB15(binfh,FileOfs[EFI_DigiBig],&bmDigiBig);
  LoadB15(binfh,FileOfs[EFI_DigiSmall],&bmDigiSmall);
  LoadB15(binfh,FileOfs[EFI_Temperature],&bmTemperature);
  
  LoadFont(&bmDate,&fontDate,int(','),14,1,1);
  LoadFont(&bmDigiBig,&fontDigiBig,int('0'),11,1,1);
  LoadFont(&bmDigiSmall,&fontDigiSmall,int('0'),11,1,1);
  LoadFont(&bmAMPM,&fontAMPM,int('0'),2,1,1);
  LoadFont(&bmTemperature,&fontTemperature,int('.'),13,1,1);
  
  ledInit();
  
  return(true);
}

void Free(void)
{
  FreeFont(&fontAMPM);
  FreeFont(&fontDate);
  FreeFont(&fontDigiBig);
  FreeFont(&fontDigiSmall);
  FreeFont(&fontTemperature);
  
  FreeB15(&bmAMPM);
  FreeB15(&bmBG);
  FreeB15(&bmDate);
  FreeB15(&bmDigiBig);
  FreeB15(&bmDigiSmall);
  FreeB15(&bmTemperature);
}

u16 ProcKeys(u16 keys)
{
 return(keys); // ignored keys.
}

static TPlugin_ClockLib_Time LastTime;

void UpdateVsync(u16 *pbmbuf,int VsyncPassedCount,bool RunningFile,bool RequestRefresh,TPlugin_ClockLib_Time Time)
{
  if(pbmbuf==NULL) return;
  if(VsyncPassedCount==0) return;
  
  if(RequestRefresh==true){
    TransBMFull(&bmBG,pbmbuf);
    
    LastTime.Year=(u32)-1;
    LastTime.Month=(u32)-1;
    LastTime.Day=(u32)-1;
    LastTime.Hour=(u32)-1;
    LastTime.Minuts=(u32)-1;
    LastTime.Second=(u32)-1;
    LastTime.SystemRTC24Hour=Time.SystemRTC24Hour;
    LastTime.Temperature12=(u32)-1;
    
    ledDrawRot(pbmbuf,(float)0,0,ELM_PowOff);
  }
  
  TiniClockPlugin *ClockPlugin=&GlobalINI.ClockPlugin;
  
  if(ClockPlugin->HourMinutsShow==true){
    if((LastTime.Hour!=Time.Hour)||(LastTime.Minuts!=Time.Minuts)){
      if(LastTime.Hour!=(u32)-1) ledDrawRot(pbmbuf,(((float)Time.Hour/12)+((float)Time.Minuts/60/12))*PI*2,48,ELM_Clear);
      if(LastTime.Minuts!=(u32)-1) ledDrawRot(pbmbuf,(float)LastTime.Minuts/60*PI*2,72,ELM_Clear);
      ledDrawRot(pbmbuf,(((float)Time.Hour/12)+((float)Time.Minuts/60/12))*PI*2,48,ELM_PowOn);
      ledDrawRot(pbmbuf,(float)Time.Minuts/60*PI*2,72,ELM_PowOn);
      LastTime.Hour=Time.Hour;
      LastTime.Minuts=Time.Minuts;
      if(Time.SystemRTC24Hour==true){
        char str[256];
        snprintf(str,256,ClockPlugin->HourMinutsFormat,Time.Hour,Time.Minuts);
        DrawFontStr(&fontDigiBig,str,pbmbuf,ClockPlugin->HourMinutsOfsX,ClockPlugin->HourMinutsOfsY);
        }else{
        u32 h=Time.Hour;
        bool nowpm=false;
        
        if(12<=h){
          h-=12;
          nowpm=true;
        }
        
        char str[256];
        snprintf(str,256,ClockPlugin->HourMinutsFormat,h,Time.Minuts);
        DrawFontStr(&fontDigiBig,str,pbmbuf,ClockPlugin->HourMinutsOfsX,ClockPlugin->HourMinutsOfsY);
        
        if(ClockPlugin->AMPMShow==true){
          if(nowpm==false){
            DrawFontStr(&fontAMPM,"0",pbmbuf,ClockPlugin->AMPMOfsX,ClockPlugin->AMPMOfsY);
            }else{
            DrawFontStr(&fontAMPM,"1",pbmbuf,ClockPlugin->AMPMOfsX,ClockPlugin->AMPMOfsY);
          }
        }
      }
    }
  }
  
  if(ClockPlugin->SecondShow==true){
    if(LastTime.Second!=Time.Second){
      if(LastTime.Second!=(u32)-1) ledDrawRot(pbmbuf,(float)LastTime.Second/60*PI*2,84,ELM_Clear);
      for(int h=0;h<12;h++){
        ledDrawRot(pbmbuf,(float)h/12*PI*2,84,ELM_PowOff);
      }
      ledDrawRot(pbmbuf,(float)Time.Second/60*PI*2,84,ELM_PowOn);
      LastTime.Second=Time.Second;
      char str[256];
      snprintf(str,256,ClockPlugin->SecondFormat,Time.Second);
      DrawFontStr(&fontDigiSmall,str,pbmbuf,ClockPlugin->SecondOfsX,ClockPlugin->SecondOfsY);
    }
  }
  
  if(ClockPlugin->DateShow==true){
    if((LastTime.Year!=Time.Year)||(LastTime.Month!=Time.Month)||(LastTime.Day!=Time.Day)){
      LastTime.Year=Time.Year;
      LastTime.Month=Time.Month;
      LastTime.Day=Time.Day;
      char str[256+1];
      formdt_FormatDate(str,256,Time.Year,Time.Month,Time.Day);
      DrawFontStr(&fontDate,str,pbmbuf,ClockPlugin->DateOfsX,ClockPlugin->DateOfsY);
    }
  }
  
  if(ClockPlugin->TemperatureShow==true){
    if((VsyncPassedCount!=0)&&(VsyncPassedCount<6)){
      if(LastTime.Temperature12!=Time.Temperature12){
        LastTime.Temperature12=Time.Temperature12;
        char str[256];
        snprintf(str,256,ClockPlugin->TemperatureFormat,((float)Time.Temperature12) / 0x1000);
        DrawFontStr(&fontTemperature,str,pbmbuf,ClockPlugin->TemperatureOfsX,ClockPlugin->TemperatureOfsY);
      }
    }
    
  }
  
}
