
#include <stdio.h>

#include <NDS.h>

#include "plugin.h"
#include "plugin_def.h"

#include "inifile.h"

TGlobalINI GlobalINI;

void InitINI(void)
{
  {
    TiniClockPlugin *ClockPlugin=&GlobalINI.ClockPlugin;
    ClockPlugin->DateShow=1;
    ClockPlugin->DateOfsX=0;
    ClockPlugin->DateOfsY=0;
    ClockPlugin->AMPMShow=1;
    ClockPlugin->AMPMOfsX=0;
    ClockPlugin->AMPMOfsY=0;
    ClockPlugin->HourMinutsShow=1;
    strcpy(ClockPlugin->HourMinutsFormat,"%d:%d");
    ClockPlugin->HourMinutsOfsX=0;
    ClockPlugin->HourMinutsOfsY=0;
    ClockPlugin->SecondShow=1;
    strcpy(ClockPlugin->SecondFormat,":%d");
    ClockPlugin->SecondOfsX=0;
    ClockPlugin->SecondOfsY=0;
    ClockPlugin->TemperatureShow=1;
    strcpy(ClockPlugin->TemperatureFormat,"%f");
    ClockPlugin->TemperatureOfsX=0;
    ClockPlugin->TemperatureOfsY=0;
  }
}

static char section[128];
static u32 readline;

static void readsection(char *str)
{
  str++;
  
  u32 ofs;
  
  ofs=0;
  while(*str!=']'){
    if((128<=ofs)||(*str==0)){
      _consolePrintf("line%d error.\nThe section name doesn't end correctly.\n",readline);
      ShowLogHalt();
    }
    section[ofs]=*str;
    str++;
    ofs++;
  }
  section[ofs]=0;
}

static void readkey(char *str)
{
  if(section[0]==0){
    _consolePrintf("line%d error.\nThere is a key ahead of the section name.\n",readline);
//    ShowLogHalt();
    return;
  }
  
  char key[128],value[128];
  
  u32 ofs;
  
  ofs=0;
  while(*str!='='){
    if((128<=ofs)||(*str==0)){
      _consolePrintf("line%d error.\nThe key name doesn't end correctly.\n",readline);
      ShowLogHalt();
    }
    key[ofs]=*str;
    str++;
    ofs++;
  }
  key[ofs]=0;
  
  str++;
  
  ofs=0;
  while(*str!=0){
    if(128<=ofs){
      _consolePrintf("line%d error.\nThe value doesn't end correctly.\n",readline);
      ShowLogHalt();
    }
    value[ofs]=*str;
    str++;
    ofs++;
  }
  value[ofs]=0;
  
  s32 ivalue=atoi(value);
  bool bvalue;
  
  if(ivalue==0){
    bvalue=false;
    }else{
    bvalue=true;
  }
  
  bool match=false;
  
  if((match==false)&&(strcmp(section,"ClockPlugin")==0)){
    TiniClockPlugin *ClockPlugin=&GlobalINI.ClockPlugin;
    
    if((match==false)&&(strcmp(key,"BGColor")==0)){
      match=true;
      ClockPlugin->BGColor=ivalue;
    }
    if((match==false)&&(strcmp(key,"DateShow")==0)){
      match=true;
      ClockPlugin->DateShow=bvalue;
    }
    if((match==false)&&(strcmp(key,"DateOfsX")==0)){
      match=true;
      ClockPlugin->DateOfsX=ivalue;
    }
    if((match==false)&&(strcmp(key,"DateOfsY")==0)){
      match=true;
      ClockPlugin->DateOfsY=ivalue;
    }
    if((match==false)&&(strcmp(key,"AMPMShow")==0)){
      match=true;
      ClockPlugin->AMPMShow=bvalue;
    }
    if((match==false)&&(strcmp(key,"AMPMOfsX")==0)){
      match=true;
      ClockPlugin->AMPMOfsX=ivalue;
    }
    if((match==false)&&(strcmp(key,"AMPMOfsY")==0)){
      match=true;
      ClockPlugin->AMPMOfsY=ivalue;
    }
    if((match==false)&&(strcmp(key,"HourMinutsShow")==0)){
      match=true;
      ClockPlugin->HourMinutsShow=bvalue;
    }
    if((match==false)&&(strcmp(key,"HourMinutsFormat")==0)){
      match=true;
      strcpy(ClockPlugin->HourMinutsFormat,value);
    }
    if((match==false)&&(strcmp(key,"HourMinutsOfsX")==0)){
      match=true;
      ClockPlugin->HourMinutsOfsX=ivalue;
    }
    if((match==false)&&(strcmp(key,"HourMinutsOfsY")==0)){
      match=true;
      ClockPlugin->HourMinutsOfsY=ivalue;
    }
    if((match==false)&&(strcmp(key,"SecondShow")==0)){
      match=true;
      ClockPlugin->SecondShow=bvalue;
    }
    if((match==false)&&(strcmp(key,"SecondFormat")==0)){
      match=true;
      strcpy(ClockPlugin->SecondFormat,value);
    }
    if((match==false)&&(strcmp(key,"SecondOfsX")==0)){
      match=true;
      ClockPlugin->SecondOfsX=ivalue;
    }
    if((match==false)&&(strcmp(key,"SecondOfsY")==0)){
      match=true;
      ClockPlugin->SecondOfsY=ivalue;
    }
    if((match==false)&&(strcmp(key,"TemperatureShow")==0)){
      match=true;
      ClockPlugin->TemperatureShow=bvalue;
    }
    if((match==false)&&(strcmp(key,"TemperatureFormat")==0)){
      match=true;
      strcpy(ClockPlugin->TemperatureFormat,value);
    }
    if((match==false)&&(strcmp(key,"TemperatureOfsX")==0)){
      match=true;
      ClockPlugin->TemperatureOfsX=ivalue;
    }
    if((match==false)&&(strcmp(key,"TemperatureOfsY")==0)){
      match=true;
      ClockPlugin->TemperatureOfsY=ivalue;
    }
    
  }
  
  if(match==false){
    _consolePrintf("line%d error.\ncurrent section [%s] unknown key=%s\n",readline,section,key);
//    ShowLogHalt();
  }
  
//  _consolePrintf("key=%s value=%s\n",key,value);
}

static void internal_LoadGlobalINI(char *pini,u32 inisize)
{
  section[0]=0;
  readline=0;
  
  u32 iniofs=0;
  
  while(iniofs<inisize){
    
    readline++;
    
    u32 linelen=0;
    
    // Calc Line Length
    {
      char *s=&pini[iniofs];
      
      while(0x20<=*s){
        linelen++;
        s++;
        if(inisize<=(iniofs+linelen)) break;
      }
      *s=0;
    }
    
    if(linelen!=0){
      char c=pini[iniofs];
      if((c==';')||(c=='/')||(c=='!')){
        // comment line
        }else{
        if(c=='['){
          readsection(&pini[iniofs]);
          }else{
          readkey(&pini[iniofs]);
        }
      }
    }
    
    iniofs+=linelen;
    
    // skip NULL,CR,LF
    {
      char *s=&pini[iniofs];
      
      while(*s<0x20){
        iniofs++;
        s++;
        if(inisize<=iniofs) break;
      }
    }
    
  }
}

void LoadINI(char *data,int size)
{
  if((data==NULL)||(size==0)) return;
  
  internal_LoadGlobalINI(data,size);
}

