
#include <stdio.h>

#include <NDS.h>

#include "_const.h"

#include "plugin.h"
#include "plugin_def.h"

// CallBack from plugin_dll.cpp

void cbLoadLibrary(void)
{
}

void cbFreeLibrary(void)
{
}

void cbQueryInterfaceLibrary(void)
{
}

// -----------------

#define BI_RGB (0)
#define BI_RLE8 (1)
#define BI_RLE4 (2)
#define BI_Bitfields (3)

typedef struct {
  u8 bfType[2];
  u32 bfSize;
  u16 bfReserved1;
  u16 bfReserved2;
  u32 bfOffset;
  u32 biSize;
  u32 biWidth;
  u32 biHeight;
  u16 biPlanes;
  u16 biBitCount;
  u32 biCopmression;
  u32 biSizeImage;
  u32 biXPixPerMeter;
  u32 biYPixPerMeter;
  u32 biClrUsed;
  u32 biCirImportant;
  
  u32 Palette[256];
  
  u32 DataWidth;
} TBMPHeader;

static TBMPHeader *pBMPHeader=NULL;

static int FileHandle;
static u8 *pReadBuf=NULL;

typedef struct {
  u16 *bm;
} TBMBuf;

static TBMBuf *pBMBuf=NULL;
static s32 LastReadY;

static u32 LastOffsetY;

// ------------------------------------------------------------------------------------

static u8 Get8bit(void)
{
  u8 res;
  fread(&res,1,1,FileHandle);
  return(res);
}

static u16 Get16bit(void)
{
  u16 res;
  fread(&res,2,1,FileHandle);
  return(res);
}

static u32 Get32bit(void)
{
  u32 res;
  fread(&res,4,1,FileHandle);
  return(res);
}

static bool GetBMPHeader(TBMPHeader *pBMPHeader)
{
  fseek(FileHandle,0,SEEK_SET);
  
  pBMPHeader->bfType[0]=Get8bit();
  pBMPHeader->bfType[1]=Get8bit();
  pBMPHeader->bfSize=Get32bit();
  pBMPHeader->bfReserved1=Get16bit();
  pBMPHeader->bfReserved2=Get16bit();
  pBMPHeader->bfOffset=Get32bit();
  pBMPHeader->biSize=Get32bit();
  pBMPHeader->biWidth=Get32bit();
  pBMPHeader->biHeight=Get32bit();
  pBMPHeader->biPlanes=Get16bit();
  pBMPHeader->biBitCount=Get16bit();
  pBMPHeader->biCopmression=Get32bit();
  pBMPHeader->biSizeImage=Get32bit();
  pBMPHeader->biXPixPerMeter=Get32bit();
  pBMPHeader->biYPixPerMeter=Get32bit();
  pBMPHeader->biClrUsed=Get32bit();
  pBMPHeader->biCirImportant=Get32bit();
  
  if((pBMPHeader->biBitCount==4)||(pBMPHeader->biBitCount==8)){
    for(int idx=0;idx<256;idx++){
      pBMPHeader->Palette[idx]=Get32bit();
    }
    }else{
    for(int idx=0;idx<256;idx++){
      pBMPHeader->Palette[idx]=0;
    }
  }
  
  if((pBMPHeader->bfType[0]!='B')||(pBMPHeader->bfType[1]!='M')){
    _consolePrintf("Error MagicID!=BM");
    return(false);
  }
  
  if(pBMPHeader->biCopmression!=BI_RGB){
    _consolePrintf("Error notsupport Compression");
    return(false);
  }
  
  if(pBMPHeader->biHeight>=0x80000000){
    _consolePrintf("Error notsupport OS/2 format");
    return(false);
  }
  
  if(pBMPHeader->biPlanes!=1){
    _consolePrintf("Error notsupport Planes!=1");
    return(false);
  }
  
  pBMPHeader->DataWidth=0;
  
  switch(pBMPHeader->biBitCount){
    case 1:
      _consolePrintf("Error notsupport 1bitcolor.");
      pBMPHeader->DataWidth=(pBMPHeader->biWidth+7)/8;
      break;
    case 4:
      pBMPHeader->DataWidth=(pBMPHeader->biWidth+1)/2;
      break;
    case 8:
      pBMPHeader->DataWidth=pBMPHeader->biWidth*1;
      break;
    case 16:
      _consolePrintf("Error notsupport 16bitcolor.");
      return(false);
    case 24:
      pBMPHeader->DataWidth=pBMPHeader->biWidth*3;
      break;
    case 32:
      _consolePrintf("Error notsupport 32bitcolor.");
      return(false);
    default:
      _consolePrintf("Error Unknown xxBitColor.");
      return(false);
  }
  
  if((pBMPHeader->DataWidth&3)!=0){
    pBMPHeader->DataWidth+=4-(pBMPHeader->DataWidth&3);
  }
  
  return(true);
}

void BMP_ShowFrameInfo(TBMPHeader *pBMPHeader)
{
  _consolePrintf("FileSize=%d\n",pBMPHeader->bfSize);
  _consolePrintf("Size=(%d,%d) %dpixel\n",pBMPHeader->biWidth,pBMPHeader->biHeight,pBMPHeader->biWidth*pBMPHeader->biHeight);
  _consolePrintf("Planes=%d\n",pBMPHeader->biPlanes);
  _consolePrintf("BitCount=%d\n",pBMPHeader->biBitCount);
  _consolePrintf("Compression=%d\n",pBMPHeader->biCopmression);
  _consolePrintf("BitmapOffset=0x%x\n",pBMPHeader->bfOffset);
  _consolePrintf("DataWidth=%d\n",pBMPHeader->DataWidth);
}

#define Gravity(c,cg) { \
  c+=cg; \
  cg=c&7; \
  c=c>>3; \
  if((c&(~0x1f))!=0) c=(c<0) ? 0x00 : 0x1f; \
}

u16* BMP_GetBMBuf(s32 y,s32 vector)
{
  if(y<0) return(NULL);
  if(pBMPHeader->biHeight<=y) return(NULL);
  
  u16 *pSrcBM;
  
  pSrcBM=pBMBuf[y].bm;
  
  if(pSrcBM!=NULL) return(pSrcBM);
  
  pSrcBM=(u16*)safemalloc(pBMPHeader->biWidth*2);
  
  if(pSrcBM==NULL){
    s32 topy=-1;
    switch(vector){
      case -1: {
        for(s32 y=pBMPHeader->biHeight-1;y>=0;y--){
          if(pBMBuf[y].bm!=NULL){
            topy=y;
            break;
          }
        }
      } break;
      case 0: case 1: {
        for(u32 y=0;y<pBMPHeader->biHeight;y++){
          if(pBMBuf[y].bm!=NULL){
            topy=y;
            break;
          }
        }
      }
    }
    if(topy==-1){
      _consolePrintf("BMP ReadCache: out of memory.\n");
      return(NULL);
    }
    if(extmem_Exists(topy)==false){
      if(extmem_Alloc(topy,pBMPHeader->biWidth*2)==true){
        extmem_Write(topy,pBMBuf[topy].bm,pBMPHeader->biWidth*2);
      }
    }
    pSrcBM=pBMBuf[topy].bm;
    pBMBuf[topy].bm=NULL;
  }
  
  pBMBuf[y].bm=pSrcBM;
  
  if(extmem_Exists(y)==true){
    extmem_Read(y,pBMBuf[y].bm,pBMPHeader->biWidth*2);
    return(pBMBuf[y].bm);
  }
  
  if(LastReadY!=y){
    LastReadY=y;
    fseek(FileHandle,pBMPHeader->bfOffset+((pBMPHeader->biHeight-y-1)*pBMPHeader->DataWidth),SEEK_SET);
  }
  LastReadY--;
  
  fread(pReadBuf,1,pBMPHeader->DataWidth,FileHandle);
  
  u32 gr=0,gg=0,gb=0;
  u32 Width=pBMPHeader->biWidth;
  
  switch(pBMPHeader->biBitCount){
    case 1: {
      u8 *PaletteTable=(u8*)pBMPHeader->Palette;
      for(u32 x=0;x<Width;x+=8){
        u32 data=pReadBuf[x/8];
        for(u32 b=0;b<8;b++){
          if((x+b)<Width){
            u32 bit;
            
            bit=(data>>(7-b))&1;
            
            pSrcBM[x+b]=RGB15(bit*31,bit*31,bit*31) | BIT(15);
          }
        }
      }
      break;
    }
    case 4: {
      u8 *PaletteTable=(u8*)pBMPHeader->Palette;
      for(u32 x=0;x<Width;x+=2){
        u32 data=pReadBuf[x/2];
        
        u8 *pal;
        u32 r,g,b;
        
        pal=&PaletteTable[(data>>4)*4];
        
        b=pal[0];
        g=pal[1];
        r=pal[2];
        
        Gravity(b,gb);
        Gravity(g,gg);
        Gravity(r,gr);
        
        pSrcBM[x]=RGB15(r,g,b) | BIT(15);
        
        if((x+1)<Width){
          pal=&PaletteTable[(data&0x0f)*4];
          
          b=pal[0];
          g=pal[1];
          r=pal[2];
          
          Gravity(b,gb);
          Gravity(g,gg);
          Gravity(r,gr);
          
          pSrcBM[x+1]=RGB15(r,g,b) | BIT(15);
        }
      }
      break;
    }
    case 8: {
      u8 *PaletteTable=(u8*)pBMPHeader->Palette;
      for(u32 x=0;x<Width;x++){
        u8 *pal;
        u32 r,g,b;
        
        pal=&PaletteTable[pReadBuf[x]*4];
        
        b=pal[0];
        g=pal[1];
        r=pal[2];
        
        Gravity(b,gb);
        Gravity(g,gg);
        Gravity(r,gr);
        
        pSrcBM[x]=RGB15(r,g,b) | BIT(15);
      }
      break;
    }
    case 24: {
      for(u32 x=0;x<Width;x++){
        u32 r,g,b;
        
        b=pReadBuf[x*3+0];
        g=pReadBuf[x*3+1];
        r=pReadBuf[x*3+2];
        Gravity(b,gb);
        Gravity(g,gg);
        Gravity(r,gr);
        
        pSrcBM[x]=RGB15(r,g,b) | BIT(15);
      }
      break;
    }
  }
  
  return(pSrcBM);
}

#undef Gravity

// ------------------------------------------------------------------------------------

bool Start(int _FileHandle)
{
  if(_FileHandle==0){
    _consolePrintf("FileHandle is NULL\n");
    return(false);
  }
  
  FileHandle=_FileHandle;
  
  pBMPHeader=(TBMPHeader*)safemalloc(sizeof(TBMPHeader));
  if(pBMPHeader==NULL){
    _consolePrintf("pBMPHeader malloc error.\n");
    return(false);
  }
  
  memset(pBMPHeader,0,sizeof(TBMPHeader));
  
  if(GetBMPHeader(pBMPHeader)==false){
    _consolePrintf("BMP LoadError.\n");
    return(false);
  }
  
  BMP_ShowFrameInfo(pBMPHeader);
  
  extmem_SetCount(pBMPHeader->biHeight);
  
  pReadBuf=(u8*)safemalloc(pBMPHeader->DataWidth*2);
  
  pBMBuf=(TBMBuf*)safemalloc(pBMPHeader->biHeight*sizeof(TBMBuf));
  memset(pBMBuf,0,pBMPHeader->biHeight*sizeof(TBMBuf));
  
  LastOffsetY=0;
  
  return(true);
}

void Free(void)
{
  if(pBMBuf!=NULL){
    for(u32 y=0;y<pBMPHeader->biHeight;y++){
      if(pBMBuf[y].bm!=NULL){
        safefree(pBMBuf[y].bm); pBMBuf[y].bm=NULL;
      }
    }
    safefree(pBMBuf); pBMBuf=NULL;
  }
  if(pReadBuf!=NULL){
    safefree(pReadBuf); pReadBuf=NULL;
  }
  if(pBMPHeader!=NULL){
    safefree(pBMPHeader); pBMPHeader=NULL;
  }
  
  FileHandle=0;
}

typedef struct {
  s32 x,y,w,h;
} TRect;

void RefreshScreen(u32 OffsetX,u32 OffsetY,u16 *pBuf,u32 BufWidth,u32 DrawWidth,u32 DrawHeight,bool HalfMode)
{
//  _consolePrintf("%d,%d 0x%x,%d %d,%d %d\n",OffsetX,OffsetY,(u32)pBuf,BufWidth,DrawWidth,DrawHeight,HalfMode);
  
  s32 Vector;
  
  if(LastOffsetY==OffsetY){
    Vector=0;
    }else{
    if(LastOffsetY<OffsetY){
      Vector=1;
      }else{
      Vector=-1;
    }
  }
  LastOffsetY=OffsetY;
  
  u32 BaseX=OffsetX,BaseY=OffsetY;
  
  TRect ClientRect;
  
  ClientRect.x=0;
  ClientRect.y=0;
  ClientRect.h=DrawHeight;
  ClientRect.w=DrawWidth;
  
  s32 DrawH,DrawW;
  
  DrawW=pBMPHeader->biWidth-BaseX;
  DrawH=pBMPHeader->biHeight-BaseY;
  
  if(DrawW>ClientRect.w) DrawW=ClientRect.w;
  if(DrawH>ClientRect.h) DrawH=ClientRect.h;
  
  if(HalfMode==false){
    for(s32 y=DrawH-1;y>=0;y--){
      u16 *pSrcBM=BMP_GetBMBuf(BaseY+y,Vector);
      u16 *pDstBM=&pBuf[ClientRect.x+((ClientRect.y+y)*BufWidth)];
      
      if(pSrcBM!=NULL){
        pSrcBM+=BaseX;
        MemCopy16DMA3(pSrcBM,pDstBM,DrawW*2);
      }
    }
    
    }else{ // half
    
    for(s32 y=(DrawH/2)-1;y>=0;y--){
      u16 *pSrcBM1;
      u16 *pSrcBM0;
      u16 *pDstBM=&pBuf[ClientRect.x+((ClientRect.y+y)*BufWidth)];
      
      pSrcBM1=BMP_GetBMBuf(BaseY+(y*2)+1,Vector);
      pSrcBM0=BMP_GetBMBuf(BaseY+(y*2)+0,Vector);
      
      if((pSrcBM0!=NULL)&&(pSrcBM1!=NULL)){
        pSrcBM0+=BaseX;
        pSrcBM1+=BaseX;
        for(int x=0;x<DrawW/2;x++){
          u32 r=0,g=0,b=0;
          u32 rgb;
          
          rgb=pSrcBM0[x*2+0];
          r+=(rgb >> 0) & 0x1f; g+=(rgb >> 5) & 0x1f; b+=(rgb >> 10) & 0x1f;
          rgb=pSrcBM0[x*2+1];
          r+=(rgb >> 0) & 0x1f; g+=(rgb >> 5) & 0x1f; b+=(rgb >> 10) & 0x1f;
          rgb=pSrcBM1[x*2+0];
          r+=(rgb >> 0) & 0x1f; g+=(rgb >> 5) & 0x1f; b+=(rgb >> 10) & 0x1f;
          rgb=pSrcBM1[x*2+1];
          r+=(rgb >> 0) & 0x1f; g+=(rgb >> 5) & 0x1f; b+=(rgb >> 10) & 0x1f;
          
          pDstBM[x]=BIT15 | RGB15(r/4,g/4,b/4);
        }
      }
      
    }
  }
}

s32 GetWidth(void)
{
  return(pBMPHeader->biWidth);
}

s32 GetHeight(void)
{
  return(pBMPHeader->biHeight);
}

int GetInfoIndexCount(void)
{
  return(4);
}

bool GetInfoStrL(int idx,char *str,int len)
{
  switch(idx){
    case 0: snprintf(str,len,"Pixels=%dx%dx%dbitColor",pBMPHeader->biWidth,pBMPHeader->biHeight,pBMPHeader->biBitCount); return(true); break;
    case 1: snprintf(str,len,"Planes=%d",pBMPHeader->biPlanes); return(true); break;
    case 2: snprintf(str,len,"Copmression=%d",pBMPHeader->biCopmression); return(true); break;
    case 3: snprintf(str,len,"PixelsPerMeter=%dx%d",pBMPHeader->biXPixPerMeter,pBMPHeader->biYPixPerMeter); return(true); break;
  }
  return(false);
}

bool GetInfoStrW(int idx,UnicodeChar *str,int len)
{
  return(false);
}

bool GetInfoStrUTF8(int idx,char *str,int len)
{
  return(false);
}

