
#include <stdio.h>

#include <NDS.h>

#include "_const.h"

#include "plugin.h"
#include "plugin_def.h"

#include "png.h"
#include "plug_png_interface.h"

// CallBack from plugin_dll.cpp

void cbLoadLibrary(void)
{
}

void cbFreeLibrary(void)
{
}

void cbQueryInterfaceLibrary(void)
{
}

// -----------------

typedef struct {
  u16 *bm;
} Tbm;

static s32 biWidth=0,biHeight=0;
static s32 DecodedHeight=0;
static Tbm *DecodedBufs=NULL;

static u32 LastOffsetY;

bool LoadPNG(int FileHandle)
{
  if(PNGINT_Init(FileHandle)==false){
    return(false);
  }
  
  biWidth=PNGINT_GetWidth();
  biHeight=PNGINT_GetHeight();
  DecodedHeight=-1;
  
  DecodedBufs=(Tbm*)safemalloc(biHeight*sizeof(Tbm));
  
  for(s32 y=0;y<biHeight;y++){
    DecodedBufs[y].bm=NULL;
  }
  
  return(true);
}

static int BackupFileHandle;

static bool InitialDrawFlag;

// ------------------------------------------------------------------------------------

void RestartPNG(void)
{
  if(DecodedBufs!=NULL){
    for(s32 y=0;y<biHeight;y++){
      if(DecodedBufs[y].bm!=NULL){
        free(DecodedBufs[y].bm); DecodedBufs[y].bm=NULL;
      }
    }
    free(DecodedBufs); DecodedBufs=NULL;
  }
  
  PNGINT_Free();
  
  DecodedHeight=-1;
  
  fseek(BackupFileHandle,0,SEEK_SET);
  
  if(PNGINT_Init(BackupFileHandle)==false){
    _consolePrintf("PNG can't restart.\n");
    ShowLogHalt();
  }
  
  DecodedBufs=(Tbm*)safemalloc(biHeight*sizeof(Tbm));
  
  for(s32 y=0;y<biHeight;y++){
    DecodedBufs[y].bm=NULL;
  }
}

static u16* PNG_GetDecodedBuf(s32 y,s32 vector)
{
  if(y<0) return(NULL);
  if(biHeight<=y) return(NULL);
  
  u16 *pSrcBM;
  
  pSrcBM=DecodedBufs[y].bm;
  
  if(pSrcBM!=NULL) return(pSrcBM);
  
  pSrcBM=(u16*)safemalloc(biWidth*2);
  
  if(pSrcBM==NULL){
    s32 topy=-1;
    switch(vector){
      case -1: {
        for(s32 y=biHeight-1;y>=0;y--){
          if(DecodedBufs[y].bm!=NULL){
            topy=y;
            break;
          }
        }
      } break;
      case 0: case 1: {
        for(u32 y=0;y<(u32)biHeight;y++){
          if(DecodedBufs[y].bm!=NULL){
            topy=y;
            break;
          }
        }
      }
    }
    if(topy==-1){
      _consolePrintf("PNG DecodeCache: out of memory.\n");
      return(NULL);
    }
    if(extmem_Exists(topy)==false){
      if(extmem_Alloc(topy,biWidth*2)==true){
        extmem_Write(topy,DecodedBufs[topy].bm,biWidth*2);
      }
    }
    pSrcBM=DecodedBufs[topy].bm;
    DecodedBufs[topy].bm=NULL;
  }
  
  DecodedBufs[y].bm=pSrcBM;
  
  if(extmem_Exists(y)==true){
    extmem_Read(y,DecodedBufs[y].bm,biWidth*2);
    return(DecodedBufs[y].bm);
  }
  
  if(y<=DecodedHeight) RestartPNG();
  
  if(DecodedHeight<(y-1)){
    int PrgBase=DecodedHeight;
    MWin_ProgressShow("SkipImage",y-PrgBase);
    
    while(DecodedHeight<(y-1)){
      if((DecodedHeight&15)==0) MWin_ProgressSetPos(DecodedHeight-PrgBase);
      PNGINT_GetNextLine(NULL);
      DecodedHeight++;
    }
    
    MWin_ProgressHide();
  }
  
  if(DecodedHeight!=(y-1)){
    _consolePrintf("seek error. %d!=(%d-1)\n",DecodedHeight,y);
    ShowLogHalt();
  }
  
  PNGINT_GetNextLine(pSrcBM);
  DecodedHeight=y;
  
  return(pSrcBM);
}

// ------------------------------------------------------------------------------------

bool Start(int FileHandle)
{
  BackupFileHandle=FileHandle;
  InitialDrawFlag=true;
  
  if(LoadPNG(FileHandle)==false){
    _consolePrintf("PNG LoadError.\n");
    return(false);
  }
  
  extmem_SetCount(biHeight);
  
  LastOffsetY=0;
  
  return(true);
}

void Free(void)
{
  if(DecodedBufs!=NULL){
    for(s32 y=0;y<biHeight;y++){
      if(DecodedBufs[y].bm!=NULL){
        free(DecodedBufs[y].bm); DecodedBufs[y].bm=NULL;
      }
    }
    free(DecodedBufs); DecodedBufs=NULL;
  }
  
  PNGINT_Free();
  
  BackupFileHandle=0;
  
  biWidth=0;
  biHeight=0;
  DecodedHeight=0;
}

typedef struct {
  s32 x,y,w,h;
} TRect;

void RefreshScreen(u32 OffsetX,u32 OffsetY,u16 *pBuf,u32 BufWidth,u32 DrawWidth,u32 DrawHeight,bool HalfMode)
{
//  _consolePrintf("%d,%d 0x%x,%d %d,%d %d\n",OffsetX,OffsetY,(u32)pBuf,BufWidth,DrawWidth,DrawHeight,HalfMode);
  
  s32 Vector;
  
  if(LastOffsetY==OffsetY){
    Vector=0;
    }else{
    if(LastOffsetY<OffsetY){
      Vector=1;
      }else{
      Vector=-1;
    }
  }
  LastOffsetY=OffsetY;
  
  u32 BaseX=OffsetX,BaseY=OffsetY;
  
  TRect ClientRect;
  
  ClientRect.x=0;
  ClientRect.y=0;
  ClientRect.h=DrawHeight;
  ClientRect.w=DrawWidth;
  
  s32 DrawH,DrawW;
  
  DrawH=biHeight-BaseY-1;
  DrawW=biWidth-BaseX-1;
  
  if(DrawH>ClientRect.h) DrawH=ClientRect.h;
  if(DrawW>ClientRect.w) DrawW=ClientRect.w;
  
  if(InitialDrawFlag==true) MWin_ProgressShow("Decode...",DrawH);
  
  if(HalfMode==false){
    for(s32 y=0;y<DrawH;y++){
      u16 *pSrcBM=PNG_GetDecodedBuf(BaseY+y,Vector);
      u16 *pDstBM=&pBuf[ClientRect.x+((ClientRect.y+y)*BufWidth)];
      
      if(pSrcBM==NULL){
        for(s32 x=0;x<DrawW;x++){
          pDstBM[x]=0;
        }
        }else{
        pSrcBM+=BaseX;
        for(s32 x=0;x<DrawW;x++){
          pDstBM[x]=pSrcBM[x];
        }
      }
      
      if((InitialDrawFlag==true)&&((y&3)==0)) MWin_ProgressSetPos(y);
    }
    
    }else{ // half
    
    for(s32 y=0;y<DrawH/2;y++){
      u16 *pSrcBM0=PNG_GetDecodedBuf(BaseY+(y*2)+0,Vector);
      u16 *pSrcBM1=PNG_GetDecodedBuf(BaseY+(y*2)+1,Vector);
      u16 *pDstBM=&pBuf[ClientRect.x+((ClientRect.y+y)*BufWidth)];
      
      if((pSrcBM0!=NULL)&&(pSrcBM1!=NULL)){
        pSrcBM0+=BaseX;
        pSrcBM1+=BaseX;
        for(s32 x=0;x<DrawW/2;x++){
          u32 r=0,g=0,b=0;
          u32 rgb;
          
          rgb=pSrcBM0[x*2+0];
          r+=(rgb >> 0) & 0x1f; g+=(rgb >> 5) & 0x1f; b+=(rgb >> 10) & 0x1f;
          rgb=pSrcBM0[x*2+1];
          r+=(rgb >> 0) & 0x1f; g+=(rgb >> 5) & 0x1f; b+=(rgb >> 10) & 0x1f;
          rgb=pSrcBM1[x*2+0];
          r+=(rgb >> 0) & 0x1f; g+=(rgb >> 5) & 0x1f; b+=(rgb >> 10) & 0x1f;
          rgb=pSrcBM1[x*2+1];
          r+=(rgb >> 0) & 0x1f; g+=(rgb >> 5) & 0x1f; b+=(rgb >> 10) & 0x1f;
          
          pDstBM[x]=BIT15 | RGB15(r/4,g/4,b/4);
        }
      }
      
      if((InitialDrawFlag==true)&&((y&3)==0)) MWin_ProgressSetPos(y*2);
    }
  }
  
  if(InitialDrawFlag==true){
    MWin_ProgressHide();
    InitialDrawFlag=false;
  }
}

s32 GetWidth(void)
{
  return(biWidth);
}

s32 GetHeight(void)
{
  return(biHeight);
}

int GetInfoIndexCount(void)
{
  return(1);
}

bool GetInfoStrL(int idx,char *str,int len)
{
  switch(idx){
    case 0: snprintf(str,len,"%s",PNGINT_GetInfoStr()); return(true); break;
  }
  return(false);
}

bool GetInfoStrW(int idx,UnicodeChar *str,int len)
{
  return(false);
}

bool GetInfoStrUTF8(int idx,char *str,int len)
{
  return(false);
}

